#include <assert.h>
#include <iostream>
#include <iomanip>
#include <vector>
using namespace std;
#include "PV.h"

void checkSame (PV a, PV b) {
  static double maxDistAB;
  double distAB = a.distance(b);
  if (distAB > maxDistAB) {
    cerr << "maxDistAB " << distAB << endl;
    maxDistAB = distAB;
  }
}

void checkPositive (double x) {
  static double maxNegative;
  if (x < -maxNegative) {
    cerr << "maxNegative " << -x << endl;
    maxNegative = -x;
  }
}

// b is between a and c
bool between (PV a, PV b, PV c) {
  if (a % c >= 0)
    return a % b > 0 && b % c > 0;
  else
    return a % b > 0 || b % c > 0;
}

istream &operator>> (istream &in, PV &p) {
  return in >> p.x >> p.y;
}

class Vert {
public:
  PV p, n;
  void read (istream &in) {
    in >> p >> n;
  }
};

class Edge {
public:
  PV c;
  double r;
  Vert *t, *h;
  void read (istream &in, Vert *vOb) {
    int iT, iH;
    in >> c >> r >> iT >> iH;
    t = &vOb[iT];
    h = &vOb[iH];
  }
  PV getP (PV n) {
    return c + n * r;
  }
  void check () {
    checkSame(t->p, getP(t->n));
    checkSame(h->p, getP(h->n));
  }
  void calcC () {
    PV th = h->p - t->p;
    double x = th.length() / 2;
    PV n = th.rot90().unit();
    PV m = (t->p + h->p) / 2;
    double y = sqrt(r * r - x * x);
    c = r > 0 ? m + n * y : m - n * y;
    checkSame(PV(c.distance(t->p), c.distance(h->p)), PV(fabs(r), fabs(r)));
    t->n = (t->p - c) / r;
    h->n = (h->p - c) / r;
    checkSame(PV(t->n.length(), h->n.length()), PV(1, 1));
  }
  Vert *getAB (int ab) { return ab == (r < 0) ? t : h; }
};

class Sub;

class Event {
public:
  PV q, p;
  int index;
  vector<Sub *> aSubs, bSubs;
  void read (istream &in, int i) {
    in >> q >> p;
    index = i;
  }
  bool between (Event *a, Event *b) {
    if (a->index <= b->index)
      return a->index <= index && index <= b->index;
    else
      return !(b->index < index && index < a->index);
  }
  void addSub (Sub *sub);
  void findNeighbors ();
};

class Pnt {
public:
  Event *a, *b;
  virtual PV getP (PV q) = 0;
};

class Sum : public Pnt {
public:
  Vert *v;
  Edge *e;
  virtual PV getN (PV q) = 0;
};

class SVE : public Sum {
public:
  void read (istream &in, Vert *vOb, Edge *eRo, Event *events) {
    int iV, iE, iA, iB;
    in >> iV >> iE >> iA >> iB;
    v = &vOb[iV];
    e = &eRo[iE];
    a = &events[iA];
    b = &events[iB];
  }
  PV getP (PV q) {
    return v->p + q.times(e->c) + v->n * e->r;
  }
  PV getN (PV q) {
    return v->n;
  }
};

class SEV : public Sum {
public:
  void read (istream &in, Edge *eOb, Vert *vRo, Event *events) {
    int iE, iV, iA, iB;
    in >> iE >> iV >> iA >> iB;
    e = &eOb[iE];
    v = &vRo[iV];
    a = &events[iA];
    b = &events[iB];
  }
  PV getP (PV q) {
    return e->c + q.times(v->n) * e->r + q.times(v->p);
  }
  PV getN (PV q) {
    return q.times(v->n);
  }
};

class SEE {
public:
  Edge *eO, *eR;
  Sum *t, *h;
  Event *a, *b;
  void read (istream &in, Edge *eOb, Edge *eRo, SVE *sves, SEV *sevs,
             Event *events) {
    int ieO, ieR, tType, iT, hType, iH, iA, iB;
    in >> ieO >> ieR >> tType >> iT >> hType >> iH >> iA >> iB;
    eO = &eOb[ieO];
    eR = &eRo[ieR];
    if (tType == 0)
      t = &sves[iT];
    else
      t = &sevs[iT];
    if (hType == 0)
      h = &sves[iH];
    else
      h = &sevs[iH];
    a = &events[iA];
    b = &events[iB];
  }
  PV getC (PV q) { return eO->c + q.times(eR->c); }
  double getR () { return eO->r + eR->r; }
  PV getP (PV q, PV n) { return getC(q) + n * getR(); }

  void checkRandomQ () {
    PV q(0.4479799);
    PV tt = t->getP(q);
    PV tee = getP(q, t->getN(q));
    checkSame(tt, tee);
    PV hh = h->getP(q);
    PV hee = getP(q, h->getN(q));
    checkSame(hh, hee);
  }

  void draw (ostream &out, SEE *sees, PV q) {
    if (!between(a->q, q, b->q))
      return;
    out << "2 1 0 1 0 7 " << (this - sees)
        << " -1 -1 0.000 0 0 -1 0 0 ";
    double r = getR();
    PV c = getC(q);
    PV tn = t->getN(q);
    PV hn = h->getN(q);
    double talpha = atan2(tn.y, tn.x);
    double halpha = atan2(hn.y, hn.x);
    double step = 0.1;
    if (r > 0 && halpha < talpha)
      halpha += 2*M_PI;
    if (r < 0 && halpha > talpha) {
      halpha -= 2*M_PI;
      step = -step;
    }

    vector<PV> ps;
    ps.push_back(t->getP(q));
    double alpha = talpha + step;
    while ((alpha - talpha) * (alpha - halpha) < 0) {
      ps.push_back(c + PV(alpha) * r);
      alpha += step;
    }
    ps.push_back(h->getP(q));

    out << ps.size() << endl << "\t";
    for (int i = 0; i < ps.size(); i++) {
      PV p = ps[i];
      out << " " << (int) rint(1200 * p.x)
          << " " << (int) rint(1200 * p.y);
    }
    cout << endl;
  }
};

class Intn : public Pnt {
public:
  Edge *eO[2], *eR[2];
  void read (istream &in, Edge *eOb, Edge *eRo, Event *events) {
    int ieO[2], ieR[2], iA, iB;
    in >> ieO[0] >> ieR[0] >> ieO[1] >> ieR[1] >> iA >> iB;
    eO[0] = &eOb[ieO[0]];
    eR[0] = &eRo[ieR[0]];
    eO[1] = &eOb[ieO[1]];
    eR[1] = &eRo[ieR[1]];
    a = &events[iA];
    b = &events[iB];
  }

  PV getP (PV q) {
    PV c[] = { eO[0]->c + q.times(eR[0]->c), eO[1]->c + q.times(eR[1]->c) };
    double s[] = { eO[0]->r + eR[0]->r, eO[1]->r + eR[1]->r };
    double r[] = { fabs(s[0]), fabs(s[1]) };
    int i0 = (r[0] > r[1]);
    int i1 = !i0;
    PV cc = c[i1] - c[i0];
    double d = cc.length();
    double x;

    if (r[i0] < d) {
      double e = d - r[i1];
      checkPositive(r[i0] + e);
      checkPositive(r[i0] - e);
      double e1 = (r[i0] * r[i0] - e * e) / d / 2;
      x = e + e1;
    }
    else {
      double f = r[i1] - r[i0];
      checkPositive(d - f);
      if (f > d)
        f = d;
      double f1 = (d * d - f * f) / d / 2;
      double f2 = f / d;
      double f3 = r[i0] * f2;
      x = f1 - f3;
    }

    double y2 = r[i0] * r[i0] - x * x;
    checkPositive(y2);
    double y = y2 < 0 ? 0 : sqrt(y2);

    PV u = cc / d;
    PV v = u.rot90();
    PV p[2];
    p[i0] = c[i0] + u * x + v * y;
    p[i1] = c[i0] + u * x - v * y;

    for (int i = 0; i < 2; i++) {
      PV pcl(p[i].distance(c[0]), p[i].distance(c[1]));
      PV rr(r[0], r[1]);
      checkSame(pcl, rr);
    }

    if ((s[0] > 0) == (s[1] > 0))
      return p[0];
    else
      return p[1];
  }

  void check () {
    getP(a->q);
    getP(b->q);
  }
};
  
class Sub {
public:
  Edge *eO, *eR;
  Pnt *t, *h;
  Event *a, *b;
  Sub *at, *ah, *bt, *bh;
  Sub () : at(0), ah(0), bt(0), bh(0) {}
  void read (istream &in, Edge *eOb, Edge *eRo,
             SVE *sves, SEV *sevs, Intn *intns, Event *events) {
    int ieO, ieR, tType, iT, hType, iH, iA, iB;
    in >> ieO >> ieR >> tType >> iT >> hType >> iH >> iA >> iB;
    eO = &eOb[ieO];
    eR = &eRo[ieR];
    if (tType == 0)
      t = &sves[iT];
    else if (tType == 1)
      t = &sevs[iT];
    else
      t = &intns[iT];
    if (hType == 0)
      h = &sves[iH];
    else if (hType == 1)
      h = &sevs[iH];
    else
      h = &intns[iH];
    a = &events[iA];
    b = &events[iB];
  }

  void check (SEE *sees) {
    SEE *see = 0;
    for (int i = 0; i < 3; i++)
      assert(sees[i].eO == eO && sees[i].eR == eR);
    assert(a->between(sees[0].a, sees[2].b));
    assert(b->between(sees[0].a, sees[2].b));

    assert(a->q % b->q > 0);
    PV q = (a->q + b->q).unit();
    assert(a->q % q > 0 && q % b->q > 0);
    for (int i = 0; i < 3; i++)
      if (sees[i].a->q % q > 0 && q % sees[i].b->q > 0) {
        assert(see == 0);
        see = &sees[i];
      }
    assert(see != 0);
    
    PV c = eO->c + q.times(eR->c);
    double r = eO->r + eR->r;
    PV nt = see->t->getN(q);
    PV nh = see->h->getN(q);
    if (typeid(*t) != typeid(Intn))
      assert(t == sees[0].t || t == sees[2].t);
    else {
      Intn *intn = dynamic_cast<Intn*>(t);
      PV p = intn->getP(q);
      PV n = ((p - c) / r).unit();
      checkSame(p, c + n * r);
      if (r > 0) {
        checkPositive(nt % n);
        checkPositive(n % nh);
      }
      else {
        checkPositive(-nt % n);
        checkPositive(-n % nh);
      }
    }
    if (typeid(*h) != typeid(Intn))
      assert(h == sees[0].h || h == sees[2].h);
    else {
      Intn *intn = dynamic_cast<Intn*>(h);
      PV p = intn->getP(q);
      PV n = ((p - c) / r).unit();
      checkSame(p, c + n * r);
      if (r > 0) {
        checkPositive(nt % n);
        checkPositive(n % nh);
      }
      else {
        checkPositive(-nt % n);
        checkPositive(-n % nh);
      }
    }
  }

  void checkNeighbors (Sub *subs) {
    if (0 && at != 0 && ah != 0) {
      assert(at->bh == 0 && ah->bt == 0);
      cout << (this - subs) << " "
           << (at - subs) << " "
           << (ah - subs) << endl;
    }
    if (0 && bt != 0 && bh != 0) {
      assert(bt->ah == 0 && bh->at == 0);
      cout << (this - subs) << " "
           << (bt - subs) << " "
           << (bh - subs) << endl;
    }
  }
};  

void Event::addSub (Sub *sub) {
  if (sub->a == this)
    aSubs.push_back(sub);
  else if (sub->b == this)
    bSubs.push_back(sub);
  else
    assert(0);
}

void Event::findNeighbors () {
  for (int i = 0; i < bSubs.size(); i++) {
    Sub *bSub = bSubs[i];
    for (int j = 0; j < aSubs.size(); j++) {
      Sub *aSub = aSubs[j];
      if (!(aSub->eO == bSub->eO && aSub->eR == bSub->eR))
        continue;
      if (aSub->t == bSub->t) {
        assert(aSub->at == 0 && bSub->bt == 0);
        aSub->at = bSub;
        bSub->bt = aSub;
      }
      if (aSub->h == bSub->h) {
        assert(aSub->ah == 0 && bSub->bh == 0);
        aSub->ah = bSub;
        bSub->bh = aSub;
      }
    }
  }
}

#ifdef CONVERT
main (int argc, char *argv[]) {
  int nvOb;
  cin >> nvOb;
  Vert vOb[nvOb];
  Vert* pvOb = vOb;
  for (int i = 0; i < nvOb; i++)
    vOb[i].read(cin);

  int neOb;
  cin >> neOb;
  // assert(nvOb == neOb);
  Edge eOb[neOb];
  Edge* peOb = eOb;
  for (int i = 0; i < neOb; i++)
    eOb[i].read(cin, vOb);

  int nvRo;
  cin >> nvRo;
  Vert vRo[nvRo];
  Vert* pvRo = vRo;
  for (int i = 0; i < nvRo; i++)
    vRo[i].read(cin);

  int neRo;
  cin >> neRo;
  // assert(nvRo == neRo);
  Edge eRo[neRo];
  Edge* peRo = eRo;
  for (int i = 0; i < neRo; i++)
    eRo[i].read(cin, vRo);

  int nEvent;
  cin >> nEvent;
  Event events[nEvent];
  for (int i = 0; i < nEvent; i++)
    events[i].read(cin, i);

  int nSVE;
  cin >> nSVE;
  SVE sves[nSVE];
  SVE *psves = sves;
  for (int i = 0; i < nSVE; i++)
    sves[i].read(cin, vOb, eRo, events);

  int nSEV;
  cin >> nSEV;
  SEV sevs[nSEV];
  SEV *psevs = sevs;
  for (int i = 0; i < nSEV; i++)
    sevs[i].read(cin, eOb, vRo, events);

  int nSEE;
  cin >> nSEE;
  SEE sees[nSEE];
  for (int i = 0; i < nSEE; i++) {
    sees[i].read(cin, eOb, eRo, sves, sevs, events);
    sees[i].checkRandomQ();
  }

  int nObRo = neOb * neRo;
  int seeIndex[nObRo];
  for (int i = 0; i < nObRo; i++)
    seeIndex[i] = -1;
  for (int i = 0; i < nSEE; i++) {
    int iO = sees[i].eO - eOb;
    int iR = sees[i].eR - eRo;
    int iOR = iO + neOb * iR;
    if (seeIndex[iOR] == -1) {
      seeIndex[iOR] = i;
      assert(i % 3 == 0);
    }
  }

  int nIntn;
  cin >> nIntn;
  Intn intns[nIntn];
  Intn *pintns = intns;
  for (int i = 0; i < nIntn; i++) {
    intns[i].read(cin, eOb, eRo, events);
    intns[i].check();
  }

  int nSub;
  cin >> nSub;
  Sub subs[nSub];
  Sub *psubs = subs;
  for (int i = 0; i < nSub; i++) {
    subs[i].read(cin, eOb, eRo, sves, sevs, intns, events);
    int iO = subs[i].eO - eOb;
    int iR = subs[i].eR - eRo;
    int iOR = iO + neOb * iR;
    assert(seeIndex[iOR] != -1);
    subs[i].check(&sees[seeIndex[iOR]]);
  }

  for (int i = 0; i < nSub; i++) {
    subs[i].a->addSub(&subs[i]);
    subs[i].b->addSub(&subs[i]);
  }

  for (int i = 0; i < nEvent; i++)
    events[i].findNeighbors();

  for (int i = 0; i < nSub; i++)
    subs[i].checkNeighbors(subs);
  

  if (argc == 2) {
    double theta = atof(argv[1]);
    cerr << "theta " << theta << endl;
    PV q(theta);

    cout << "#FIG 3.2  Produced by xfig version 3.2.5b" << endl
         << "Landscape" << endl
         << "Center" << endl
         << "Inches" << endl
         << "Letter  " << endl
         << "100.00" << endl
         << "Single" << endl
         << "-2" << endl
         << "1200 2" << endl;

    for (int i = 0; i < nSEE; i++)
      sees[i].draw(cout, sees, q);
  }

}
#else
main (int argc, char *argv[]) {
  int nvOb;
  cin >> nvOb;
  Vert vOb[nvOb];
  Vert* pvOb = vOb;
  for (int i = 0; i < nvOb; i++) {
    PV p;
    cin >> p.x >> p.y;
    vOb[i].p = p;
  }

  int neOb;
  cin >> neOb;
  Edge eOb[neOb];
  Edge* peOb = eOb;
  for (int i = 0; i < neOb; i++) {
    double r;
    int t, h;
    cin >> r >> t >> h;
    Edge &edge = eOb[i];
    edge.r = r;
    edge.t = &vOb[t];
    edge.h = &vOb[h];
    edge.calcC();
  }

  int nvRo;
  cin >> nvRo;
  Vert vRo[nvRo];
  Vert* pvRo = vRo;
  for (int i = 0; i < nvRo; i++) {
    PV p;
    cin >> p.x >> p.y;
    vRo[i].p = p;
  }

  int neRo;
  cin >> neRo;
  Edge eRo[neRo];
  Edge* peRo = eRo;
  for (int i = 0; i < neRo; i++) {
    double r;
    int t, h;
    cin >> r >> t >> h;
    Edge &edge = eRo[i];
    edge.r = r;
    edge.t = &vRo[t];
    edge.h = &vRo[h];
    edge.calcC();
  }

  int nEvent = 0;
  int nSVE = 0;
  int nSEV = 0;
  int nSEE = 0;
  for (int i = 0; i < neOb; i++) {
    Edge &eO = eOb[i];
    for (int j = 0; j < neRo; j++) {
      Edge &eR = eRo[j];
      if ((eO.r < 0 || eR.r < 0) && eO.r + eR.r > 0)
        continue;
      nEvent += 4;
      nSVE += 2;
      nSEV += 2;
      nSEE += 3;
    }
  }

  Event events[nEvent];
  Event *pevents = events;
  SVE sves[nSVE];
  SVE *psves = sves;
  SEV sevs[nSEV];
  SEV *psevs = sevs;
  SEE sees[nSEE];
  SEE *psees = sees;

  int iEvent = 0;
  int iSVE = 0;
  int iSEV = 0;
  int iSEE = 0;
  for (int i = 0; i < neOb; i++) {
    Edge *eO = &eOb[i];
    for (int j = 0; j < neRo; j++) {
      Edge *eR = &eRo[j];
      if ((eO->r < 0 || eR->r < 0) && eO->r + eR->r > 0)
        continue;
      Vert *aO = eO->getAB(0);
      Vert *bO = eO->getAB(1);
      Vert *aR = eR->getAB(0);
      Vert *bR = eR->getAB(1);
      Event *ab = &events[iEvent];
      ab->index = iEvent++;
      Event *ba = &events[iEvent];
      ab->index = iEvent++;
      Event *aa = &events[iEvent];
      ab->index = iEvent++;
      Event *bb = &events[iEvent];
      ab->index = iEvent++;
      ab->q = aO->n.times(bR->n.conjugate());
      ba->q = bO->n.times(aR->n.conjugate());
      aa->q = aO->n.times(aR->n.conjugate());
      bb->q = bO->n.times(bR->n.conjugate());      
      ab->p = aO->p + bR->p.times(ab->q);
      ba->p = bO->p + aR->p.times(ba->q);
      aa->p = aO->p + bO->p.times(aa->q);
      bb->p = bO->p + bR->p.times(bb->q);

      SVE *ae = &sves[iSVE++];
      ae->v = aO;
      ae->e = eR;
      ae->a = ab;
      ae->b = aa;

      SVE *be = &sves[iSVE++];
      be->v = bO;
      be->e = eR;
      be->a = bb;
      be->b = ba;

      SEV *ea = &sevs[iSEV++];
      ea->e = eO;
      ea->v = aR;
      ea->a = aa;
      ea->b = ba;

      SEV *eb = &sevs[iSEV++];
      eb->e = eO;
      eb->v = bR;
      eb->a = ab;
      eb->b = bb;

      SEE *see0 = &sees[iSEE++];
      SEE *see1 = &sees[iSEE++];
      SEE *see2 = &sees[iSEE++];
      
      see0->eO = eO;
      see1->eO = eO;
      see2->eO = eO;

      see0->eR = eR;
      see1->eR = eR;
      see2->eR = eR;

      see0->t = ae;
      see0->h = eb;
      see2->t = ea;
      see2->h = be;

      see0->a = ab;
      see2->b = ba;

      see1->t = see0->t;
      see1->h = see0->h;

      if (aa->q % bb->q > 0) {
        see0->b = aa;
        see1->a = aa;
        see1->b = bb;
        see2->a = bb;
        see1->t = ea;
      }
      else {
        see0->b = bb;
        see1->a = bb;
        see1->b = aa;
        see2->a = aa;
        see1->h = be;
      }

      assert(see0->a->q % see0->b->q >= 0);
      assert(see1->a->q % see1->b->q >= 0);
      assert(see2->a->q % see2->b->q >= 0);
    }
  }

  assert(iEvent == nEvent);
  assert(iSVE == nSVE);
  assert(iSEV == nSEV);
  assert(iSEE == nSEE);

  cout << setprecision(23);

  cout << nvOb << endl;
  for (int i = 0; i < nvOb; i++) {
    Vert *v = &vOb[i];
    cout << v->p.x << " " << v->p.y << " " << v->n.x << " " << v->n.y << endl;
  }
  cout << endl;

  cout << neOb << endl;
  for (int i = 0; i < neOb; i++) {
    Edge *e = &eOb[i];
    cout << e->c.x << " " << e->c.y << " " << e->r << " "
         << (e->t - vOb) << " " << (e->h - vOb) << endl;
  }
  cout << endl;
  
  cout << nvRo << endl;
  for (int i = 0; i < nvRo; i++) {
    Vert *v = &vRo[i];
    cout << v->p.x << " " << v->p.y << " " << v->n.x << " " << v->n.y << endl;
  }
  cout << endl;

  cout << neRo << endl;
  for (int i = 0; i < neRo; i++) {
    Edge *e = &eRo[i];
    cout << e->c.x << " " << e->c.y << " " << e->r << " "
         << (e->t - vRo) << " " << (e->h - vRo) << endl;
  }
  cout << endl;
  
  cout << nEvent << endl;
  for (int i = 0; i < nEvent; i++) {
    Event *e = &events[i];
    cout << e->q.x << " " << e->q.y << " " << e->p.x << " " << e->p.y << endl;
  }
  cout << endl;

  cout << nSVE << endl;
  for (int i = 0; i < nSVE; i++) {
    SVE *s = &sves[i];
    cout << (s->v - vOb) << " " << (s->e - eRo) << " "
         << (s->a - events) << " " << (s->b - events) << endl;
  }
  cout << endl;

  cout << nSEV << endl;
  for (int i = 0; i < nSEV; i++) {
    SEV *s = &sevs[i];
    cout << (s->e - eOb) << " " << (s->v - vRo) << " "
         << (s->a - events) << " " << (s->b - events) << endl;
  }
  cout << endl;

  cout << nSEE << endl;
  for (int i = 0; i < nSEE; i++) {
    SEE *s = &sees[i];
    cout << (s->eO - eOb) << " " << (s->eR - eRo) << " ";
    if (typeid(*s->t) == typeid(SVE))
      cout << "0 " << (dynamic_cast<SVE*>(s->t) - sves) << " ";
    else if (typeid(*s->t) == typeid(SEV))
      cout << "1 " << (dynamic_cast<SEV*>(s->t) - sevs) << " ";
    else
      cout << "2 " << (dynamic_cast<SEE*>(s->t) - sees) << " ";
    if (typeid(*s->h) == typeid(SVE))
      cout << "0 " << (dynamic_cast<SVE*>(s->h) - sves) << " ";
    else if (typeid(*s->h) == typeid(SEV))
      cout << "1 " << (dynamic_cast<SEV*>(s->h) - sevs) << " ";
    else
      cout << "2 " << (dynamic_cast<SEE*>(s->h) - sees) << " ";
    cout << (s->a - events) << " " << (s->b - events) << endl;
  }
  cout << endl;

  cout << 0 << endl;
  cout << endl;
  cout << 0 << endl;
}
#endif
