#include "PV.h"
#include "point.h"

//typedef RParameter Parameter;
//typedef GPoint2<Parameter> GP2;

inline bool lesserZ (const double &a, const double &b) {
  if (a == 0)
    return true;
  if (b == 0)
    return false;
  if (a == b)
    return false;
  return a < b;
}

inline bool greaterZ (const double &a, const double &b) {
  if (a == 0)
    return false;
  if (b == 0)
    return true;
  if (a == b)
    return false;
  return b < a;
}

inline bool isZero (const Parameter &a) {
  return a.l == 0 && a.u.r == 0;
}

inline bool operator== (const Parameter &a, const Parameter &b) {
  return a.l == b.l && a.u.r == b.u.r;
}

inline bool lesserZ (const Parameter &a, const Parameter &b) {
  if (isZero(a))
    return true;
  if (isZero(b))
    return false;
  if (a == b)
    return false;
  return a < b;
}

inline bool greaterZ (const Parameter &a, const Parameter &b) {
  if (isZero(a))
    return false;
  if (isZero(b))
    return true;
  if (a == b)
    return false;
  return b < a;
}

inline PV P2PV (const Point2 &p) { return PV(p.x.round(), p.y.round()); }

bool intersectCircles (const Parameter &r1, Point2 c, const Parameter &r2,
                       Point2 q[2], Point2 cq[2]);

class GeoObj {
public:
  virtual void increasePrecision () = 0;
  virtual void decreasePrecision () = 0;
};

class Param : public GeoObj {
public:
  virtual Parameter getX () = 0;
};

class ParamX : public Param {
public:
  Parameter x;

  ParamX (const Parameter &x) : x(x) {
    // cout << "ParamX " << x.round() << endl;
  }
  Parameter getX () { return x; }
  void increasePrecision () {
    if (x.extended())
      return;
    x.increasePrecision();
  }
  void decreasePrecision () {
    if (!x.extended())
      return;
    x.decreasePrecision();
  }
};

class Point : public GeoObj {
public:
  virtual Point2 getP () = 0;
};

class PointXY : public Point {
public:
  Point2 p;
  
  PointXY (const Point2 &p) : p(p) {}
  Point2 getP () { return p; }
  void increasePrecision () {
    if (p.x.extended())
      return;
    p.increasePrecision();
  }
  void decreasePrecision () {
    if (!p.x.extended())
      return;
    p.decreasePrecision();
  }
};

class Circle;

class Circle : public GeoObj {
public:
  virtual Point2 getC () = 0;
  virtual Parameter getS () = 0;
};

class CirclePS : public Circle {
public:
  Point *c;
  Param *s;
  
  CirclePS (Point *c, Param *s) : c(c), s(s) {}

  Point2 getC () { return c->getP(); }
  Parameter getS () { return s->getX(); }
  
  void increasePrecision () { c->increasePrecision(); s->increasePrecision(); }
  void decreasePrecision () { c->decreasePrecision(); s->decreasePrecision(); }
};

class CircleCO : public Circle {
public:
  Circle *circle;
  Param *offset;
  Parameter s;
  
  CircleCO (Circle *circle, Param *offset)
    : circle(circle), offset(offset), s(circle->getS() + offset->getX()) {}

  Point2 getC () { return circle->getC(); }
  virtual Parameter getS () { return s; }
  
  void increasePrecision () {
    if (s.extended())
      return;
    circle->increasePrecision();
    offset->increasePrecision();
    s = circle->getS() + offset->getX();
  }
  void decreasePrecision () {
    if (!s.extended())
      return;
    circle->decreasePrecision();
    offset->decreasePrecision();
    s.decreasePrecision();
  }
};

class CircleTHS : public Circle {
public:
  Point *t;
  Point *h;
  Param *s;
  Point2 c;

  CircleTHS (Point *t, Point *h, Param *s) : t(t), h(h), s(s) { calculate(); }
  Point2 getC () { return c; }
  virtual Parameter getS () { return s->getX(); }

  void calculate ();

  void increasePrecision () {
    if (c.x.extended())
      return;
    t->increasePrecision();
    h->increasePrecision();
    s->increasePrecision();
    calculate();
  }
  void decreasePrecision () {
    if (!c.x.extended())
      return;
    t->decreasePrecision();
    h->decreasePrecision();
    s->decreasePrecision();
    c.decreasePrecision();
  }
};
  
class IntCC : public Point {
public:
  Circle *c[2];
  Point2 cp; // p - c[iSmaller], circle with smaller radius
  int iSmaller;
  
  static int construct (Circle *a, Circle *b, IntCC *out[2]);
  
  Point2 getP () { return c[iSmaller]->getC() + cp; }
  Point2 getCP (int i) {
    if (i == iSmaller)
      return cp;
    else
      return getP() - c[!iSmaller]->getC();
  }
  Point2 getN (int i) { return getCP(i) / c[i]->getS(); }

  Point2 getN (Circle *circ) {
    if (c[0] == circ)
      return getN(0);
    assert(c[1] == circ);
    return getN(1);
  }

  void increasePrecision ();
  void decreasePrecision ();
  
private:
  IntCC (const Point2 &cp, int iSmaller, Circle *a, Circle *b)
    : cp(cp), iSmaller(iSmaller) {
    c[0] = a;
    c[1] = b;
  }
};

class Normal : public Point {
public:
  virtual Point2 getN () = 0;
};

class NormalNN : public Normal {
public:
  Normal *n1, *n2;
  Point2 c, m;

  Point2 getP() { return c; }
  Point2 getN() { return m; }

  NormalNN (Normal *n1, Normal *n2) : n1(n1), n2(n2) { calculate(); }

  void increasePrecision () {
    if (c.x.extended())
      return;
    n1->increasePrecision();
    n2->increasePrecision();
    calculate();
  }
  void decreasePrecision () {
    if (!c.x.extended())
      return;
    n1->decreasePrecision();
    n2->decreasePrecision();
    c.decreasePrecision();
    m.decreasePrecision();
  }

  void calculate () {
    Point2 p1 = n1->getP();
    Point2 q1 = n1->getN();
    Point2 p2 = n2->getP();
    Point2 q2 = n2->getN();
    
    Point2 q12 = q2.divides(q1);
    Point2 qMid = (q1.innerproduct(q2).sign() > 0 ?
                   Point2(q1.x + q2.x, q1.y + q2.y) :
                   Point2(q2.y - q1.y, q1.x - q2.x)).unit();
    Point2 q12o2 = qMid.divides(q1);
    
    double two = 2;
    Point2 pMid = (p1 + p2) / two;
    Point2 pNorm = (p2 - p1).rot90() / two;
    c = pMid + pNorm * (q12o2.x / q12o2.y);
    m = (p1 - c).divides(q1);
  }
};    

class NormalXYQ : public Normal {
public:
  Point2 p;
  Parameter q;
  Point2 n;

  NormalXYQ (Parameter x, Parameter y, Parameter q) : p(x, y), q(q) {
    n = cis(q);
  }

  Point2 getP() { return p; }
  Point2 getN () { return n; }

  void increasePrecision () {
    if (n.x.extended())
      return;
    p.increasePrecision();
    q.increasePrecision();
    n = cis(q);
  }
  void decreasePrecision () {
    if (!n.x.extended())
      return;
    p.decreasePrecision();
    q.decreasePrecision();
    n.decreasePrecision();
  }

  static Point2 cis (Parameter q) {
    double theta = q.round();
    double c = cos(theta), s = sin(theta);
    Point2 cs;
    if (fabs(c) < fabs(s)) {
      cs.x = cs.x + c;
      if (q.extended())
        cs.x.increasePrecision();
      if (s > 0)
        cs.y = (1 - cs.x * cs.x).sqrt();
      else
        cs.y = -(1 - cs.x * cs.x).sqrt();
    }
    else {
      cs.y = cs.y + s;
      if (q.extended())
        cs.y.increasePrecision();
      if (c > 0)
        cs.x = (1 - cs.y * cs.y).sqrt();
      else
        cs.x = -(1 - cs.y * cs.y).sqrt();
    }
    return cs;
  }
};

class NormalCP : public Normal {
public:
  Circle *c;
  Point *p;
  Point2 n;

  NormalCP (Circle *c, Point *p)
    : c(c), p(p), n((p->getP() - c->getC()) / c->getS()) {}

  Point2 getP () { return p->getP(); }
  Point2 getN () { return n; }
  
  void increasePrecision () {
    if (n.x.extended())
      return;
    c->increasePrecision();
    p->increasePrecision();
    n = (p->getP() - c->getC()) / c->getS();
  }
  void decreasePrecision () {
    if (!n.x.extended())
      return;
    c->decreasePrecision();
    p->decreasePrecision();
    n.decreasePrecision();
  }
};

class NormalNO : public Normal {
public:
  Normal *pn;
  Param *offset;
  Point2 p;

  NormalNO (Normal *pn, Param *offset)
    : pn(pn), offset(offset),
      p(pn->getP() + pn->getN() * offset->getX()) {}

  Point2 getP () { return p; }
  Point2 getN () { return pn->getN(); }
  
  void increasePrecision () {
    if (p.x.extended())
      return;
    pn->increasePrecision();
    offset->increasePrecision();
    p = pn->getP() + pn->getN() * offset->getX();
  }
  void decreasePrecision () {
    if (!p.x.extended())
      return;
    pn->decreasePrecision();
    offset->decreasePrecision();
    p.decreasePrecision();
  }
};

class NormalIi : public Normal {
public:
  IntCC *intCC;
  int i;

  NormalIi (IntCC *intCC, int i) : intCC(intCC), i(i) {}

  Point2 getP () { return intCC->getP(); }
  Point2 getN () { return intCC->getN(i); }

  void increasePrecision () { intCC->increasePrecision(); }
  void decreasePrecision () { intCC->decreasePrecision(); }
};

class PolyArc {
public:
  // created from input file
  vector<Point *> points;
  vector<Param *> radii;

  // convolution with circle
  Param *o;

  // convolution or level set
  PolyArc *parent;

  PolyArc (const char *fileName, double angle=0);

  PolyArc (vector<PV> &points, vector<double> &radii,
           vector<int> &tails, vector<int> &heads)
    : parent(0), o(0) {
    init(points, radii, tails, heads);
  }

  // convolve poly with circle of radius o
  PolyArc (PolyArc *poly, double o);

  PolyArc (PolyArc *poly) : parent(poly), o(0) {}    

  class Vert;
  class Intn;

  class Edge : public GeoObj {
  public:
    PolyArc *poly;
    int id;
    bool concave;

    Circle *c;
    Vert *t;
    Vert *h;
    
    Edge (PolyArc *poly, Vert *t, Vert *h);

    Edge (PolyArc *poly, Vert *t, Vert *h, Edge *e);

    Edge (PolyArc *poly, Vert *t, Vert *h, bool concave);

    Edge (PolyArc *poly, Vert *tv, Vert *hv, Point *tp, Point *hp, Param *s);

    Edge (PolyArc *poly, Vert *t, Vert *h, Circle *c);

    void increasePrecision ();
    void decreasePrecision ();

    // Does intCC lies on this?
    bool contains (IntCC *intCC);

    // Compare tail or head to intersection on e
    int compare (Vert *v, IntCC *intCC);

    // Compare two intersections on e.
    int compare (Intn *a, Intn *b);

    // Is p closer to this than d?
    bool closer (Point *p, Param *d);
  };
  
  class Vert : public GeoObj {
  public:
    PolyArc *poly;
    int id;

    Normal *n;
    Edge *t;
    Edge *h;
    
    Vert (PolyArc *poly, Normal *n = 0)
      : poly(poly), id(poly->verts.size()), t(0), h(0), n(n) {
      poly->verts.push_back(this);
    }
    
    Vert (PolyArc *poly, IntCC *intCC, int i)
      : poly(poly), id(poly->verts.size()), t(0), h(0),
        n(new NormalIi(intCC, i)) {
      poly->verts.push_back(this);
    }

    void increasePrecision () {
      n->increasePrecision();
      t->increasePrecision();
      h->increasePrecision();
    }
    void decreasePrecision () {
      n->decreasePrecision();
      t->decreasePrecision();
      h->decreasePrecision();
    }
  };
  
  class Intn : public GeoObj {
  public:
    Edge *e[2];
    IntCC *intCC;
    int ids[2];

    // Put intersections into intns.  Return number of intersections.
    static int construct (Edge *a, Edge *b, Intn *intns[2]);

    ~Intn () { delete intCC; }      

    // Insert this into intersection list of e.
    void insert (Edge *e, vector<Intn *> &intns);

    void increasePrecision () {
      intCC->increasePrecision();
      e[0]->increasePrecision();
      e[1]->increasePrecision();
    }
    void decreasePrecision () {
      intCC->decreasePrecision();
      e[0]->decreasePrecision();
      e[1]->decreasePrecision();
    }

  private:
    Intn (IntCC *intCC, Edge *a, Edge *b) : intCC(intCC) {
      e[0] = a, e[1] = b;
    }
  };
  
  class SubE {
  public:
    Edge *e;
    // i in 0..m for edge e with m intersections on it
    int i;

    SubE (Edge *e, int i) : e(e), i(i) {}

    bool isNull () { return e == 0; }
    bool operator== (SubE &sub) { return e == sub.e && i == sub.i; }
    bool operator!= (SubE &sub) { return e != sub.e || i != sub.i; }
    SubE next (vector<vector<Intn *> > &intnss);
    SubE prev (vector<vector<Intn *> > &intnss);
  };

  vector<Vert *> verts;
  vector<Edge *> edges;
  
  // Edge i has radius radii[i], tail points[tails[i]],
  // and head points[heads[i]].
  void init (vector<PV> &points, vector<double> radii, vector<int> &tails, vector<int> &heads);

  // p is inside this expanded by o, ignoring edges which generated p
  bool insideD (Point *p, Param *o);

  // Create level set from convolution.
  PolyArc *levelSet ();

  void drawFig (const char *fileName);

  void write (const char *fileName);

  // hack
  void write (const char *fileName, double sOff);
};
