#ifndef ACP
#define ACP

#include <gmp.h>
#include <mpfr.h>
#include <assert.h>
#include <iostream>
#include <fenv.h>
#include <exception>
#include <math.h>
#include <algorithm>
#include <sys/time.h>
#include <float.h>

using namespace std;

double randomNumber (double rmin, double rmax);

double getTime ();

class SignException: public exception {
 public:
  virtual const char* what() const throw() {
    return "Not enough precision";
  }
};

extern SignException signException;

class MValue {
public:
  MValue (double rl, double ru);
  void incRef () { refCnt++; }
  void decRef ();

  unsigned int refCnt;
  mpfr_t l, u;
};

#define Sentinel 1e20

class Parameter {
 public:
  static double delta;
  static unsigned int precision;
  static void enable () { fesetround(FE_UPWARD); }
  static void disable () { fesetround(FE_TONEAREST); }

  static Parameter input (double a) {
    double p = a + delta*(1.0 + fabs(a))*randomNumber(-1.0, 1.0);
    return Parameter(p, p);
  }

  Parameter () : l(0.0) { u.r = 0.0; }
  Parameter (double il, double iu) : l(il) { u.r = iu; }    

  Parameter (const Parameter &p) : l(p.l)
    {
      if (l != Sentinel) 
	u.r = p.u.r;
      else {
	u.m = p.u.m;
	u.m->incRef();
      }
    }

  Parameter (const Parameter &il, const Parameter &iu) {
    if (il.l != Sentinel) {
      l = il.l;
      u = iu.u;
    }
    else {
      l = Sentinel;
      u.m = new MValue(0.0, 0.0);
      mpfr_set(u.m->l, il.u.m->l, GMP_RNDD);
      mpfr_set(u.m->u, iu.u.m->u, GMP_RNDU);
    }
  }
  
  ~Parameter () { if (l == Sentinel) u.m->decRef(); }
  
  Parameter & operator= (const Parameter &p) {
    if (p.l == Sentinel)
      p.u.m->incRef();
    if (l == Sentinel)
      u.m->decRef();
    l = p.l;
    u = p.u;
    return *this;
  }

  bool extended () const { return l == Sentinel; }

  double mu () const { 
    if (l != Sentinel)
      return 1.1e-16;
    Parameter p;
    p.increasePrecision();
    mpfr_set_d(p.u.m->l, 1.0, GMP_RNDN);
    mpfr_div_2ui(p.u.m->l, p.u.m->l, precision, GMP_RNDN);
    return mpfr_get_d(p.u.m->l, GMP_RNDN);
  }

  void increasePrecision () {
    if (l == Sentinel)
      return;
    u.m = new MValue(l, u.r);
    l = Sentinel;
  }

  void decreasePrecision () {
    if (l == Sentinel) {
      double dl = lb(), du = ub();
      u.m->decRef();
      l = dl;
      u.r = du;
    }
  }

  double intervalWidth () const {
    if (l != Sentinel)
      return u.r - l;
    mpfr_t w;
    mpfr_init2(w, precision);
    mpfr_sub(w, u.m->u, u.m->l, GMP_RNDN);
    double wr = mpfr_get_d(w, GMP_RNDN);
    mpfr_clear(w);
    return wr;
  }

  double lb () const { 
    if (l != Sentinel)
      return l;
    return mpfr_get_d(u.m->l, GMP_RNDD);
  }

  double ub () const { 
    if (l != Sentinel)
      return u.r;
    return mpfr_get_d(u.m->u, GMP_RNDU);
  }

  double round () const { return 0.5*(lb() + ub()); }

  int sign (bool fail = true) const {
    extern long int nn3, nn4;
    if (l != Sentinel) {
      if (l > 0.0) {
	++nn3;
	return 1;
      }
      if (u.r < 0.0) {
	++nn3;
	return -1;
      }
      if (fail)
	throw signException;
      ++nn3;
      return 0;
    }
    ++nn4;
    if (mpfr_sgn(u.m->l) > 0)
      return 1;
    if (mpfr_sgn(u.m->u) < 0)
      return -1;
    if (fail)
      assert(0);
      // throw signException;
    return 0;
  }

  bool operator< (const Parameter &b) const { return (b - *this).sign() == 1; }
  bool operator< (double b) const;
  bool operator> (const Parameter &b) const { return b < *this; }
  bool operator> (double b) const;
  Parameter perturb () const { return input(round()); }

  Parameter operator+ (const Parameter &b) const {
    if (l != Sentinel && b.l != Sentinel)
      return Parameter(- (- l - b.l), u.r + b.u.r);
    assert(l == Sentinel && b.l == Sentinel);
    Parameter p;
    p.increasePrecision();
    mpfr_add(p.u.m->l, u.m->l, b.u.m->l, GMP_RNDD);
    mpfr_add(p.u.m->u, u.m->u, b.u.m->u, GMP_RNDU);
    return p;
  }

  Parameter operator+ (double b) const {
    if (l != Sentinel)
      return Parameter(- (- l - b), u.r + b);
    Parameter p;
    p.increasePrecision();
    mpfr_add_d(p.u.m->l, u.m->l, b, GMP_RNDD);
    mpfr_add_d(p.u.m->u, u.m->u, b, GMP_RNDU);
    return p;
  }
  
  Parameter operator- (const Parameter &b) const {
    if (l != Sentinel && b.l != Sentinel)
      return Parameter(-(b.u.r - l), u.r - b.l);
    assert(l == Sentinel && b.l == Sentinel);
    Parameter p;
    p.increasePrecision();
    mpfr_sub(p.u.m->l, u.m->l, b.u.m->u, GMP_RNDD);
    mpfr_sub(p.u.m->u, u.m->u, b.u.m->l, GMP_RNDU);
    return p;
  }
  
  Parameter operator- (double b) const {
    if (l != Sentinel)
      return Parameter(-(b - l), u.r - b);
    Parameter p;
    p.increasePrecision();
    mpfr_sub_d(p.u.m->l, u.m->l, b, GMP_RNDD);
    mpfr_sub_d(p.u.m->u, u.m->u, b, GMP_RNDU);
    return p;
  }

  Parameter operator- () const {
    if (l != Sentinel)
      return Parameter(- u.r, - l);
    Parameter p;
    p.increasePrecision();
    mpfr_neg(p.u.m->l, u.m->u, GMP_RNDD);
    mpfr_neg(p.u.m->u, u.m->l, GMP_RNDU);
    return p;
  }

  Parameter operator* (const Parameter &b) const {
    if (l != Sentinel && b.l != Sentinel) {
      Parameter s = u.r < 0.0 ? - *this : *this, t = u.r < 0.0 ? - b : b;
      if (s.l > 0.0) {
	double k = t.l > 0.0 ? - s.l : - s.u.r;
	return Parameter(- (k*t.l), t.u.r > 0.0 ? s.u.r*t.u.r : s.l*t.u.r);
      }
      if (t.l > 0.0) {
	double k = - s.l;
	return Parameter(- (k*t.u.r), s.u.r*t.u.r);
      }
      if (t.u.r < 0.0) {
	double k = - s.u.r;
	return Parameter(- (k*t.l), s.l*t.l);
      }
      double k1 = - s.l, k2 = - s.u.r, cl1 = k1*t.u.r, cl2 = k2*t.l, 
	cu1 = s.l*t.l, cu2 = s.u.r*t.u.r,
	cl = cl1 < cl2 ? - cl2 : - cl1, cu = cu1 < cu2 ? cu2 : cu1;
      return Parameter(cl, cu);
    }
    assert(l == Sentinel && b.l == Sentinel);
    bool pflag = ub() < 0.0;
    Parameter p, s =  pflag ? - *this : *this, t = pflag ? - b : b;
    p.increasePrecision();
    if (mpfr_sgn(s.u.m->l) > 0.0) {
      mpfr_mul(p.u.m->l, mpfr_sgn(t.u.m->l) > 0 ? s.u.m->l : s.u.m->u, 
	       t.u.m->l, GMP_RNDD);
      mpfr_mul(p.u.m->u, mpfr_sgn(t.u.m->u) > 0 ? s.u.m->u : s.u.m->l, 
	       t.u.m->u, GMP_RNDU);
    }
    else if (mpfr_sgn(t.u.m->l) > 0) {
      mpfr_mul(p.u.m->l, s.u.m->l, t.u.m->u, GMP_RNDD);
      mpfr_mul(p.u.m->u, s.u.m->u, t.u.m->u, GMP_RNDU);
    }
    else if (mpfr_sgn(t.u.m->u) < 0) {
      mpfr_mul(p.u.m->l, s.u.m->u, t.u.m->l, GMP_RNDD);
      mpfr_mul(p.u.m->u, s.u.m->l, t.u.m->l, GMP_RNDU);
    }
    else {
      mpfr_t cl1, cl2, cu1, cu2;
      mpfr_init2(cl1, precision);
      mpfr_init2(cl2, precision);
      mpfr_init2(cu1, precision);
      mpfr_init2(cu2, precision);
      mpfr_mul(cl1, s.u.m->l, t.u.m->u, GMP_RNDD);
      mpfr_mul(cl2, s.u.m->u, t.u.m->l, GMP_RNDD);
      mpfr_mul(cu1, s.u.m->l, t.u.m->l, GMP_RNDU);
      mpfr_mul(cu2, s.u.m->u, t.u.m->u, GMP_RNDU);
      mpfr_set(p.u.m->l, mpfr_less_p(cl1, cl2) ? cl1 : cl2, GMP_RNDD);
      mpfr_set(p.u.m->u, mpfr_less_p(cu1, cu2) ? cu2 : cu1, GMP_RNDU);
      mpfr_clear(cl1);
      mpfr_clear(cl2);
      mpfr_clear(cu1);
      mpfr_clear(cu2);
    }
    return p;
  }
  
  Parameter operator* (double b) const {
    if (l != Sentinel) {
      double k = - b;
      return b > 0.0 ? Parameter(- (k*l), b*u.r) : Parameter(- (k*u.r), b*l);
    }
    Parameter p;
    p.increasePrecision();
    if (b > 0.0) {
      mpfr_mul_d(p.u.m->l, u.m->l, b, GMP_RNDD);
      mpfr_mul_d(p.u.m->u, u.m->u, b, GMP_RNDU);
    }
    else {
      mpfr_mul_d(p.u.m->l, u.m->u, b, GMP_RNDD);
      mpfr_mul_d(p.u.m->u, u.m->l, b, GMP_RNDU);
    }
    return p;
  }

  Parameter operator/ (const Parameter &b) const { return *this*b.reciprocal(); }

  Parameter operator/ (double b) const { return *this*(1.0/b); }
  
  Parameter reciprocal () const {
    sign();
    if (l != Sentinel) {
      double k = - u.r;
      return Parameter(-1.0/k, 1.0/l);
    }
    Parameter p;
    p.increasePrecision();
    mpfr_d_div(p.u.m->l, 1.0, u.m->u, GMP_RNDD);
    mpfr_d_div(p.u.m->u, 1.0, u.m->l, GMP_RNDU);
    return p;
  }

  Parameter sqrt () const {
    assert(sign() == 1);
    if (l != Sentinel) {
      double sl = ::sqrt(l);
      while (- ((- sl)*sl) > l)
	sl = nextafter(sl, - DBL_MAX);
      double su = ::sqrt(u.r);
      while (su*su < u.r)
	su = nextafter(su, DBL_MAX);
      return Parameter(sl, su);
    }
    Parameter p;
    p.increasePrecision();
    mpfr_sqrt(p.u.m->l, u.m->l, GMP_RNDD);
    mpfr_sqrt(p.u.m->u, u.m->u, GMP_RNDU);
    return p;
  }
  
  Parameter abs () const { 
    return sign() == 1 ? *this : - *this; 
  }

  Parameter mid () const {
    if (l != Sentinel) {
      double v = round();
      return Parameter(v, v);
    }
    Parameter p;
    p.increasePrecision();
    mpfr_add(p.u.m->l, u.m->l, u.m->u, GMP_RNDN);
    mpfr_div_2ui(p.u.m->l, p.u.m->l, 1, GMP_RNDN);
    mpfr_set(p.u.m->u, p.u.m->l, GMP_RNDN);
    return p;
  }

  bool subset (const Parameter &b) const {
    if (l != Sentinel)
      return l > b.lb() && u.r < b.ub();
    if (b.l != Sentinel)
      return mpfr_cmp_d(u.m->l, b.l) > 0 && mpfr_cmp_d(u.m->u, b.u.r) < 0;
    return mpfr_cmp(u.m->l, b.u.m->l) > 0 && mpfr_cmp(u.m->u, b.u.m->u) < 0;
  }

  double l;
  union {
    double r;
    MValue *m;
  } u;
};

Parameter operator+ (double a, const Parameter &b);

Parameter operator- (double a, const Parameter &b);

Parameter operator* (double a, const Parameter &b);

bool operator< (double a, const Parameter &b);

bool operator> (double a, const Parameter &b);

#endif
