#include "eigen.h"

int realRoots (const PFunPoly &f, double *res)
{
  int n = f.extended() ? realRootsE(f, res) : realRootsR(f, res);
  if (n == -1)
    return -1;
  sort(res, res + n);
  return n;
}

int realRootsR (const PFunPoly &f, double *res)
{
  int d = f.d;
  double *a = new double [d*d];
  companionMatrixR(f, a);
  int nr = eigenvaluesR(a, d, res);
  delete [] a;
  return nr;
}

void companionMatrixR (const PFunPoly &f, double *a)
{
  int d = f.d;
  double *ptr = a, s = -1.0/f.a[d].value();
  for (int i = 0; i + 1 < d; ++i)
    for (int j = 0; j < d; ++j, ++ptr)
      *ptr = i + 1 == j ? 1.0 : 0.0;
  for (int i = 0; i < d; ++i, ++ptr)
    *ptr = s*f.a[i].value();
}

int eigenvaluesR (double *a, int k, double *res)
{
  double evalsr[k], evalsi[k], evecs[k*k], maxcond = 1e6;
  if (!eigen(k, a, evalsr, evalsi, evecs, maxcond))
    return -1;
  int nr = 0;
  for (int i = 0; i < k; ++i)
    if (evalsi[i] == 0.0)
      res[nr++] = evalsr[i];
  return nr;
}

Monomial::Monomial (int in) : n(in)
{
  elts = n ? new int [n] : 0;
  for (int i = 0; i < n; ++i)
    elts[i] = 0;
}

Monomial::Monomial (const Monomial &b) : n(b.n)
{
  elts = n ? new int [n] : 0;
  for (int i = 0; i < n; ++i)
    elts[i] = b.elts[i];
}

Monomial & Monomial::operator= (const Monomial &b)
{
  if (n < b.n) {
    delete [] elts; 
    elts = new int [b.n]; 
  }
  n = b.n;
  for (int i = 0; i < n; ++i)
    elts[i] = b.elts[i]; 
  return *this;
}

int Monomial::degree () const
{
  int res = 0;
  for (int i = 0; i < n; ++i)
    res += elts[i];
  return res;
}

bool Monomial::operator< (const Monomial &b) const
{
  int sa = degree(), sb = b.degree();
  if (sa > sb)
    return true;
  if (sa < sb)
    return false;
  for (int i = 0; i < n; ++i)
    if (elts[i] > b.elts[i])
      return true;
    else if (elts[i] < b.elts[i])
      return false;
  return false;
}
 
bool Monomial::inI0 (PPoly *f) const
{
  for (int i = 0; i < n; ++i)
    if (elts[i] >= f[i].d)
      return false;
  return true;
}

bool MonomialOrder::operator() (const Monomial &a, const Monomial &b) const
{
  bool fa = a.inI0(f), fb = b.inI0(f);
  if (!fa && fb)
    return true;
  if (fa && !fb)
    return false;
  return a < b;
}

int systemDegree (int n, PPoly *f)
{
  int m = 1;
  for (int i = 0; i < n; ++i)
    m *= f[i].d;
  return m;
}

int maxMonomialDegree (int n, PPoly *f)
{
  int dmax = 1 - n;
  for (int i = 0; i < n; ++i)
    dmax += f[i].d;
  return dmax;
}

int formMonomials (int n, PPoly *f, int dmax, Monomial *&z)
{  
  int N = choose(n, n + dmax);
  z = new Monomial [N];
  z[0] = Monomial(n);
  int nprev = 1, npp = 1;
  for (int i = 0; i < n; ++i) {
    for (int j = 0; j < nprev; ++j) {
      int dd = dmax - z[j].degree();
      for (int k = 1; k <= dd; ++k) {
	z[npp] = z[j];
	z[npp].elts[i] = k;
	++npp;
      }
    }
    nprev = npp;
  }
  sort(z, z + N, MonomialOrder(n, f));
  return N;
}

int choose (int k, int n)
{
  int res = 1;
  for (int i = 0; i < k; ++i)
    res *= n - i;
  for (int i = 2; i <= k; ++i)
    res /= i;
  return res;
}

int Zindex (const Monomial &m, int N, Monomial *z, const MonomialOrder &mo)
{
  return lower_bound(z, z + N, m, mo) - z;
}

int * varIndex (int n, PPoly *f, int N, Monomial *z)
{
  int *ind = new int [n];
  Monomial s(n);
  MonomialOrder mo(n, f);
  for (int i = 0; i < n; ++i) {
    s.elts[i] = 1;
    ind[i] = Zindex(s, N, z, mo);
    s.elts[i] = 0;
  }
  return ind;
}

int realRoots (int n, PPoly *f, double *res)
{
  return f->extended() ? realRootsE(n, f, res) 
    : realRootsR(n, f, res);
}

int realRootsR (int n, PPoly *f, double *res)
{
  Monomial *z;
  int m = systemDegree(n, f), dmax = maxMonomialDegree(n, f), 
    N = formMonomials(n, f, dmax, z), nr = N - m, k = -1;
  double *a11 = setupA1R(n, f, m, N, z), *a10 = a11+nr*nr,
    *a01 = setupA0R(n, f, m, N, z), *a00 = a01 + nr*m, maxcond = 1e6;
  if (linsolve(a11, a10, nr, m, maxcond)) {
    cblas_dgemm(CblasColMajor, CblasNoTrans, CblasNoTrans, m, m, nr, -1.0,
		a01, m, a10, nr, 1.0, a00, m);
    double evalsr[m], evalsi[m], evecs[m*m];
    if (eigen(m, a00, evalsr, evalsi, evecs, maxcond)) {
      int *ind = varIndex(n, f, N, z);
      k = eigenRootsR(n, f, m, evalsr, evalsi, evecs, nr, a10, ind, res);
      delete [] ind;
    }
  }
  delete [] a11;
  delete [] a01;
  delete [] z;
  return k;
}

double * setupA1R (int n, PPoly *f, int m, int N, Monomial *z)
{
  int nr = N - m;
  double *a = new double [nr*N];
  MonomialOrder mo(n, f);
  for (int i = 0; i < nr*N; ++i)
    a[i] = 0.0;
  Monomial s(n);
  for (int i = 0; i < nr; ++i) {
    Monomial zi = z[i];
    for (int j = n - 1; j >= 0; --j) {
      PPoly fj = f[j];
      if (zi.elts[j] >= fj.d) {
	zi.elts[j] -= fj.d;
	for (int k = 0; k < fj.nt; ++k) {
	  int *tk = fj.m + k*n;
	  for (int l = 0; l < n; ++l)
	    s.elts[l] = zi.elts[l] + tk[l];
	  int ns = Zindex(s, N, z, mo);
	  a[ns*nr+i] = fj.a[k].value();
	}
	break;
      }
    }
  }
  return a;
}

double * setupA0R (int n, PPoly *f, int m, int N, Monomial *z)
{
  double *a = new double [m*N];
  for (int i = 0; i < m*N; ++i)
    a[i] = 0.0;
  MonomialOrder mo(n, f);
  for (int i = 0; i < m; ++i) {
    Monomial s = z[N-m+i];
    s.elts[0] += 1;
    int ns = Zindex(s, N, z, mo);
    a[ns*m+i] = 1.0;
  }
  return a;
}

int eigenRootsR (int n, PPoly *f, int m, double *evalsr, double *evalsi, 
		 double *evecs, int nr, double *a10, int *ind, double *res)
{
  int k = 0;
  for (int i = 0; i < m; ++i)
    if (evalsi[i] == 0.0) {
      eigenRootR(n, m, evalsr[i], evecs + m*i, nr, a10, ind, res + k*n);
      ++k;
    }
  return k;
}

void eigenRootR (int n, int m, double eval, double *evec, int nr, 
		 double *a10, int *ind, double *res)
{
  res[0] = eval;
  if (n > 1) {
    double d = 1.0/evec[m-1];
    for (int j = 1; j < n; ++j) {
      int ij = ind[j];
      if (ij < nr) {
	res[j] = 0.0;
	for (int l = 0; l < m; ++l)
	  res[j] -= a10[ij+nr*l]*evec[l]*d;
      }
      else
	res[j] = evec[ij-nr]*d;
    }
  }
}

// LAPACK

bool linsolve (double *a, double *b, int m, int k, double maxcond)
{
  int n = m, nrhs = k, lda = m, ipiv[m], ldb = m, info, iwork[n];
  char norm = 'I', trans = 'N';
  double anorm = maxnorm(a, m, m), rcond, work[4*n];
  dgetrf_(&n, &n, a, &lda, ipiv, &info);
  if (info != 0)
    return false;
  dgecon_(&norm, &n, a, &lda, &anorm, &rcond, work, iwork, &info);
  if (info != 0 || rcond*maxcond < 1.0)
    return false;
  dgetrs_(&trans, &n, &nrhs, a, &lda, ipiv, b, &ldb, &info);
  return info == 0;
}

double maxnorm (double *a, int m, int n)
{
  double res = 0.0;
  for (int i = 0; i < m; ++i) {
    double ri = 0.0;
    for (int j = 0; j < n; ++j)
      ri += fabs(a[i+m*j]);
    if (ri > res)
      res = ri;
  }
  return res;
}

bool eigen (int k, double *a, double *evalsr, double *evalsi, double *evecs, 
	    double maxcond)
{
  char balanc = 'B', jobvl = 'V', jobvr = 'V', sense = 'E';
  int n = k, lda = k, ldvl = k, ldvr = k, ilo, ihi,
    lwork = 10*k*(k+6), iwork[2*k-2], info;
  double *wr = evalsr, *wi = evalsi, vl[k*k], *vr = evecs, scale[k], abnrm = 0.0,
    rcond[k], work[lwork];
  Parameter::disable();
  dgeevx_(&balanc, &jobvl, &jobvr, &sense, &n, a, &lda, wr, wi,
	  vl, &ldvl, vr, &ldvr, &ilo, &ihi, scale, &abnrm,
	  rcond, 0, work, &lwork, iwork, &info);
  Parameter::enable();
  if (info != 0)
    return false;
  for (int i = 0; i < k; ++i)
    if (maxcond*rcond[i] < 1.0)
      return false;
  return true;
}

// end LAPACK

int realRootsE (const PFunPoly &f, double *res)
{
  int d = f.d;
  matrixE a;
  a.setlength(d, d);
  companionMatrixE(f, a);
  int nr = eigenvaluesE(a, d, res);
  return nr;
}

void companionMatrixE (const PFunPoly &f, matrixE &a)
{
  int d = f.d;
  for (int i = 0; i < d; ++i)
    for (int j = 0; j + 1 < d; ++j)
      a(i, j) = i == j + 1 ? 1.0 : 0.0;
  for (int i = 0; i < d; ++i) {
    Parameter p = (- f.a[i]/f.a[d]).mid();
    mpfr_set(a(i, d - 1).getWritePtr(), p.u.m->l, GMP_RNDN);
  }
}

int eigenvaluesE (matrixE &a, int k, double *res)
{
  vectorE evalsr, evalsi;
  evalsr.setlength(k);
  evalsi.setlength(k);
  matrixE evecsl, evecsr;
  if (!evd::rmatrixevd<PRECISION>(a, k, 0, evalsr, evalsi, evecsl, evecsr))
    return -1;
  int nr = 0;
  for (int i = 0; i < k; ++i)
    if (evalsi(i).isZero())
      res[nr++] = evalsr(i).toDouble();
  return nr;
}

int realRootsE (int n, PPoly *f, double *res)
{
  Monomial *z;
  int m = systemDegree(n, f), dmax = maxMonomialDegree(n, f), 
    N = formMonomials(n, f, dmax, z), nr = N - m, k = -1;
  matrixE a11, a10, a01, a00, x;
  setupA1E(n, f, m, N, z, a11, a10);
  setupA0E(n, f, m, N, z, a01, a00);
  if (linsolveE(a11, nr, a10, m, x)) {
    ablas::rmatrixgemm<PRECISION>(m, m, nr, -1.0, a01, 0, 0, 0, x, 0, 0, 0, 1.0, 
				  a00, 0, 0);
    vectorE evalsr, evalsi;
    evalsr.setlength(m);
    evalsi.setlength(m);
    matrixE evecsl, evecsr;
    evecsr.setlength(m, m);
    if (evd::rmatrixevd<PRECISION>(a00, m, 1, evalsr, evalsi, evecsl, evecsr)) {
      int *ind = varIndex(n, f, N, z);
      k = eigenRootsE(n, f, m, evalsr, evalsi, evecsr, nr, x, ind, res);
      delete [] ind;
    }
  }
  delete [] z;
  return k;
}

void setupA1E (int n, PPoly *f, int m, int N, Monomial *z, matrixE &a11, 
	       matrixE &a10)
{
  int nr = N - m;
  a11.setlength(nr, nr);
  a10.setlength(nr, m);
  MonomialOrder mo(n, f);
  for (int i = 0; i < nr; ++i)
    for (int j = 0; j < nr; ++j)
      a11(i, j) = 0.0;
  for (int i = 0; i < m; ++i)
    for (int j = 0; j < nr; ++j)
      a10(i, j) = 0.0;
  Monomial s(n);
  for (int i = 0; i < nr; ++i) {
    Monomial zi = z[i];
    for (int j = n - 1; j >= 0; --j) {
      PPoly fj = f[j];
      if (zi.elts[j] >= fj.d) {
	zi.elts[j] -= fj.d;
	for (int k = 0; k < fj.nt; ++k) {
	  int *tk = fj.m + k*n;
	  for (int l = 0; l < n; ++l)
	    s.elts[l] = zi.elts[l] + tk[l];
	  int ns = Zindex(s, N, z, mo);
	  Parameter p = fj.a[k].mid();
	  if (ns < nr)
	    mpfr_set(a11(i, ns).getWritePtr(), p.u.m->l, GMP_RNDN);
	  else
	    mpfr_set(a10(i, ns - nr).getWritePtr(), p.u.m->l, GMP_RNDN);
	}
	break;
      }
    }
  }
}

void setupA0E (int n, PPoly *f, int m, int N, Monomial *z, matrixE &a01, 
	       matrixE &a00)
{
  int nr = N - m;
  a01.setlength(m, nr);
  a00.setlength(m, m);
  for (int i = 0; i < m; ++i)
    for (int j = 0; j < nr; ++j)
      a01(i, j) = 0.0;
  for (int i = 0; i < m; ++i)
    for (int j = 0; j < m; ++j)
      a00(i, j) = 0.0;
  MonomialOrder mo(n, f);
  for (int i = 0; i < m; ++i) {
    Monomial s = z[N-m+i];
    s.elts[0] += 1;
    int ns = Zindex(s, N, z, mo);
    if (ns < nr)
      a01(i, ns) = 1.0;
    else
      a00(i, ns - nr) = 1.0;
  }
}

bool linsolveE (const matrixE &a, int n, const matrixE &b, int m, matrixE &x)
{
  densesolver::densesolverreport<PRECISION> rep;
  int info;
  densesolver::rmatrixsolvem(a, n, b, m, true, info, rep, x);
  return info == 1;
}

int eigenRootsE (int n, PPoly *f, int m, const vectorE &evalsr, 
		 const vectorE &evalsi, const matrixE &evecs, int nr, 
		 const matrixE &a10, int *ind, double *res)
{
  int k = 0;
  for (int i = 0; i < m; ++i)
    if (evalsi(i).isZero()) {
      res[0] = evalsr(i).toDouble();
      if (n > 1) {
	for (int j = 1; j < n; ++j) {
	  int ij = ind[j];
	  doubleE rj = 0.0;
	  if (ij < nr)
	    for (int l = 0; l < m; ++l)
	      rj -= a10(ij, l)*evecs(l, i);
	  else
	    rj = evecs(ij - nr, i);
	  rj /= evecs(m - 1, i);
	  res[j] = rj.toDouble();
	}
      }
      res += n;
      ++k;
    }
  return k;
}
