#include "implicit.h"

Parameter PFun::polish (const Parameter &x) const
{
  double xub = x.ub();
  // vjm 20120807
  // int sub = value(Parameter(xub,xub)).sign();
  Parameter xubP(xub,xub);
  if (extended())
    xubP.increasePrecision();
  int sub = value(xubP).sign();
  return shrinkBracket(this, sub, x.lb(), xub);
}

PFunPoly::PFunPoly (int di, Parameter *ai) : d (di)
{
  a = new Parameter [d + 1];
  if (ai)
    for (int i = 0; i <= d; ++i)
      a[i] = ai[i];
}

PFunPoly::PFunPoly (const PFunPoly &f) : d (f.d)
{
  a = new Parameter [d + 1];
  for (int i = 0; i <= d; ++i)
    a[i] = f.a[i];
}

PFunPoly & PFunPoly::operator= (const PFunPoly &f)
{
  delete [] a;
  d = f.d;
  a = new Parameter [d+1];
  for (int i = 0; i <= d; ++i)
    a[i] = f.a[i];
  return *this;
}

Parameter PFunPoly::value (const Parameter &x) const
{
  Parameter y = a[d];
  for (int i = d - 1; i >= 0; --i)
    y = x*y + a[i];
  return y;
}

Parameter PFunPoly::der (const Parameter &x) const
{
  Parameter y = d*a[d];
  for (int i = d - 1; i > 0; --i)
    y = x*y + i*a[i];
  return y;
}

PFunPoly PFunPoly::partial () const
{
  PFunPoly g(d - 1);
  for (int i = 1; i <= d; ++i)
    g.a[i-1] = i*a[i];
  return g;
}

void PFunPoly::increasePrecision ()
{
  for (int i = 0; i <= d; ++i)
    a[i].increasePrecision();
}

void PFunPoly::decreasePrecision ()
{
  for (int i = 0; i <= d; ++i)
    a[i].decreasePrecision();
}

int implicit (const PFunPoly &f, double lb, double ub, Parameter *res)
{
  if (f.d < 3)
    return implicit2(f, lb, ub, res);
  PFunPoly df = f.partial();
  Parameter *dres = new Parameter [df.d];
  int dn = implicit(df, lb, ub, dres), n = 0;
  Parameter lbi(lb, lb);
  if (f.extended())
    lbi.increasePrecision();
  int slbi = f.value(lbi).sign();
  for (int i = 0; i <= dn; ++i) {
    Parameter ubi = i < dn ? dres[i] : Parameter(ub, ub);
    if (i == dn && f.extended())
      ubi.increasePrecision();
    int subi = f.value(ubi).sign();
    if (slbi == - subi)
      res[n++] = shrinkBracket(&f, subi, lbi.lb(), ubi.ub());
    lbi = ubi;
    slbi = subi;
  }
  delete [] dres;
  return n;
}

int implicit2 (const PFunPoly &f, double lb, double ub, Parameter *res)
{
  if (f.d == 0)
    return 0;
  if (f.d == 1) {
    res[0] = - f.a[0]/f.a[1];
    return lb < res[0] && res[0] < ub ? 1 : 0;
  }
  const Parameter &a = f.a[2], &b = f.a[1], &c = f.a[0];
  Parameter disc = b*b - 4.0*a*c;    
  if (disc.sign() == -1)
    return 0;
  Parameter d = disc.sqrt(), x[2];

  if (b.sign() == -1) {
    x[0] = 0.5*(d - b)/a;
    x[1] = 2.0*c/(d - b);
  }
  else {
    x[0] = -0.5*(d + b)/a;
    x[1] = -2.0*c/(d + b);
  }
  int n = 0;
  for (int i = 0; i < 2; ++i)
    if (lb < x[i] && x[i] < ub)
      res[n++] = x[i];
  sort(res, res + n);
 return n;
}

Parameter implicit (double a, const PFun *f, double lb, double ub)
{
  Parameter aa(a, a);
  if (f->extended())
    aa.increasePrecision();
  int sub = f->der(aa).sign();
  double xlb = findBracket(a, f, - sub, lb), xub = findBracket(a, f, sub, ub);
  return shrinkBracket(f, sub, xlb, xub);
}

double findBracket (double a, const PFun *f, int sb, double b)
{
  Parameter aa(a, a);
  if (f->extended())
    aa.increasePrecision();
  double ga = f->value(aa).round(), dga = f->der(aa).round(), k = sb*1.1e-15;
  while (true) {
    double x = a + (k - ga)/dga;
    if (b > a ? x >= b : x <= b)
      return b;
    Parameter xx(x, x);
    if (f->extended())
      xx.increasePrecision();
    if (f->value(xx).sign(false) == sb)
      return x;
    k *= 2.0;
  }
}

Parameter shrinkBracket (const PFun *f, int sub, double lb, double ub)
{
  bool bflag = false;
  Parameter x(lb, ub);
  if (f->extended())
    x.increasePrecision();
  double mu = x.mu();
  while (x.intervalWidth() > mu) {
    bool flag = true;
    Parameter xm = x.mid(), fm = f->value(xm), df = f->der(x);
    if (df.sign(false) == 0)
      flag = false;
    else {
      Parameter nx = xm - fm/df;
      if (nx.subset(x))
	x = nx;
      else
	flag = false;
    }
    if (!flag) {
      int sm = fm.sign(false);
      if (sm == sub) {
	Parameter nx(x, xm);
	if (nx.intervalWidth() == x.intervalWidth())
	  break;
	x = nx;
      }
      else if (sm == - sub) {
	Parameter nx(xm, x);
	if (nx.intervalWidth() == x.intervalWidth())
	  break;
	x = nx;
      }
      else if (!bflag) {
	bflag = true;
	x = shrinkBracketB(f, x, sub);
      }
      else
	break;
    }
  }
  return x;
}

Parameter shrinkBracketB (const PFun *f, const Parameter &x, int sub)
{
  double xlb = x.lb(), xm = x.round(), xub = x.ub();
  while (xlb < xm) {
    double nx = 0.5*(xlb + xm);
    if (nx == xlb || nx == xm)
      break;
    int sm = f->value(Parameter(nx, nx)).sign(false);
    if (sm == 0)
      xm = nx;
    else
      xlb = nx;
  }
  xm = x.round();
  while (xm < xub) {
    double nx = 0.5*(xm + xub);
    if (nx == xm || nx == xub)
      break;
    int sm = f->value(Parameter(nx, nx)).sign(false);
    if (sm == 0)
      xm = nx;
    else
      xub = nx;
  }
  return xub - xlb < x.intervalWidth() ? Parameter(xlb, xub) : x;
}
