#ifndef POINT
#define POINT

#include <vector>
#include <iomanip>
#include <fstream>
#include "acp.h"

class Point2 {
 public:
  Point2 () {}
  Point2 (double ix, double iy) : x(Parameter(ix, ix)), y(Parameter(iy, iy)) {}
  Point2 (const Parameter &ix, const Parameter &iy) : x(ix), y(iy) {}
  Point2 (const Point2 &p) : x(p.x), y(p.y) {}
  Point2 & operator= (const Point2 &p) { x = p.x; y = p.y; return *this; }
  Point2 round () const { return Point2(x.round(), y.round()); }
  Point2 perturb () const { return Point2(x.perturb(), y.perturb()); }
  void increasePrecision () { x.increasePrecision(); y.increasePrecision(); }
  void decreasePrecision () { x.decreasePrecision(); y.decreasePrecision(); }
  Parameter getX () const { return x; }
  Parameter getY () const { return y; }
  Parameter innerproduct (const Point2 &b) const { return x*b.x + y*b.y; }
  Parameter crossproduct (const Point2 &b) const { return x*b.y - y*b.x; }
  Parameter normSquared () const { return x*x + y*y; }
  Parameter norm () const { return normSquared().sqrt(); }
  Parameter distance (const Point2 &b) const { return (b - *this).norm(); }
  Point2 operator+ (const Point2 &b) const { return Point2(x + b.x, y + b.y); }
  Point2 operator- (const Point2 &b) const { return Point2(x - b.x, y - b.y); }
  Point2 operator- () const { return Point2(- x, - y); }    

  Point2 unit () const {
    Parameter d = norm();
    return Point2(x/d, y/d);
  }

  Point2 times (const Point2 &b) const {
    Point2 c(b.y, b.x);
    return Point2(crossproduct(c), innerproduct(c));
  }

  Point2 divides (const Point2 &b) const {
    return Point2(innerproduct(b), b.crossproduct(*this));
  }

  Point2 operator* (const Parameter &k) const { return Point2(k*x, k*y); }
  Point2 operator* (double k) const { return Point2(k*x, k*y); }
  Point2 operator/ (const Parameter &k) const { return Point2(x/k, y/k); }
  Point2 operator/ (double k) const { return Point2(x/k, y/k); }
  Point2 rot90 () const { return Point2(- y, x); }

  Parameter circulation (const Point2 &b, const Point2 &c) const {
    return (c - b).crossproduct(*this - b);
  }

  bool inTriangle (const Point2 &a, const Point2 &b, const Point2 &c) {
    return circulation(a, b).sign() == 1 && circulation(b, c).sign() == 1 &&
      circulation(c, a).sign() == 1;
  }

  bool inArc (const Point2 &s, const Point2 &e) const {
    if (s.crossproduct(e).sign() == 1)
      return crossproduct(s).sign() == -1 && crossproduct(e).sign() == 1;
    else
      return crossproduct(s).sign() == -1 || crossproduct(e).sign() == 1;
  }

  Parameter x, y;
};

typedef vector<Point2> Points2;

Point2 operator* (const Parameter &k, const Point2 &a);

Point2 operator* (double k, const Point2 &a);

bool pointInCircumcircle (const Point2 &p, const Point2 &a, 
			  const Point2 &b, const Point2 &c);

Point2 circumCenter (const Point2 &a, const Point2 &b, const Point2 &c);

void linsolve2 (Parameter a, Parameter b, Parameter c, Parameter d, 
		Parameter e, Parameter f, Parameter *xy);

Parameter signedArea (const Points2 &pts);

class Point3 {
 public:
  Point3 () {}
  Point3 (double ix, double iy, double iz) : 
  x(Parameter(ix, ix)), y(Parameter(iy, iy)), z(Parameter(iz, iz)) {}
  Point3 (const Parameter &ix, const Parameter &iy, const Parameter &iz)
    : x(ix), y(iy), z(iz) {}
  Point3 (const Point3 &p) : x(p.x), y(p.y), z(p.z) {}
  Point3 & operator= (const Point3 &p) 
    { x = p.x; y = p.y; z = p.z; return *this; }
  Point3 round () const { return Point3(x.round(), y.round(), z.round()); }
  Point3 perturb () const { return Point3(x.perturb(), y.perturb(), z.perturb()); }

  void increasePrecision () { 
    x.increasePrecision(); y.increasePrecision(); z.increasePrecision();
  }

  void decreasePrecision () { 
    x.decreasePrecision(); y.decreasePrecision(); z.decreasePrecision();
  }

  Parameter getX () const { return x; }
  Parameter getY () const { return y; }
  Parameter getZ () const { return z; }

  int projectionCoordinate () const {
    double vx = x.round(), vy = y.round(), vz = z.round();
    if (fabs(vx) >= max(fabs(vy), fabs(vz)))
      return vx > 0.0 ? 1 : -1;
    if (fabs(vy) >= max(fabs(vx), fabs(vz)))
      return vy > 0.0 ? 2 : -2;
    return vz > 0.0 ? 3 : -3;
  }

  Point2 project (int coord) const {
    bool sflag = coord > 0;
    switch (sflag ? coord : - coord) {
    case 1: return Point2(sflag ? y : - y, z);
    case 2: return Point2(sflag ? z : - z, x);
    case 3: return Point2(sflag ? x : - x, y);
    }
  }

  Parameter coordinate (int coord) const {
    switch (coord) {
    case 1: return x;
    case 2: return y;
    case 3: return z;
    }
  }

  Parameter innerproduct (const Point3 &b) const { 
    return x*b.x + y*b.y + z*b.z; 
  }

  Parameter normSquared () const { return x*x + y*y + z*z; }
  Parameter norm () const { return normSquared().sqrt(); }
  Parameter distance (const Point3 &b) const { return (b - *this).norm(); }

  Point3 operator+ (const Point3 &b) const {
    return Point3(x + b.x, y + b.y, z + b.z);
  }

  Point3 operator- () const {
    return Point3(- x, - y, - z);
  }

  Point3 operator- (const Point3 &b) const {
    return Point3(x - b.x, y - b.y, z - b.z);
  }

  Point3 operator* (const Parameter &k) const {
    return Point3(k*x, k*y, k*z);
  }

  Point3 operator* (double k) const {
    return Point3(k*x, k*y, k*z);
  }

  Point3 operator/ (const Parameter &k) const {
    return Point3(x/k, y/k, z/k);
  }

  Point3 operator/ (double k) const {
    return Point3(x/k, y/k, z/k);
  }

  Point3 unit () const {
  Parameter d = norm();
  return Point3(x/d, y/d, z/d);
  }

  Point3 normal (const Point3 &b) const {
    return crossproduct(b).unit();
  }

  Point3 crossproduct (const Point3 &b) const {
    return Point3(y*b.z - z*b.y, z*b.x - x*b.z, x*b.y - y*b.x);
  }

  Parameter crossproduct (const Point3 &b, const int coord) const {
    int s = coord > 0 ? 1 : -1;
    switch (s*coord) {
    case 1:
      return s*(y*b.z - z*b.y);
    case 2:
      return s*(z*b.x - x*b.z);
    case 3:
      return s*(x*b.y - y*b.x);
    }
  }

  Parameter circulation (const Point3 &b, const Point3 &c) const {
    return x*(b.y*c.z - b.z*c.y) + y*(b.z*c.x - b.x*c.z) + 
      z*(b.x*c.y - b.y*c.x);
  }
  
  Parameter x, y, z;
};

typedef vector<Point3> Points3;

Point3 operator* (const Parameter &k, const Point3 &a);

Point3 operator* (double k, const Point3 &a);

Point3 segmentPlaneIntersectionPoint (const Point3 &e, const Point3 &u, 
				      const Point3 &n, const Point3 &a);

Point3 lineIntersectionPoint (const Point3 &a, const Point3 &u,
			      const Point3 &b, const Point3 &v, int coord);

class Quat {
 public:
  Quat (const Parameter &ix, const Parameter &iy, const Parameter &iz, 
	const Parameter &is)
    : x(ix), y(iy), z(iz), s(is) {}

  Quat (const Point2 &angle, const Point3 &axis) {
    Parameter ch = (0.5*angle.x + 0.5).sqrt(), sh = (0.5 - 0.5*angle.x).sqrt();
    s = ch;
    x = sh*axis.x;
    y = sh*axis.y;
    z = sh*axis.z;
  }

  Point2 angle () const {
    Parameter cos = 1.0 - s*s*2.0, sin2 = 1.0 - cos*cos, sin = sin2.sqrt();
    return Point2(cos, sin);
  }
  
  Quat operator* (const Quat &q) const {
    Parameter rx = s*q.x + q.s*x + y*q.z - z*q.y,
      ry = s*q.y + q.s*y + z*q.x - x*q.z,
      rz = s*q.z + q.s*z + x*q.y - y*q.x,
      rs = s*q.s - (x*q.x + y*q.y + z*q.z);
    return Quat(rx, ry, rz, rs);
  }

  Point3 operator* (const Point3 &p) const {
    Point3 u(x, y, z), up = u.crossproduct(p), uup = u.crossproduct(up);
    Parameter  rx = p.x + (s*up.x + uup.x)*2.0,
      ry = p.y + (s*up.y + uup.y)*2.0, rz = p.z + (s*up.z + uup.z)*2.0;
    return Point3(rx, ry, rz);
  }

  Point3 axis () const { return Point3(x, y, z).unit(); }

  Quat inverse () const { return Quat(x, y, z, - s); }

  Parameter x, y, z, s;
};

class Se3 {
 public:
  Se3 (const Quat &ir, const Point3 &it) : r(ir), t(it) {}

  Se3 inverse () const { 
    Quat ir = r.inverse(); 
    Point3 w = ir*t; 
    return Se3(ir, - w); 
  }
  
  Se3 operator* (const Se3 &s) const { 
    return Se3(r*s.r, r*s.t + t); 
  }

  Point3 operator* (const Point3 &p) const { return r*p + t; }
  
  Point3 t;
  Quat r;
};

Point2 randomPoint2 ();

Point3 randomPoint3 ();

ostream& operator<< (ostream &ostr, const Point2 &a);

void readPoints2 (Points2 &points, istream &istr);

void writePoints2 (const Points2 &points, ostream &ostr);

ostream& operator<< (ostream &ostr, const Point3 &a);

// debug

void pp (const Point2 &p);

void pp (const Point3 &p);

void ppv (const Point2 &p);

void ppv (const Point3 &p);

#endif
