#ifndef SUM
#define SUM

#include <set>
#include <map>

#include "point.h"

class Vertex {
 public:
  Vertex () : parent(0) {}
  Vertex (const Point3 &ip) : p(ip), parent(0) {}
  bool orderZ (Vertex *v);
  bool steeperZ (Vertex *v, Vertex *w);
  Parameter cosZ (Vertex *v) const;
  virtual void increasePrecision () = 0;
  virtual void decreasePrecision () = 0;

  Point3 p;
  Vertex *parent;
};  

class PEdge;
class PFace;
class HFace;

class PVertex : public Vertex {
 public:
  PVertex (const Point3 &ip) : Vertex(ip) {}
  bool compatible (PFace *f) const;
  bool onFace (PFace *f) const;
  void increasePrecision ();
  void decreasePrecision ();
  void increasePrecisionCone ();
  void decreasePrecisionCone ();

  vector<PEdge *> edges;
  set<int> fids;
};

typedef vector<PVertex *> PVertices;

typedef set<PVertex *> PVertexset;

typedef map<PVertex *, PVertex *> PVPVMap;

class PVertexD : public PVertex {
 public:
  PVertexD (Vertex *ia, Vertex *ib) : a(ia), b(ib), PVertex(ia->p - ib->p) {}
  void increasePrecision ();
  void decreasePrecision ();

  Vertex *a, *b;
};

class HFace {
 public:
  HFace (PEdge *ie, HFace *iprev, HFace *inext) 
    : e(ie), prev(iprev), next(inext) {}
  bool conflict (PEdge *f) const;
  bool inCset (PEdge *f) const 
  { return find(cset.begin(), cset.end(), f) != cset.end(); }
  void updateCset (HFace *h, PEdge *f);
  void cone (Points3 &g) const;

  vector<PEdge *> cset;
  PEdge *e;
  HFace *prev, *next;
};

class PVertexCone {
 public:
  PVertexCone (PVertex *iv) : v(iv) {}
  void increasePrecision ();
  void decreasePrecision ();

  PVertex *v;
  Points3 g;
};

typedef vector<PVertexCone *> PVertexCones;

PVertexCone * convexCone (PVertex *v);

HFace * initHull (vector<PEdge *> &edges);

HFace * updateHull (HFace *hull, PEdge *e, bool &flag);

HFace * updateHullAux (HFace *fs, HFace *fe, PEdge *e);

void deleteHull (HFace *hull);

class PEdit;

class PEdge {
 public:
  PEdge () : tail(0), twin(0), face(0), cedit(0), fedit(0) {}
  PEdge (PVertex *itail, PEdge *itwin = 0, PFace *iface = 0)
    : tail(itail), twin(itwin), face(iface), cedit(0), fedit(0) {}
  PVertex * head () const { return twin->tail; }
  Point3 u () const { return twin->tail->p - tail->p; }
  bool convex ();
  bool compatible (PEdge *f, bool &aflag);
  bool compatibleAux (PEdge *f, bool &aflag) const;
  void fids (set<int> &res) const;
  bool coplanar (PEdge *e) const;
  bool coplanar (PEdge *e, PEdge *f) const;
  bool colinear (PEdge *e) const;
  bool colinear (const PVertex *v) const;
  bool acute () const;
  PVertex * oppositeVertex () const;
  int faceIndex () const;
  void increasePrecision ();
  void decreasePrecision ();

  PVertex *tail;
  PEdge *twin;
  PFace *face;
  PEdit *cedit, *fedit;
};

typedef vector<PEdge *> PEdges;

class PFace {
 public:
  PFace () : id(-1) { boundary[0] = boundary[1] = boundary[2] = 0; }
  bool sharedEdge (PFace *g) const;
  bool intersects (PEdge *e) const;
  bool rayIntersection (const Point3 &a, const Point3 &u, Parameter &k) const;
  void increasePrecision();
  void decreasePrecision();

  PEdge *boundary[3];
  Point3 n;
  bool flag;
  int id;
};

typedef vector<PFace *> PFaces;

class PEdit {
 public:
  PEdit (bool icflag, PEdge *iedge, const Parameter &id) 
    : valid(true), cflag(icflag), edge(iedge), d(id) {}
  bool operator< (PEdit &b);
  void increasePrecision ();
  void decreasePrecision ();

  bool valid, cflag;
  PEdge *edge;
  Parameter d;
};

typedef vector<PEdit *> PEdits;

class PEditOrder {
 public:
  bool operator() (PEdit *a, PEdit *b) const;
};

PEdit * nextPEdit (PEdits &edits);

Parameter minAltitude (PVertex *a, PVertex *b, PVertex *c, 
		       PVertex *d, bool &cflag);

enum PFormat { VTK, OFF, TRI };

class Polyhedron {
 public:
  ~Polyhedron ();
  PVertex * addVertex (const Point3 &p);
  PVertex * addVertex (double x, double y, double z) {
    return addVertex(Point3(Parameter(x, x), Parameter(y, y), Parameter(z, z)));
  }
  void addVertex (PVertex *v);
  PEdge * getEdge (PVertex *a, PVertex *b);
  void addFace (PVertex *a, PVertex *b, PVertex *c, int id = -1);
  void convexEdges (PEdges &convex) const;
  Polyhedron * translate (const Point3 &t) const;
  Polyhedron * rotate (const Quat &q) const;
  Polyhedron * negative () const;
  Polyhedron * negativeTranslate (const Point3 &t) const;
  bool intersects (const Polyhedron *b, PVertex *at = 0, PVertex *bt = 0) const;
  bool contains (PVertex *v) const;
  Point3 interiorPoint () const;
  PFace * largestFace () const;
  void simplify (double dmin);
  void simplifyInit (double dmin, PEdits &edits) const;
  void checkEdits (PEdge *e, double dmin, PEdits &edits) const;
  void checkCollapse (PEdge *e, double dmin, PEdits &edits) const;
  void checkFlip (PEdge *e, double dmin, PEdits &edits) const;
  bool checkFlipAux (PEdge *e, double dmin, Parameter &h, bool &cflag) const;
  void collapseEdge (PEdge *e, double dmin, PEdits &edits);
  void mergeVertices (PVertex *a, PVertex *b) const;
  void removeEdge (PEdge *e) const;
  void flipEdge (PEdge *e, double dmin, PEdits &edits);
  void invalidateEdits (PEdge *e) const;
  void invalidateCollapse (PEdge *e) const;
  void invalidateFlip (PEdge *e) const;
  void updateEdits (PVertex *v, bool oflag, double dmin, PEdits &edits) const;  
  void simplifyEnd ();
  double maxBBoxDimension () const;
  bool freeSpace (double *box, double *res) const;
  void output (PFormat format, ostream &ostr) const;

  PVertices vertices;
  PEdges edges;
  PFaces faces;
  double bbox[6];
};

typedef map<Vertex *, int> VIMap;

bool bboxContains (const double *bbox, const Point3 &p);

int getPoint (VIMap &vimap, Points3 &pts, Vertex *v);

void outputVTK (const Points3 &pts, const vector<int> &data, ostream &ostr);

void outputOFF (const Points3 &pts, const vector<int> &data, ostream &ostr);

void outputTRI (const Points3 &pts, const vector<int> &data, ostream &ostr);

Polyhedron * readPolyhedron (istream &istr);

void readPolyhedronVTK (istream &istr, int &nv, double *&v, int &nt, int *&t);

void readPolyhedronOFF (istream &istr, int &nv, double *&v, int &nt, int *&t);

void readPolyhedronTRI (istream &istr, int &nv, double *&v, int &nt, int *&t);

Polyhedron * polyhedron (int nv, double *v, int nt, int *t);

class BSPTree {
 public:
  BSPTree () : left(0), right(0) {}
  BSPTree (const Point3 &ni, BSPTree *ileft, BSPTree *iright) 
    : n(ni), left(ileft), right(iright) {}
  ~BSPTree () { if (left) delete left; if (right) delete right; }
  void find (const Point3 &p, vector<BSPTree *> &res);
  void insert (PVertexCone *c);
  void insert (PEdge *e, bool aflag);
  void leafs (vector<BSPTree *> &res);
  
  Point3 n;
  BSPTree *left, *right;
  PVertices vertices;
  PEdges aedges, bedges;
};

typedef vector<BSPTree *> BSPTrees;

BSPTree * BSPTreeBuild (PVertexCones &cones);

BSPTree * BSPTreeBuild (PEdges &aedges, PEdges &bedges);

BSPTree * BSPTreeBuild (Points3 &pts, int nmax = 1);

class MetaCEdge;
class CEdge;
class CFace;

class CVertex : public Vertex {
 public:
  CVertex () : Vertex(), face(0), va(0), vb(0), fflag(true) 
    { iedges[0] = iedges[1] = 0; }
  CVertex (const Point3 &ip, PVertex *iva = 0, PVertex *ivb = 0)
    : Vertex(ip), face(0), va(iva), vb(ivb), fflag(true)
    { iedges[0] = iedges[1] = 0; }
  CVertex (const Point3 &ip, CFace *iface, CEdge *e1, CEdge *e2 = 0)
    : Vertex(ip), face(iface), va(0), vb(0), fflag(true)
    { iedges[0] = e1; iedges[1] = e2; }

  void increasePrecision ();
  void decreasePrecision ();

  CFace *face;
  CEdge *iedges[2];
  PVertex *va, *vb;
  bool fflag;
};

typedef vector<CVertex *> CVertices;

typedef map<CVertex *, PVertex *> CVPVMap;
typedef map<PVertex *, CVertex *> PVCVMap;
typedef pair<PVertex *, PVertex *> VVPair;
typedef map<VVPair, CVertex *> VVMap;
typedef pair<PEdge *, PEdge *> EEPair;

class Triangle {
 public:
  Triangle () : a(0), b(0), c(0) {}
  Triangle (CVertex *ia, CVertex *ib, CVertex *ic) 
    : a(ia), b(ib), c(ic) {}
  void increasePrecision ();
  void decreasePrecision ();

  CVertex *a, *b, *c;
};

typedef vector<Triangle> Triangles;

class MetaCEdge {
 public:
  MetaCEdge () : tail(0), head(0), fflags(0), edge(0), ne(0), eedges(0) 
    { pthread_mutex_init(&mutex, 0); }
  MetaCEdge (CVertex *itail, CVertex *ihead, bool ibflag, 
	     bool iavflag, PEdge *iedge) 
    : tail(itail), head(ihead), bflag(ibflag), avflag(iavflag),
    fflags(0), edge(iedge), ne(0), nmax(2), eedges(0) 
    { setBBox(); pthread_mutex_init(&mutex, 0); }
  ~MetaCEdge () { delete [] fflags; delete [] eedges; }
  void setBBox ();
  void addEdge (CEdge *e);
  CEdge * getEdge (int i) const 
  { return i < 2 ? edges[i] : eedges[i-2]; }
  Point3 u () const { return head->p - tail->p; }
  void sortVertices ();
  CFace * intersectionFace (CVertex *v) const;
  void formSubedges (bool allfaces, bool fail);
  void subedgesAll () const;
  void prune1 (bool fail);
  bool manifold () const;
  int intersectionSign (CVertex *v, bool fail);
  void prune2 () const;
  void subedgesPruned () const;
  void increasePrecision ();
  void decreasePrecision ();

  CVertex *tail, *head;
  bool bflag, avflag, *fflags;
  PEdge *edge;
  double bbox[6];
  int ne, nmax;
  CEdge *edges[2], **eedges;
  CVertices vertices;
  pthread_mutex_t mutex;
};

typedef vector<MetaCEdge *> MetaCEdges;

class MetaCEdgeOrder {
 public:
  bool operator() (MetaCEdge *a, MetaCEdge *b) const {
    return a->tail < b->tail || 
      a->tail == b->tail && a->head < b->head;
  }
};

typedef set<MetaCEdge *, MetaCEdgeOrder> MetaCEdgeSet;

class EdgeVertexOrder {
 public:
  EdgeVertexOrder (MetaCEdge *ie, bool ifail) : e(ie), fail(ifail) {}
  bool operator() (CVertex *v, CVertex *w) const;
  bool orderAux (CVertex *v, CVertex *w) const;

  MetaCEdge *e;
  bool fail;
};

class Subedge;
class SFace;

class CEdge {
 public:
  CEdge (MetaCEdge *iedge, bool iforward, CFace *iface) 
    : edge(iedge), forward(iforward), face(iface) 
  { edge->addEdge(this); }
  ~CEdge ();
  CVertex * tail () const { return forward ? edge->tail : edge->head; }
  CVertex * head () const { return forward ? edge->head : edge->tail; }
  Point3 u () const { return forward ? edge->u() : - edge->u(); }
  CFace * otherFace () const;
  bool incident (CEdge *e) const;
  int ccw (CFace *f, CFace *g);
  void addSubedge (int i);
  Subedge * findSubedge (CVertex *t, CVertex *h) const;
  void increasePrecision () const { edge->increasePrecision();}
  void decreasePrecision () const { edge->decreasePrecision();}

  MetaCEdge *edge;
  bool forward;
  CFace *face;
  vector<Subedge *> subedges;
};

typedef vector<CEdge *> CEdges;

class VertexOrderF {
 public:
  bool operator() (CVertex *v, CVertex *w) const {
    return v->face < w->face;
  }
};

typedef map<CVertex *, vector<Subedge *> *> SubedgeMap;

class CFace {
 public:
  CFace (PVertex *v, PFace *f) : fv(v), ffa(f), fe(0), ff(0), nb(3), 
    ns(0), id(-1), sfaces(0) { pthread_mutex_init(&mutex, 0); }
  CFace (PEdge *e, PEdge *f) : fv(0), ffa(0), fe(e), ff(f), nb(4), 
    ns(0), id(-1), sfaces(0) { pthread_mutex_init(&mutex, 0); }
  CFace (bool iflag) : fv(0), ffa(0), fe(0), ff(0), nb(3),  ns(0), id(-1), 
    sfaces(0), flag(iflag) { pthread_mutex_init(&mutex, 0); }
  ~CFace ();
  void setBBox ();
  bool coplanar(CFace *g) const;
  bool canIntersect (CFace *g) const;
  bool sharedBoundaryEdge (CFace *g) const;
  bool intersection (CFace *g, bool justreport = false);
  bool intersection (CFace *g, CFace **ifaces, CEdge **iedges, int &ni) const;
  bool coplanarVertex (CVertex *v) const;
  bool canIntersect (CEdge *e) const;
  bool intersection (CEdge *e, Point3 &p) const;
  void formEdges (vector<CFace *> &gfaces, CVertices &vf, CVertices &vg);
  MetaCEdge * formEdge (CFace *g, CVertex *v, CVertex *w);
  void sortEdges ();
  void intersectEdges1 (CEdges &e1, CEdges &e2, CEdges &e3) const;
  void intersectEdges2 (CEdge *ej, CEdge *ek, CEdge *el, CEdges &e1bad, 
			CEdges &e2bad, CEdges &e3bad);
  bool intersection (CEdge *e, CEdge *f) const;
  bool intersectionAux (CEdge *e, CEdge *f) const;
  Point3 edgeIntersectionPoint (CEdge *e, CEdge *f) const;
  bool boundaryVertex (PVertex *v) const;
  bool boundaryVertex (CVertex *v) const;
  CEdge * knownIntersection (CFace *g);
  void sortVertices ();
  void subdivide (bool allfaces, bool fail);
  void subdivideInit (bool fail, SubedgeMap &subedges) const;
  void subdivideInitEdge (SubedgeMap &smap, CEdge *e, bool tflag) const;
  void subdivideInitVertex (bool fail, vector<Subedge *> &subedges) const;
  bool loop (Subedge *e, bool allfaces, bool fail, bool &outer) const;
  SFace * enclosingFace (CVertex *v, bool fail) const;
  void increasePrecision ();
  void decreasePrecision ();
  void decreasePrecisionBV () const;

  PVertex *fv;
  PEdge *fe, *ff;
  PFace *ffa;
  int nb, ns, id;
  CEdge *boundary[4];
  CVertices vertices;
  CEdges edges;
  Point3 n;
  double bbox[6];
  SFace **sfaces;
  bool flag;
  pthread_mutex_t mutex;
};

typedef vector<CFace *> CFaces;

class CFaceEdgeOrder {
 public:
  bool operator() (CEdge *e, CEdge *f) const {
    return e != f && e->otherFace() < f->otherFace();
  }
};

class Shell;

class Cell;

class Convolution {
 public:
  Convolution () {}
  ~Convolution ();
  void addFace (CFace *f);
  void sum (const Polyhedron *a, const Polyhedron *b);
  void sumVF (const Polyhedron *a, const Polyhedron *b, bool avflag, VVMap &vvmap,
	      MetaCEdgeSet &eset);
  void sumVF (PVertex *v, PFace *f, bool avflag, VVMap &vvmap, MetaCEdgeSet &eset);
  void sumEE (const Polyhedron *a, const Polyhedron *b, VVMap &vvmap, 
	      MetaCEdgeSet &eset);  
  void sumEE (PEdge *e, PEdge *f, bool aflag, VVMap &vvmap, MetaCEdgeSet &eset);
  CEdge * sumVE (PVertex *v, PEdge *e, bool avflag, VVMap &vvmap,
		 MetaCEdgeSet &eset, CFace *face);
  CVertex * sumVV (PVertex *a, PVertex *b, bool flag, VVMap &vvmap);
  void subdivide (bool allfaces, bool rbflag);
  void intersectFaces (bool rbflag) const;
  void KDTreeBuild (bool rbflag, CFaces &fa, CFaces &ga) const;
  void sortVerticesF () const;
  void formEdges ();
  void sortEdges () const;
  void intersectEdges () const;
  void sortVertices ();
  void formSubedges (bool allfaces);
  void subdivideFaces (bool allfaces) const;
  void merge (const Polyhedron *p, bool aflag, bool perturb);
  CVertex * getVertex (PVertex *v, PVCVMap &vmap, bool perturb);
  CEdge * addEdge (CVertex *t, CVertex *h, CFace *face, bool avflag, 
		   PEdge *e, MetaCEdgeSet &eset);
  void formCells (bool allfaces, vector<Cell *> &cells) const;
  void formShells (bool allfaces, vector<Shell *> &shells) const;
  bool testIntersection (PVertex *at, PVertex *bt) const;
  void output (ostream &ostr) const;

  CFaces faces;
  MetaCEdges edges;
  CVertices vertices;
};

class FNode {
 public:
  FNode (CFace *iface) : face(iface), label(0u) {}
  bool compatible (const FNode &n) const
  { return (label & n.label) == 0u; }

  CFace *face;
  unsigned int label;
};

typedef vector<FNode> FNodes;

void KDTreeBuild (CVertices &pts1, CVertices &pts2, CVertices &pts3,
		  const FNodes &nodes, int nmax, int coord, bool rbflag,
		  CFaces &fa, CFaces &ga);

void KDTreeLeaf (const FNodes &nodes, bool rbflag, CFaces &fa, CFaces &ga);

void partition (const CVertices &pts, CVertex *p, int coord,
		CVertices &pts1, CVertices &pts2);

void partition (const FNodes &nodes, CVertex *p, int coord,
		FNodes &nodes1, FNodes &nodes2);

class AxisOrder {
 public:
  AxisOrder (int icoord) : coord(icoord) {}
  bool operator() (CVertex *a, CVertex *b) const {
    if (a == b) return false;
    return a->p.coordinate(coord).ub() < b->p.coordinate(coord).ub();
  }

  int coord;
};

bool bboxOverlap (const double *a, const double *b);

#define NPROC 2

enum ThreadType { IntersectFaces, SortVerticesF, FormEdges, SortEdges,
		  IntersectEdges1, IntersectEdges2, SortVertices, 
		  FormSubedges, SubdivideFaces };

class ThreadData {
 public:
  ThreadData () : s(0), d(0), fa(0), ga(0), edges(0), e1(0), e2(0), e3(0) {}

  int *s, d, n;
  ThreadType type;
  const CFaces *fa, *ga;
  CFaces fbad, gbad;
  MetaCEdges *edges;
  MetaCEdges ebad;
  CEdges *e1, *e2, *e3, e1bad, e2bad, e3bad;
  CVertices vbad, wbad;
  bool allfaces;
};

void runThreaded (ThreadData *td, ThreadType type);

void * runThread (void *ptr);

void intersectFaces (CFace *f, CFace *g, CFaces &fbad, CFaces &gbad);

void formEdges (CFace *f, MetaCEdges &edges, CFaces &fbad, CFaces &gbad,
		CVertices &vbad, CVertices &wbad);

void sortVertices (MetaCEdge *e, MetaCEdges &ebad);

void formSubedges (MetaCEdge *e, bool allfaces, MetaCEdges &ebad);

void subdivideFaces (CFace *f, bool allfaces, CFaces &fbad);

class Subedge {
 public:
  Subedge (int nti, bool iforward, CEdge *iedge, Subedge *itwin)
    : nt(nti), forward(iforward), traversed(false), edge(iedge), 
    twin(itwin), next(0), face(0) {}
  CVertex * tail () const;
  CVertex * head () const;
  Point3 u () const {
    return forward == edge->forward ? edge->edge->u() : - edge->edge->u();
  }

  int nt;
  bool forward, traversed;
  CEdge *edge;
  Subedge *twin, *next;
  SFace *face;
};

typedef vector<Subedge *> Subedges;

class SubedgeOrder {
 public:
  SubedgeOrder (const CFace *iface, bool ifail) : face(iface), fail(ifail) {}
  bool operator() (Subedge *e, Subedge *f) const;
  bool order (CVertex *v, Subedge *e, Subedge *f) const;

  const CFace *face;
  bool fail;
};

class HSFace {
 public:
  HSFace () : face(0), flag(false), cell(0) {}
  HSFace * twin () const;
  Shell * formShell (bool allfaces);
  bool neighbors (bool allfaces, vector<HSFace *> &res) const;
  HSFace * neighbor (Subedge *e, bool allfaces) const;
  HSFace * neighborB (Subedge *e, bool allfaces) const;
  HSFace * neighborI (Subedge *e, bool allfaces) const;
  bool aboveZ (CVertex *v) const;
  void triangulate (Triangles &tr) const;

  SFace *face;
  bool outer, flag;
  Cell *cell;
};

typedef vector<HSFace *> HSFaces;

class SFace {
 public:
  SFace (CFace *iface);
  void addBoundary (Subedge *e);
  bool contains (CVertex *v, int coord, bool fail, Parameter &k) const;
  void triangulate (Triangles &tr) const;

  CFace *face;
  Subedges boundary;
  HSFace hfaces[2];
};

void triangulate (vector<CVertices *> &reg, int coord, Triangles &tr);

class Shell {
 public:
  void init ();
  bool isOuter (CVertex *w, HSFace *g1, HSFace *g2);
  CVertex * boundaryVertex () const;

  HSFaces faces;
  CVertex *v;
  bool outer;
};

typedef vector<Shell *> Shells;

class ShellOrder {
 public:
  bool operator() (Shell *a, Shell *b) const {
    return a->v != b->v && b->v->orderZ(a->v);
  }
};

class Cell {
 public:
  Cell () : flag(true), ain(0), bin(0) {}
  ~Cell ();
  void addShell (Shell *s);
  void setBBox ();
  bool contains (CVertex *v) const;
  void triangulate (Triangles &tr) const;

  Shells shells;
  bool flag;
  int ain, bin;
  double bbox[6];
};

typedef vector<Cell *> Cells;

void formCells (const Shells &shells, Cells &cells);

Polyhedron * polyhedron (const Triangles &tr);

PVertex * getVertex (CVertex *v, CVPVMap &vmap, Polyhedron *p);

Polyhedron * repair (Polyhedron *p, Convolution &con);

Polyhedron * repairPoly (const Convolution &con);

Polyhedron * repairPolyInit (const Convolution &con, CVPVMap &vmap);

void repairPolyFace (Polyhedron *p, CVPVMap &vmap, const HSFace &f);

void checkVertices (Polyhedron *p);

void checkVertex (Polyhedron *p, PVertex *v);

void updateCone (Polyhedron *p, PEdges &edges);

Polyhedron * minkowskiSum (Polyhedron *a, Polyhedron *b, double dmin, 
			   bool inner);

void minkowskiShells (Polyhedron *a, Polyhedron *b, bool inner,
		      const Shells &shells, Shells &res);

bool minkowskiShell (Polyhedron *a, Polyhedron *b, Shell *shell);

Polyhedron * negativeTranslate (Polyhedron *p, CVertex *t, PVertex *&at);

enum SetOp { Union, Intersection, Complement };

Polyhedron * setOperation (Polyhedron *a, Polyhedron *b, SetOp op, double dmin);

Cell * enclosingCell (CVertex *v, const Cells &cells);

void labelCells (Cells &cells);

bool inSet (int a, int b, SetOp op);

double pack3 (Polyhedron *a, Polyhedron *b, Polyhedron *c, double dmin, 
	      Point3 *t);

bool pack3 (Polyhedron *a, Polyhedron *b, Polyhedron *c, double dmin, 
	    Point3 *t, double bs);

void freeSpace (double *a, double *b, double *res);

void translateBox (double *a, const Point3 &t, double *res);

void intersectBoxes (double *a, double *b, double *res);

Point3 midpointBox (double *a);

Polyhedron * makeBox (double *b);

void pack3output (Polyhedron *a, Polyhedron *b, Polyhedron *c,
		  Point3 *t, ostream &ostr);

Polyhedron * makeWell (double s, double d);

Polyhedron * pyramid (double s, double h);

Polyhedron * ipyramid (double s, double h);

#endif
