#include "triangulate.h"

bool TVertex::operator< (TVertex &w)
{ 
  bool res;
  try {
    res = p.y < w.p.y;
  }
  catch (SignException &exc) {
    increasePrecision();
    w.increasePrecision();
    res = p.y < w.p.y;
    decreasePrecision();
    w.decreasePrecision();
  }
  return res;
}

void TVertex::link (TVertex *w, bool twinflag)
{
  TEdge *e = new TEdge(this, 0, true, false), 
    *f = new TEdge(w, e, false, twinflag);
  e->twin = f;
  insert(e);
  w->insert(f);
}

void TVertex::insert (TEdge *e)
{
  if (!edge) {
    edge = e;
    edge->next = e;
    return;
  }
  TEdge *a = edge;
  if (a->ccw(e)) {
    while (a->next != edge)
      a = a->next;
    e->next = edge;
    a->next = e;
    edge = e;
    return;
  }
  while (true) {
    TEdge *b = a->next;
    if (b == edge || e->inArc(b, a)) {
      a->next = e;
      e->next = b;
      return;
    }
    a = b;
  }
}

void TVertex::increasePrecision ()
{
  v->increasePrecision();
  p = v->p.project(coord);
}

void TVertex::decreasePrecision ()
{ 
  v->decreasePrecision();
  p.decreasePrecision(); 
}

bool TEdgeOrder::operator() (TEdge *a, TEdge *b) const
{ 
  if (a == b)
    return false;
  if (a->tail == b->tail)
    return b->ccw(a);
  if (a->head() == b->head())
    return a->twin->ccw(b->twin);
  return *a->tail < *b->tail ? a->left(b->tail) : !b->left(a->tail);
}

bool TEdge::left (TVertex *v)
{
  bool res;
  try {
  Parameter dx = head()->p.x - tail->p.x, dy = head()->p.y - tail->p.y,
    x = tail->p.x + (v->p.y - tail->p.y)*dx/dy;
  res = x < v->p.x;
  }
  catch (SignException &exc) {
    increasePrecision();
    v->increasePrecision();
    Parameter dx = head()->p.x - tail->p.x, dy = head()->p.y - tail->p.y,
      x = tail->p.x + (v->p.y - tail->p.y)*dx/dy;
    res = x < v->p.x;
    decreasePrecision();
    v->decreasePrecision();
  }
  return res;
}

bool TEdge::ccw (TEdge *a)
{
  bool res;
  try {
    res = ccwAux(a);
  }
  catch (SignException &exc) {
    increasePrecision();
    a->increasePrecision();
    res = ccwAux(a);
    decreasePrecision();
    a->decreasePrecision();
  }
  return res;
}

bool TEdge::ccwAux (TEdge *a) const
{
  Point2 u = head()->p - tail->p, v = a->head()->p - tail->p;
  int su = u.y.sign(), sv = v.y.sign();
  if (su == -1 && sv == 1)
    return true;
  if (su == 1 && sv == -1)
    return false;
  return u.crossproduct(v).sign() == 1;
}

bool TEdge::inArc (TEdge *a, TEdge *b)
{
  return !ccw(a) && ccw(b);
}

void TEdge::monotoneLoop (TVertices &v)
{
  TEdge *e = this;
  bool flag = false;
  while (!e->flag) {
    flag = flag || *e->tail < *e->head();
    e->tail->flag = flag;
    v.push_back(e->tail);
    e->flag = true;
    e = e->twin->next;
  }
  sort(v.begin(), v.end(), TVertexOrder());
}

void TEdge::increasePrecision ()
{
  tail->increasePrecision();
  head()->increasePrecision();
}

void TEdge::decreasePrecision ()
{
  tail->decreasePrecision();
  head()->decreasePrecision();
}

void triangulate (vector<CVertices *> &reg, int coord, Triangles &tr)
{
  TVertices vertices;
  triangulateInit(reg, coord, vertices);
  makeMonotone(vertices);
  triangulateMonotone(vertices, tr);
  deleteVertices(vertices);
}

void triangulateInit (vector<CVertices *> &reg, int coord, TVertices &vertices)
{
  TVertexMap vmap;
  for (int i = 0; i < reg.size(); ++i) {
    CVertices &b = *reg[i];
    int n = b.size();
    for (int ie = 0; ie < n; ++ie) {
      int ien = ie + 1 == n ? 0 : ie + 1;
      TVertex *t = getTVertex(b[ie], coord, vmap),
	*h = getTVertex(b[ien], coord, vmap);
      t->link(h, true);
    }
  }
  for (TVertexMap::iterator x = vmap.begin(); x != vmap.end(); ++x)
    vertices.push_back(x->second);
}

TVertex * getTVertex (CVertex *v, int coord, TVertexMap &vmap)
{
  TVertexMap::iterator x = vmap.find(v);
  if (x != vmap.end())
    return x->second;
  TVertex *w = new TVertex(v, coord);
  vmap.insert(pair<CVertex *, TVertex *>(v, w));
  return w;
}

void makeMonotone (TVertices &vertices)
{
  sort(vertices.begin(), vertices.end(), TVertexOrder());
  TEdgeSet sweep;
  for (TVertices::iterator v = vertices.begin(); v != vertices.end(); ++v)
    makeMonotone(*v, sweep);
}

void makeMonotone (TVertex *v, TEdgeSet &sweep)
{
  TVertices d;
  bool dflag = true;
  TEdge *e = v->edge;
  while (e->increasingY()) {
    dflag = false;
    if (e->helper->flag)
      d.push_back(e->helper);
    sweep.erase(e);
    e = e->next;
    if (e == v->edge)
      break;
  }
  TEdge *f = findLeft(v, sweep);
  if (f) {
    if (f->helper->flag || dflag)
      d.push_back(f->helper);
    f->helper = v;
  }
  v->flag = true;
  while (!e->increasingY()) {
    v->flag = false;
    e->twin->helper = v;
    sweep.insert(e->twin);
    e = e->next;
    if (e == v->edge)
      break;
  }
  for (TVertices::iterator w = d.begin(); w != d.end(); ++w)
    v->link(*w, false);
}

TEdge * findLeft (TVertex *v, TEdgeSet &sweep)
{
  if (sweep.empty())
    return 0;
  TEdge *e = v->edge->increasingY() ? v->edge : v->edge->twin;
  TEdgeSet::iterator f = sweep.lower_bound(e);
  if (f == sweep.begin())
    return 0;
  --f;
  return (*f)->in ? 0 : *f;
}

void deleteVertices (TVertices &vertices)
{
  for (unsigned int i = 0; i < vertices.size(); ++i) {
    vector<TEdge *> edges;
    TEdge *e = vertices[i]->edge, *e0 = e;
    do {
      edges.push_back(e);
      e = e->next;
    }
    while (e != e0);
    for (unsigned int j = 0; j < edges.size(); ++j)
      delete edges[j];
    delete vertices[i];
  }
}

void triangulateMonotone (TVertices &vertices, Triangles &tr)
{
  for (TVertices::iterator w = vertices.begin(); w != vertices.end(); ++w) {
    TEdge *e0 = (*w)->edge, *e = e0;
    do {
      if (!e->flag) {
	TVertices v;
	e->monotoneLoop(v);
	triangulateLoop(v, tr);
      }
      e = e->next;
    }
    while (e != e0);
  }
}

void triangulateLoop (const TVertices &v, Triangles &tr)
{
  TVertices stack;
  stack.push_back(v[0]);
  stack.push_back(v[1]);
  int n = v.size();
  for (int i = 2; i + 1 < n; ++i)
    if (v[i]->flag == (*(stack.end()-1))->flag)
      triangulateAdd(v[i], stack, tr);
    else
      triangulatePop(v[i], stack, tr, false);
  triangulatePop(v[n-1], stack, tr, true);
}

/* generic version
void triangulateAdd (TVertex *v, TVertices &stack, Triangles &tr)
{
  int n = stack.size() - 1;
  while (n > 0) {
    bool aflag;
    try {
      aflag = stack[n-1]->p.circulation(stack[n]->p, v->p).sign() == 1;
    }
    catch (SignException &exc) {
      stack[n-1]->increasePrecision();
      stack[n]->increasePrecision();
      v->increasePrecision();
      aflag = stack[n-1]->p.circulation(stack[n]->p, v->p).sign() == 1;
      stack[n-1]->decreasePrecision();
      stack[n]->decreasePrecision();
      v->decreasePrecision();
    }
    if (v->flag ? aflag : !aflag)
      break;
    tr.push_back(triangle(v, stack[n], stack[n-1], v->flag));
    stack.pop_back();
    --n;
  }
  stack.push_back(v);
}
*/

void triangulateAdd (TVertex *v, TVertices &stack, Triangles &tr)
{
  int n = stack.size() - 1;
  while (n > 0) {
    int s;
    try {
      s = stack[n-1]->p.circulation(stack[n]->p, v->p).sign();
    }
    catch (SignException &exc) {
      stack[n-1]->increasePrecision();
      stack[n]->increasePrecision();
      v->increasePrecision();
      s = stack[n-1]->p.circulation(stack[n]->p, v->p).sign(false);
      stack[n-1]->decreasePrecision();
      stack[n]->decreasePrecision();
      v->decreasePrecision();
    }
    if (v->flag ? s >= 0 : s <= 0)
      break;
    tr.push_back(triangle(v, stack[n], stack[n-1], v->flag));
    stack.pop_back();
    --n;
  }
  stack.push_back(v);
}

void triangulatePop (TVertex *v, TVertices &stack, Triangles &tr,
		      bool done)
{
  int n = stack.size() - 1;
  TVertex *q = stack[n];
  bool flag = q->flag;
  while (n > 0) {
    tr.push_back(triangle(v, stack[n], stack[n-1], flag));
    --n;
  }
  stack.clear();
  if (done)
    return;
  stack.push_back(q);
  stack.push_back(v);
}

Triangle triangle (TVertex *a, TVertex *b, TVertex *c, bool flag)
{
  return flag ? Triangle(a->v, b->v, c->v) 
    : Triangle(a->v, c->v, b->v);
}

void pt (const Triangle &tr)
{
  cerr << "(";
  pp(tr.a->p);
  pp(tr.b->p);
  pp(tr.c->p);
  cerr << ")" << endl;
}
