static const char *copyright = "copyright (c) 2011 Victor Milenkovic and Elisha Sacks.  Permission required for commercial use.  Contact vjm@cs.miami.edu";

// set b for oldSubs and a for newSubs to event
// SEE/SEE speedup
// neighbors

#include "PolyArc.h"
#include "eigen.h"
#include "crit.h"
#include <limits>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <set>
#include <map>
#include <sys/time.h>
using namespace std;

#ifdef BLEEN
double getTime ()
{
  timeval tv;
  gettimeofday(&tv, 0);
  return tv.tv_sec + 1e-6*tv.tv_usec;
}
#endif

// b is between a and c
bool between (Point2 a, Point2 b, Point2 c) {
  if (a.crossproductSign(c) > 0)
    return a.crossproductSign(b) > 0 && b.crossproductSign(c) > 0;
  else
    return a.crossproductSign(b) > 0 || b.crossproductSign(c) > 0;
}

// (1,0) is between a and c
bool between (Point2 a, Point2 c) {
  if (a.crossproductSign(c) > 0)
    return a.y.sign() < 0 && c.y.sign() > 0;
  else
    return a.y.sign() < 0 || c.y.sign() > 0;
}

bool ignoreDissentingVotes = false;

class CSpace {
public:
  enum Type { V, E, SVE, SEV, SEE, I };

  class SumEE;

  class Obj : public GeoObj {
  public:
    virtual Type type () = 0;
    virtual PV getP (PV theta) { assert(0); return PV(); }
  };
  
  class Edge;

  class Vert : public Obj {
  public:
    Type type () { return V; }
    CSpace *csp;
    int id;
    bool moving;
    Normal *n;
    Edge *t, *h;

    Vert (CSpace *csp, PolyArc::Vert *v, bool moving)
      : csp(csp), id(v->id), moving(moving), n(v->n), t(0), h(0) {}

    virtual void increasePrecision () { n->increasePrecision(); }
    virtual void decreasePrecision () { n->decreasePrecision(); }

    int index () { return id; }
    bool isMoving () { return moving; }
    Parameter distance (Vert *that) {
      if (this == that ||
          this->t == that->h && this->t->c == 0 ||
          that->t == this->h && that->t->c == 0)
        return Parameter(0);
      return this->n->getP().distance(that->n->getP());
    }          
  };

  class Edge : public Obj {
  public:
    Type type () { return E; }
    CSpace *csp;
    int id;
    bool moving;
    bool concave;
    Circle *c;
    Vert *t, *h;

    Edge (CSpace *csp, PolyArc::Edge *e, bool moving)
      : csp(csp), id(e->id), moving(moving), c(e->c), concave(e->concave),
        t(0), h(0) {}

    virtual void increasePrecision () { c->increasePrecision(); }
    virtual void decreasePrecision () { c->decreasePrecision(); }

    int index () { return id; }
    bool isMoving () { return moving; }
    bool isConcave () { return concave; }

    // ab = 0/1 returns vertex with smaller/larger angle
    Vert *getAB (int ab) { return isConcave() == ab ? t : h; }
#ifdef BLEEN
    Vert *getAB (int ab) {
      if (isConcave())
        ab = 1 - ab;
      if (isMoving())
        return csp->mvs[(index() + ab) % csp->mvs.size()];
      else
        return csp->fvs[(index() + ab) % csp->fvs.size()];
    }
#endif
  };

  class EVDist {
    vector<double> vv;
    vector<double> veMin, veMax;
    vector<double> eeMin, eeMax;
    int size, sizeVV, sizeVE;

  public:
    EVDist () : size(0), sizeVV(0), sizeVE(0) {}
    void init(vector<Vert *> verts, vector<Edge *> edges);

    int indVV (int i, int j) {
      assert(i < j && j < size);
      return sizeVV - (size - i) * (size - i - 1) / 2 - i + j - 1;
    }
    int indVE (int i, int j) {
      assert(i < size && j < size);
      return size * i + j;
    }

    const double &getVV (int i, int j) {
      return i < j ? vv[indVV(i,j)] : vv[indVV(j, i)];
    }
    const double &getVEmin (int i, int j) { return veMin[indVE(i,j)]; }
    const double &getVEmax (int i, int j) { return veMax[indVE(i,j)]; }
    const double &getEEmin (int i, int j) {
      return i < j ? eeMin[indVV(i,j)] : eeMin[indVV(j,i)];
    }
    const double &getEEmax (int i, int j) {
      return i < j ? eeMax[indVV(i,j)] : eeMax[indVV(j,i)];
    }
  };

  EVDist evdfm[2];

  class Intn;

  class EventID {
    static int nEvents;
  public:
    int id;

    EventID () : id(nEvents++) {
      if (id == -1 || id == -1)
        cout << "EventID this is it" << endl;
    }
  };

  class Type3;

  class Event : public EventID, public GeoObj {
  public:
    Point2 theta, p;
    Obj *a, *b, *c;
    Point2 n1, n2;
    // bump:  0:+outer, 1:-outer, 2:+inner, 3:-inner
    // stab:  0:+, 1:-
    int type2dir;
    int index;
    Type3 *type3;

    virtual void increasePrecision ();
    virtual void decreasePrecision ();

    Event (Point2 theta)
      : theta(theta), a(0), b(0), c(0), type2dir(-1), index(-1) {}
    Event (Vert *fv, Vert *mv);
    Event (Point2 theta, Point2 p, SumEE *see1, Obj *obj2,
           Point2 n1, Point2 n2, int type2dir)
      : theta(theta), p(p), a(see1), b(obj2), c(0), n1(n1), n2(n2),
        type2dir(type2dir), index(-1) {
      // cout << "Event::Event(7) " << theta.getX() << endl;
      // if (fabs(theta.getX() - 0.90986) < 0.00002)
      // cout << "Event::Event(7) this is it" << endl;
    }
    Event (Point2 theta, Point2 p, Intn *intns[3], Type3 *type3)
      : theta(theta), p(p), a(intns[0]), b(intns[1]), c(intns[2]), type2dir(-1), type3(type3) {
      // cout << "Event::Event(3) " << theta.getX() << endl;
    }

    bool isType1 () const { return a->type() == V; }
    bool isType2 () const { return !isType1() && !isType3(); }
    bool isBumpOrUnbump () const { return c == 0 && type2dir >= 0 && b->type() == SEE; }
    bool isStab () const { return c == 0 && type2dir >= 0 && b->type() != SEE; }
    bool isBump () const {
      return c == 0 && (type2dir == 0 || type2dir == 3) && b->type() == SEE;
    }
    bool isUnbump () const {
      return c == 0 && (type2dir == 1 || type2dir == 2) && b->type() == SEE;
    }
    bool isType3 () const { return c != 0; }

    // Does this bump contain that stab?
    bool contains (const Event *that) const {
      SumEE *see = 0;
      if (a == that->a)
        see = dynamic_cast<SumEE *>(b);
      else if (b == that->a)
        see = dynamic_cast<SumEE *>(a);
      else
        return false;

      if (that->b->type() == SEV) {
        SumEV *sev = dynamic_cast<SumEV *>(that->b);
        return (see->fe == sev->fe &&
                (sev->mv == see->me->t || sev->mv == see->me->h));
      }
      else {
        SumVE *sve = dynamic_cast<SumVE *>(that->b);
        return (see->me == sve->me &&
                (sve->fv == see->fe->t || sve->fv == see->fe->h));
      }
    }

    // v1 on e1 stabbing e2 is related to v2 on e2 stabbing e1
    bool related (const Event *that) const {
      SumV *thisSumV = dynamic_cast<SumV *>(this->b);
      SumV *thatSumV = dynamic_cast<SumV *>(that->b);
      return ((thisSumV->tailOf == that->a || thisSumV->headOf == that->a) &&
              (thatSumV->tailOf == this->a || thatSumV->headOf == this->a));
    }
    
    bool bumpLTstab (const Event *that) const;
    bool compareStab (const Event *that) const;

    bool isType1Type3special (const Event *that) const;
    int type1type3crossProductSign (const Event *that) const;
  };

  static bool betweenEvents (Event *a, Event *b, Event *c);

  class Config;

  class Compare {
  public:
    bool operator() (Event *a, Event *b) {
      if (a->id == 3045 && b->id == 3042 ||
          a->id == 3042 && b->id == 3045) {
        cout << "a->id " << a->id << " b->id " << b->id << endl;
        // cout << "this is it 162" << endl;
      }

      int sa = a->theta.y.sign();
      int sb = b->theta.y.sign();
      if (sa > 0 && sb < 0)
        return true;
      if (sa < 0 && sb > 0)
        return false;

      if (a->isBumpOrUnbump() && b->isStab() && a->contains(b) ||
          b->isBumpOrUnbump() && a->isStab() && b->contains(a)) {
        bool bumpLTstab =
          a->isBumpOrUnbump() ? a->bumpLTstab(b) : !b->bumpLTstab(a);
        
        double axb = a->theta.x.value() * b->theta.y.value() - a->theta.y.value() * b->theta.x.value();
        // int axbSign = a->theta.crossproductSign(b->theta);
        int axbSign = axb > 0 ? 1 : -1;
        
        if (!((axbSign > 0) == bumpLTstab))
          cout << "bad bumpLTstab " << axb << endl;
            
        return bumpLTstab;
      }

      if (a->isStab() && b->isStab() && a->related(b)) {
        double axb = a->theta.x.value() * b->theta.y.value() - a->theta.y.value() * b->theta.x.value();
        // int axbSign = a->theta.crossproductSign(b->theta, true);
        int axbSign = axb > 0 ? 1 : -1;
        // if (axbSign != 0)
        // return axbSign > 0;

        bool aCSb = a->compareStab(b);
        if (!(axbSign == 0 || (axbSign > 0) == aCSb))
          cout << "bad compareStab" << endl;
        return a->compareStab(b);
      }

#ifdef TYPE1TYPE3
      static int count;
      if (++count == 5048)
        cout << "type1type3 this is it";

      if (a->isType1Type3special(b)) {
        bool ret = (a->type1type3crossProductSign(b) > 0);
        Parameter::signFail = false;
        if (ret != (a->theta.crossproductSign(b->theta) > 0))
          cout << "type1type3 conflict" << endl;
        Parameter::signFail = true;
        return ret;
      }
#endif

      try {
        Parameter axb = a->theta.crossproduct(b->theta);
        return axb.sign() > 0;
      } catch (SignException se) {
        a->increasePrecision();
        b->increasePrecision();
        bool ret = a->theta.crossproductSign(b->theta) > 0;
        a->decreasePrecision();
        b->decreasePrecision();
        return ret;
      }
    }

    bool operator() (SumEE **a, SumEE **b) {
      for (int i = 0; i < 3; i++) {
        if (a[i]->lessThan(b[i]))
          return true;
        if (b[i]->lessThan(a[i]))
          return false;
      }
      return false;
    }

    bool operator() (const Config &a, const Config &b) {
      if (a.x != b.x)
        return a.x < b.x;
      if (a.y != b.y)
        return a.y < b.y;
      return a.z < b.z;
    }
  };

  class SumV : public Obj {
  public:
    SumEE *tailOf, *headOf;

    SumV () : tailOf(0), headOf(0) {}
    virtual Point2 getN (Point2 theta) = 0;
  };

  class SumVE : public SumV {
  public:
    Type type () { return SVE; }
    Vert *fv;
    Edge *me;

    SumVE (Vert *fv, Edge *me) : fv(fv), me(me) {}
    virtual void increasePrecision () {
      fv->increasePrecision();
      me->increasePrecision();
    }
    virtual void decreasePrecision () {
      fv->decreasePrecision();
      me->decreasePrecision();
    }
    
    bool isLive ();
    bool isLive (Point2 theta);
    PV getP (PV theta);
    Point2 getN (Point2 theta) { return fv->n->getN(); }
  };

  class SumEV : public SumV {
  public:
    Type type () { return SEV; }
    Edge *fe;
    Vert *mv;

    SumEV (Edge *fe, Vert *mv) : fe(fe), mv(mv) {}
    virtual void increasePrecision () {
      fe->increasePrecision();
      mv->increasePrecision();
      fe->csp->gettype1event(fe->getAB(0)->index(), mv->index())->increasePrecision();
      fe->csp->gettype1event(fe->getAB(1)->index(), mv->index())->increasePrecision();
    }
    virtual void decreasePrecision () {
      fe->decreasePrecision();
      mv->decreasePrecision();
      fe->csp->gettype1event(fe->getAB(0)->index(), mv->index())->decreasePrecision();
      fe->csp->gettype1event(fe->getAB(1)->index(), mv->index())->decreasePrecision();
    }
    
    bool isLive ();
    bool isLive (Point2 theta);
    PV getP (PV theta);
    Point2 getN (Point2 theta) { return theta.times(mv->n->getN()); }
  };

  class SubE;
  class INN;

  class SumEE : public Obj {
  public:
    Type type () { return SEE; }
    Edge *fe, *me;
    Parameter s, r;
  private:
    vector<SubE *> subs;
  public:
    vector<SubE *> &getSubs (bool check=false) {
      if (check && (index() == -1 || index() == 53 || index() == 136))
        cout << "getSubs this is it" << endl;
      return subs;
    }
    vector<SubE *> subs0, allSubs;

    SumEE (Edge *fe, Edge *me) : fe(fe), me(me) {
      s = fe->c->getS() + me->c->getS();
      if (fe->isConcave() || me->isConcave())
        r = -s;
      else
        r = s;
    }

    virtual void increasePrecision () {
      if (s.l == 1e20)
        return;
      fe->increasePrecision();
      me->increasePrecision();
      getEvent(0, 0)->increasePrecision();
      getEvent(0, 1)->increasePrecision();
      getEvent(1, 0)->increasePrecision();
      getEvent(1, 1)->increasePrecision();
      s = fe->c->getS() + me->c->getS();
      if (fe->isConcave() || me->isConcave())
        r = -s;
      else
        r = s;
    }

    virtual void decreasePrecision () {
      if (s.l != 1e20)
        return;
      fe->decreasePrecision();
      me->decreasePrecision();
      getEvent(0, 0)->decreasePrecision();
      getEvent(0, 1)->decreasePrecision();
      getEvent(1, 0)->decreasePrecision();
      getEvent(1, 1)->decreasePrecision();
      s.decreasePrecision();
      r.decreasePrecision();
    }

    Point2 getC (Point2 theta) {
      return fe->c->getC() + me->c->getC().times(theta);
    }

    Point2 getC () {
      return fe->c->getC() + me->c->getC();
    }

    PV getC (PV q) {
      return P2PV(fe->c->getC()) + P2PV(me->c->getC()).times(q);
    }

    PV getN (Obj *o, double z) {
      PV q(z);
      PV c = P2PV(getC());
      PV p = o->getP(q);
      return (p - c) / s.value();
    }      

    PV getP (PV q, PV n) {
      return getC(q) + n * s.value();
    }

    Event *getEvent (int abf, int abm);

    bool isConcave () { return s.sign() < 0; }

    bool isLive (Point2 theta);
    bool contains (Point2 theta, Point2 n);
    bool contains (Event *type3, Point2 n);
    Point2 type3theta (Event *e, SumEE *see012[3]);
    bool isLive ();
    bool contains (Point2 n);

    bool lessThan (const SumEE *that) const {
      return index() < that->index();
      /*
      if (this->fe->id < that->fe->id)
        return true;
      if (this->fe->id > that->fe->id)
        return false;
      if (this->me->id < that->me->id)
        return true;
      if (this->me->id > that->me->id)
        return false;
      return false;
      */
    }

    int intersect(SumEE *that, INN inn[2]);

    SumVE *getAB (int ab) {
      CSpace *csp = fe->csp;
      return csp->sves[csp->sumInd(fe->getAB(ab)->index(), me->index())];
    }

    SumV *getTH (int th);

    int findSub (Point2 theta, Point2 n);
    int index () const { return fe->csp->sumInd(fe->index(), me->index()); }

    int crossproductSign (SumEE *that, Point2 &p, Point2 &q);
  };

  class Intn : public Obj {
  public:
    Type type () { return I; }
    SumEE *const a, *const b;

    Intn (SumEE *a, SumEE *b) : a(a), b(b) {
      if (a->index() == -1 && b->index() == 104)
        cout << "this is it Intn" << endl;
    }

    virtual void increasePrecision () {
      a->increasePrecision();
      b->increasePrecision();
    }

    virtual void decreasePrecision () {
      a->decreasePrecision();
      b->decreasePrecision();
    }

    Point2 getN (Point2 theta, SumEE *see);
    bool operator== (Intn &intn) { return a == intn.a && b == intn.b; }
    bool operator== (Obj &obj) {
      return obj.type() == I && *this == dynamic_cast<Intn &>(obj);
    }
    PV getP (PV theta);
  };

  class INN {
  public:
    Intn *intn;
    Point2 n[2];

    INN () : intn(0) {}

    INN (Intn *intn, Point2 n0, Point2 n1) : intn(intn) {
      this->n[0] = n0;
      this->n[1] = n1;
    }

    Point2 getN (SumEE *see) {
      if (see == intn->a)
        return n[0];
      assert(see == intn->b);
      return n[1];
    }

    void insert(SumEE *see, vector<INN> &inns);
  };

  class Type3 : public GeoObj {
  public:
    Point2 theta, p, qd;
    SumEE *sees[3];
    int index;


    static vector<Type3 *> construct (SumEE *sees[3]);
    static vector<Type3 *> constructSameFE (SumEE *sees[3], int iSameFE);
    static vector<Type3 *> constructSameME (SumEE *sees[3], int iSameME);

    virtual void increasePrecision ();
    virtual void decreasePrecision ();

  private:
    Type3 (SumEE *sees[3], Point2 p, Point2 theta, int index)
      : theta(theta), p(p), index(index) {
      this->sees[0] = sees[0];
      this->sees[1] = sees[1];
      this->sees[2] = sees[2];
    }
    Type3 (SumEE *sees[3], Point2 p, Point2 theta, Point2 qd, int index)
      : theta(theta), p(p), qd(qd), index(index) {
      this->sees[0] = sees[0];
      this->sees[1] = sees[1];
      this->sees[2] = sees[2];
    }
  };

  class ObjAB {
  public:
    Obj *obj; // SVE, SEV, or Intn
    Event *a, *b;

    ObjAB (Obj *obj, Event *a, Event *b) : obj(obj), a(a), b(b) {}

    ObjAB interpolate (ObjAB &next, SumEE *see, double maxStep,
                       vector<PV> &ps, vector<PV> &qs);
  };    

  class Corner {
  public:
    Event *e;
    Obj *o; // SVE, SEV, or Intn
    double ze, zo; // angles or event and object

    Corner (Event *e, Obj *o, double ze=DBL_INF, double zo=DBL_INF)
      : e(e), o(o), ze(ze), zo(zo) {}
  };

  class Crossing;

  class SubE {
  public:
    SumEE *see;
    Obj *t, *h;
    Event *a, *b;
    vector<SubE *> neighbors;
    bool dangling, zerolength, visited;
    SubE *next, *qnext;
    int index;
    double zab[2]; // angles of a and b
    double zabth[2][2]; // angles of t and h at a and b
    int component;

    SubE (SumEE *see)
      : see(see), t(0), h(0), a(0), b(0), next(0), qnext(0),
        dangling(false), zerolength(false), visited(false), component(-1) {
      index = see->fe->csp->allSubs.size();
      see->allSubs.push_back(this);
      see->fe->csp->allSubs.push_back(this);
      zab[0] = zab[1] =
        zabth[0][0] = zabth[0][1] = zabth[1][0] = zabth[1][1] =
        DBL_INF;
    }

    void increasePrecision () {
      see->increasePrecision();
      t->increasePrecision();
      h->increasePrecision();
      if (a) a->increasePrecision();
      if (b) b->increasePrecision();
    }
    void decreasePrecision () {
      see->decreasePrecision();
      t->decreasePrecision();
      h->decreasePrecision();
      if (a) a->decreasePrecision();
      if (b) b->decreasePrecision();
    }

    Obj *getTH (int th) { return th == 0 ? t : h; }
    Intn *getITH (int th) { return dynamic_cast<Intn *>(getTH(th)); }

    void setT (Obj *t) {
      if (t->type() == I) {
        Intn *intn = dynamic_cast<Intn *>(t);
        assert(see == intn->a || see == intn->b);
      }
      this->t = t;
    }

    void setH (Obj *h) {
      if (h->type() == I) {
        Intn *intn = dynamic_cast<Intn *>(h);
        assert(see == intn->a || see == intn->b);
      }
      this->h = h;
    }

    SubE *getNext ();
    SubE *getPrev ();
    void addNeighbor (SubE *that);
    bool isNeighbor  (SubE *that);
    bool isLive (Point2 theta);
    ObjAB shared (SubE *that);
    void writeGrid (ostream &out, int n);
    void drawTight ();

    // Set angles (zab, zabth) based on angles of corner;
    void setZs (Corner corner);
    double dZo (const Corner &c1,  const Corner &c2);
    double dZoLR (const Corner &cLeft,  const Corner &cRight);
    double dZe (const Corner &cLo,  const Corner &cHi);

    // Get shared corners with neighbor.
    void shared (SubE *that, Corner corners[2]);

    Point2 getN (Point2 theta, int th);
    bool contains(Point2 theta, Point2 n);

    bool contains(Crossing *cr);

    bool check(Event *event);
  };

  vector<SubE *> allSubs;
  int notDangling;

  class DSubE {
  public:
    SubE *e;
    int d; // 0 or 1

    DSubE () : e(0), d(-1) {}
    DSubE (SubE *e, int d) : e(e), d(d) {}
    bool operator== (DSubE &b) { return e == b.e && d == b.d; }
    bool operator!= (DSubE &b) { return e != b.e || d != b.d; }

    SumEE *getSEE () { return e->see; }
    Obj *getTH (int th) { return th == d ? e->t : e->h; }
    DSubE nextRight ();
  };

  // CLP *clp;

  static const double DBL_INF;
  static const double randomAngle;

  // Arcs
  vector<Vert *> fvs, mvs;
  vector<Edge *> fes, mes;

  // Constructor
  CSpace () {
    // clp = new CLP();
    // clp = new CLP(3.3703718849344102e-15);
    // clp = new CLP(2.3162591518678383e-14);
    // clp = new CLP(1.1491841918841484e-13);
    // clp = new CLP(6.5973912064555917e-13);
    nSee = 0;
  }

  int nSee;

  CSpace (PolyArc *polyFixed, PolyArc *polyMoving);

  void createArcs (PolyArc *poly,
                   vector<Vert *> &verts, vector<Edge *> & edges);

  void createArcs (vector<PCR> points,
                   vector<Vert *> &verts, vector<Edge *> & edges);

  void createArcs (vector<PCR> &pFixed, vector<PCR> &pMoving) {
    createArcs(pFixed, fvs, fes);
    createArcs(pMoving, mvs, mes);
  }

  void checkArcs (vector<Vert *> &verts, vector <Edge *> &edges);

  void checkArcs () {
    checkArcs(fvs, fes);
    checkArcs(mvs, mes);
  }


  // Events
  set<Event *, Compare> events;


  // Things that are fixed + moving
  int sumInd (int jf, int jm) { return jf + fvs.size() * jm; }
  int sumInd (Edge *e0, Edge *e1) {
    if (e0->isMoving()) {
      assert(!e1->isMoving());
      return sumInd(e1->index(), e0->index());
    }
    assert(e1->isMoving());
    return sumInd(e0->index(), e1->index());
  }

  vector<Event *> type1events;
  Event *gettype1event (int jf, int jm) {
    return type1events[sumInd(jf, jm)];
  }

  void createType1Events ();

  vector<SumEE *> sees;
  vector<SumVE *> sves;
  vector<SumEV *> sevs;

  // Things that are (fixed + moving)^2
  void createSums ();

  void createType2Events ();
  void createOuterBumps (SumEE *see1, SumEE *see2, int iStart=0);
  void createInnerBumps (SumEE *see1, SumEE *see2, int iStart=0);
  void createStabs (SumEE *see, SumVE *sve, int iStart=0);
  void createStabs (SumEE *see, SumEV *sev, int iStart=0);


  // Type 3
  set<SumEE **, Compare> testedTriples;
  void createType3Events (SumEE *see012[3]);
  bool testType3 (Type3 *type3, Intn *intns[3]);

  // Initial Arrangement
  void makeArrangement0 ();

  void checkTriangle (DSubE dsube);
  void checkTriangles (SubE *e) {
    checkTriangle(DSubE(e, 0));
    checkTriangle(DSubE(e, 1));
  }

  // Process Event

  // Type 1
  void processType1Event (Event *event);
  int type1stab (Edge *ffmm[4], int ba);
  void swap (Edge *&a, Edge *&b) {
    Edge *temp = a;
    a = b;
    b = temp;
  }

  // Type 2
  void processBump (Event *event);
  void processStab (Event *event);

  // Type 3
  void processFlip (Event *event);

  // ALL
  void updateSubs (vector<SubE *> &oldSubs, vector<SubE *> &newSubs,
                   Event *event);
  void processEvents ();

  // Graph search
  void findDangling ();
  int findComponents ();
  SubE *findNearest (PV pos, double ang, PV &nearP);
  bool findPath (SubE *start, SubE *finish);

  class Crossings {
  public:
    Point2 theta[2], n[2];
    bool onSee[2];
    Crossings () { onSee[0] = onSee[1] = false; }

    SumEE *see;
    NormalNN *cm;

    Crossings (SumEE *see, NormalNN *cm) : see(see), cm(cm) {
      onSee[0] = onSee[1] = false;
      calculate();
    }
      
    void calculate () {
      Point2 p1 = cm->n1->getP();
      Point2 q1 = cm->n1->getN();
      Point2 p2 = cm->n2->getP();
      Point2 q2 = cm->n2->getN();
      Point2 fc = cm->getP() - see->fe->c->getC();
      Point2 mc = cm->getN() - see->me->c->getC();
      Parameter mcnorm = mc.norm();
      Point2 q[2], cq[2];
      if (!intersectCircles(see->r, fc, mcnorm, q, cq))
        return;
      
      Point2 mcu = mc / mcnorm;
      for (int j = 0; j < 2; j++) {
        theta[j] = cq[j].unit().divides(mcu);
        if (!between(q1, theta[j], q2))
          continue;
        n[j] = q[j] / see->s;
        onSee[j] = see->contains(theta[j], n[j]);
      }
    }

    void increasePrecision () {
      if (n[0].x.l == 1e20 || n[1].x.l == 1e20)
        return;
      see->increasePrecision();
      cm->increasePrecision();
      calculate();
    }
    void decreasePrecision () {
      if (n[0].x.l != 1e20 && n[1].x.l != 1e20)
        return;
      see->decreasePrecision();
      cm->decreasePrecision();
      for (int i = 0; i < 2; i++) {
        theta[i].decreasePrecision();
        n[i].decreasePrecision();
      }
    }      
  };

  class Crossing {
  public:
    SubE *sub;
    Crossings *crs;
    int j;

    Point2 getTheta () { return crs->theta[j]; }
    Point2 getN () { return crs->n[j]; }
    bool isOutward () { return sub->see->isConcave() != j; }

    Crossing (SubE *sub, Crossings *crs, int j)
      : sub(sub), crs(crs), j(j) {}

    bool between (Normal *n, Crossing *that) {
      try {
        return ::between(n->getN(), getTheta(), that->getTheta());
      } catch (SignException e) {
        n->increasePrecision();
        this->increasePrecision();
        that->increasePrecision();
        bool ret = between(n, that);
        n->decreasePrecision();
        this->decreasePrecision();
        that->decreasePrecision();
        return ret;
      }
    }

    void increasePrecision () {
      sub->increasePrecision();
      crs->increasePrecision();
    }

    void decreasePrecision () {
      sub->decreasePrecision();
      crs->decreasePrecision();
    }

    PV getP () { return sub->see->getP(P2PV(getTheta()), P2PV(getN())); }
    PV getQ () { return P2PV(getTheta()); }
  };

  bool findPath (double x1, double y1, double t1,
                 double x2, double y2, double t2,
                 vector<PV> &ps, vector<PV> &qs);
  void angleRange (Point2 p, Point2 q, Point2 qs[]);
  void interpolate (vector<PV> &ps, vector<PV> &qs, PV p2, PV q2);

  int nFigs;
  void drawArcs (PV theta);
  void drawBody (int fm);

  void writePath (const char *file, vector<PV> &ps, vector<PV> &qs);

  void writePath(const char *name,
                 PV startP, PV startQ, PV nearP1, SubE *start,
                 SubE *finish, PV nearP2, PV finishP, PV finishQ,
                 double maxStep);
  ObjAB interpolate (PV p, PV q, ObjAB next, SumEE *see, double maxStep,
                     vector<PV> &ps, vector<PV> &qs);
  void interpolate (PV startP, PV startQ, PV nearP, double maxStep,
                    vector<PV> &ps, vector<PV> &qs);

  void writeGrids (const char *name, int n);

  // tight mesh generation
  void drawTight(double maxStep);
  double maxStep;
  static PV q360;

  static double TwoPI () {
    return 2 * M_PI;
  }

  class Config {
  public:
    double x, y, z;
    Config (PV p, PV q, bool top=false) {
      x = p.x;
      y = p.y;
      z = atan2(q.y, q.x);
      if (!(q == q360) && z < 0 || q == q360 && top)
        z += TwoPI();
    }

    double distance (const Config &b) {
      double dx = x - b.x;
      double dy = y - b.y;
      double dz = z - b.z;
      return sqrt(dx*dx + dy*dy + dz*dz);
    }
  };

  class Triangle {
  public:
    int a, b, c;
    Triangle (int a, int b, int c) : a(a), b(b), c(c) {}
  };

  vector<Config> configs;
  map<Config, int, Compare> conInd;
  vector<Triangle> triangles;

  int addConfig (PV p, PV q, bool top=false) {
    Config config(p, q, top);
    if (conInd.find(config) == conInd.end()) {
      conInd[config] = configs.size();
      configs.push_back(config);
    }
    return conInd[config];
  }

  void addTriangles (int *a, int m, int *b, int n);
  void addTriangle (int a, int b, int c);

  void vertInterp (Obj *obj, Point2 theta0, Point2 theta1, bool top,
                   vector<int> &cons);
  void horizInterp (SumEE *see, Point2 theta, PV pt, PV ph, bool top,
                    vector<int> &cons);
};

PV CSpace::q360;
const double CSpace::DBL_INF = numeric_limits<double>::infinity();
const double CSpace::randomAngle = 0.0123456789;
int CSpace::EventID::nEvents;

void CSpace::EVDist::init(vector<Vert *> verts, vector<Edge *> edges) {
  assert(verts.size() == edges.size());
  size = verts.size();
  sizeVV = size * (size - 1) / 2;
  sizeVE = size * size;
  vv.resize(sizeVV);
  veMin.resize(sizeVE);
  veMax.resize(sizeVE);
  eeMin.resize(sizeVV);
  eeMax.resize(sizeVV);
  
  for (int i = 0; i < size; i++)
    for (int j = i+1; j < size; j++)
      vv[indVV(i,j)] = verts[i]->distance(verts[j]).value();
  
  for (int i = 0; i < size; i++) {
    Vert *v = verts[i];
    for (int j = 0; j < size; j++) {
      Edge *e = edges[j];
      int ind = indVE(i,j);
      double vtDist = v->distance(e->t).value();
      double vhDist = v->distance(e->h).value();
      if (vtDist < vhDist) {
        veMin[ind] = vtDist;
        veMax[ind] = vhDist;
      }
      else {
        veMin[ind] = vhDist;
        veMax[ind] = vtDist;
      }
      
      if (e->c == 0)
        continue;
      
      if (v == e->t || v == e->h ||
          e->t->h->c == 0 && v == e->t->h->t ||
          e->h->t->c == 0 && v == e->h->t->h) {
        veMin[ind] = 0;
        if (v == e->t || e->t->h->c == 0 && v == e->t->h->t)
          veMax[ind] = v->distance(e->h).value();
        else
          veMax[ind] = v->distance(e->t).value();
        continue;
      }
      
      Point2 n = (e->c->getC() - v->n->getP()).unit();
      bool tnxn = e->t->n->getN().crossproduct(n).sign() > 0;
      bool hnxn = e->h->n->getN().crossproduct(n).sign() > 0;
      if (tnxn == hnxn)
        continue;
      
      Point2 p = tnxn ?
        e->c->getC() + n * e->c->getS().abs() :
        e->c->getC() - n * e->c->getS().abs();
      double veDist = v->n->getP().distance(p).value();
      if (veDist < veMin[ind])
        veMin[ind] = veDist;
      else if (veMax[ind] < veDist)
        veMax[ind] = veDist;
    }
  }
  
  for (int i = 0; i < size; i++) {
    Edge *ei = edges[i];
    for (int j = i+1; j < size; j++) {
      Edge *ej = edges[j];
      int ind = indVV(i,j);
      
      if (i == -1 && j == 13)
        cout << "eiej this is it" << endl;
      
      double teMin = getVEmin(ei->t->index(), j);
      double heMin = getVEmin(ei->h->index(), j);
      double etMin = getVEmin(ej->t->index(), i);
      double ehMin = getVEmin(ej->h->index(), i);
      
      eeMin[ind] = teMin;
      if (heMin < eeMin[ind])
        eeMin[ind] = heMin;
      if (etMin < eeMin[ind])
        eeMin[ind] = etMin;
      if (ehMin < eeMin[ind])
        eeMin[ind] = ehMin;

      double teMax = getVEmax(ei->t->index(), j);
      double heMax = getVEmax(ei->h->index(), j);
      double etMax = getVEmax(ej->t->index(), i);
      double ehMax = getVEmax(ej->h->index(), i);
      
      eeMax[ind] = teMax;
      if (heMax > eeMax[ind])
        eeMax[ind] = heMax;
      if (etMax > eeMax[ind])
        eeMax[ind] = etMax;
      if (ehMax > eeMax[ind])
        eeMax[ind] = ehMax;

      if (ei->t == ej->h || ei->h == ej->t)
        continue;

      if (ei->c == 0 || ej->c == 0)
        continue;

      //Point2 n = ei->c->getC().unit(ej->c->getC());
      Point2 n = (ej->c->getC() - ei->c->getC()).unit();

      bool tinxn = ei->t->n->getN().crossproductSign(n) > 0;
      bool hinxn = ei->h->n->getN().crossproductSign(n) > 0;
      if (tinxn == hinxn)
        continue;

      bool tjnxn = ej->t->n->getN().crossproductSign(n) > 0;
      bool hjnxn = ej->h->n->getN().crossproductSign(n) > 0;
      if (tjnxn == hjnxn)
        continue;

      Point2 pi = tinxn ?
        ei->c->getC() + n * ei->c->getS().abs() :
        ei->c->getC() - n * ei->c->getS().abs();
      Point2 pj = tjnxn ?
        ej->c->getC() + n * ej->c->getS().abs() :
        ej->c->getC() - n * ej->c->getS().abs();
      double eeDist = pi.distance(pj).value();

      if (eeDist < eeMin[ind])
        eeMin[ind] = eeDist;
      else if (eeDist > eeMax[ind])
        eeMax[ind] = eeDist;
    }
  }

  for (int i = 0; i < sizeVE; i++) {
    if (veMin[i] > 1e-6)
      veMin[i] -= 1e-6;
    veMax[i] += 1e-6;
  }

  for (int i = 0; i < sizeVV; i++) {
    if (eeMin[i] > 1e-6)
      eeMin[i] -= 1e-6;
    eeMax[i] += 1e-6;
  }
}

void reportEvent2 () {
    cout << "Event::Event(2) this is it" << endl;
}

CSpace::Event::Event (Vert *fv, Vert *mv) : a(fv), b(mv), c(0), type2dir(-1) {
  // mv *theta = fv
  // theta = fv / mv
  theta = fv->n->getN().divides(mv->n->getN());
  p = fv->n->getP() + mv->n->getP().times(theta);
  // cout << "Event::Event(2) " << theta.getX() << endl;
  // if (fabs(theta.getX() - 0.90986) < 0.00002)
  // reportEvent2();
}

void CSpace::Event::increasePrecision () {
  if (id == 598)
    cout << "this is it Event::increasePrecision" << endl;
  if (theta.x.l == 1e20)
    return;

  if (c == 0) {
    a->increasePrecision();
    b->increasePrecision();
  }

  if (isType1()) {
    Vert *fv = dynamic_cast<Vert*>(a);
    Vert *mv = dynamic_cast<Vert*>(b);
    theta = fv->n->getN().divides(mv->n->getN());
    p = fv->n->getP() + mv->n->getP().times(theta);
    return;
  }

  if (isBumpOrUnbump() && type2dir <= 1) {
    SumEE *see1 = dynamic_cast<SumEE*>(a);
    SumEE *see2 = dynamic_cast<SumEE*>(b);

    // |(fc2 + mc2 t) - (fc1 + mc1 t)| = r1 + r2
    // |(fc2 - fc1) + (mc2 - mc1) t| = r1 + r2
    // |fc + mc t| = r
    Point2 fc = see2->fe->c->getC() - see1->fe->c->getC();
    Point2 mc = see2->me->c->getC() - see1->me->c->getC();
    Parameter mcnorm = mc.norm();
    Parameter r = see1->r + see2->r;
    Point2 q[2], cq[2];
    if (!intersectCircles(r, fc, mcnorm, q, cq))
      assert(0);
    
    Point2 mcu = mc / mcnorm;
    int i = type2dir;
    Point2 n = q[i].unit();
    Point2 nneg = -n;
      
    theta = cq[i].unit().divides(mcu);
    n1 = see1->isConcave() ? nneg : n;
    n2 = see2->isConcave() ? n : nneg;
    p = see1->getC(theta) + n1 * see1->s;
    return;
  }
    
  if (isBumpOrUnbump() && type2dir >= 2) {
    SumEE *see1 = dynamic_cast<SumEE*>(a);
    SumEE *see2 = dynamic_cast<SumEE*>(b);

    // |(fc2 + mc2 t) - (fc1 + mc1 t)| = r1 - r2
    // |(fc2 - fc1) + (mc2 - mc1) t| = r1 - r2
    // |fc + mc t| = r
    Point2 fc = see2->fe->c->getC() - see1->fe->c->getC();
    Point2 mc = see2->me->c->getC() - see1->me->c->getC();
    Parameter mcnorm = mc.norm();
    Parameter r = see1->r - see2->r;
    Point2 q[2], cq[2];
    if (!intersectCircles(r, fc, mcnorm, q, cq))
      assert(0);

    Point2 mcu = mc / mcnorm;
    int i = type2dir - 2;
    Point2 n = q[i].unit();
    Point2 nneg = -n;
    
    theta = cq[i].unit().divides(mcu);
    n1 = see1->isConcave() ? nneg : n;
    n2 = see2->isConcave() ? nneg : n;
    p = see1->getC(theta) + n1 * see1->s;
    return;
  }

  if (isStab() && b->type() == SVE) {
    SumEE *see = dynamic_cast<SumEE*>(a);
    SumVE *sve = dynamic_cast<SumVE*>(b);

    // c1 = fc1 + mc1 t
    // v2 = p2 + c2 t + n2 s2
    // |(p2 + c2 t + n2 s2) - (fc1 + mc1 t)| = r1
    // |(p2 + n2 s2 - fc1) + (c2 - mc1) t| = r1
    // |fc + mc t| = r
    Point2 fc = sve->fv->n->getP() + sve->fv->n->getN() * sve->me->c->getS() - see->fe->c->getC();
    Point2 mc = sve->me->c->getC() - see->me->c->getC();
    Parameter mcnorm = mc.norm();
    Parameter r = see->r;
    Point2 q[2], cq[2];
    if (!intersectCircles(r, fc, mcnorm, q, cq))
      assert(0);
    
    Point2 mcu = mc / mcnorm;
    int i = type2dir;
    theta = cq[i].unit().divides(mcu);
    Point2 n = q[i].unit();
    Point2 nneg = -n;
    n1 = see->isConcave() ? nneg : n;
    n2 = sve->fv->n->getN();
    p = see->getC(theta) + n1 * see->s;
    return;
  }

  if (isStab() && b->type() == SEV) {
    SumEE *see = dynamic_cast<SumEE*>(a);
    SumEV *sev = dynamic_cast<SumEV*>(b);

    // c1 = fc1 + mc1 t
    // v2 = c2 + n2 t s2 + p2 t
    // |(c2 + n2 t s2 + p2 t) - (fc1 + mc1 t)| = r1
    // |(c2 - fc1) + (n2 s + p2 - mc1) t| = r1
    // |fc + mc t| = r
    Point2 fc = sev->fe->c->getC() - see->fe->c->getC();
    Point2 mc = sev->mv->n->getN() * sev->fe->c->getS() + sev->mv->n->getP() - see->me->c->getC();
    Parameter mcnorm = mc.norm();
    Parameter r = see->r;
    Point2 q[2], cq[2];
    if (!intersectCircles(r, fc, mcnorm, q, cq))
      assert(0);
    
    Point2 mcu = mc / mcnorm;
    int i = type2dir;
    theta = cq[i].unit().divides(mcu);
    Point2 n = q[i].unit();
    Point2 nneg = -n;
    n1 = see->isConcave() ? nneg : n;
    n2 = sev->mv->n->getN().times(theta);
    // Point2 p = linearCombination(see->getC(theta), see->s, n1);
    p = see->getC(theta) + n1 * see->s;
    return;
  }

  type3->increasePrecision();
  theta = type3->theta;
  p = type3->p;
}

void CSpace::Event::decreasePrecision () {
  if (id == 598)
    cout << "this is it Event::decreasePrecision" << endl;
  if (theta.x.l != 1e20)
    return;

  if (c == 0) {
    a->decreasePrecision();
    b->decreasePrecision();
  }
  else
    type3->decreasePrecision();

  theta.decreasePrecision();
  p.decreasePrecision();
  if (isType2()) {
    n1.decreasePrecision();
    n2.decreasePrecision();
  }
}

bool CSpace::Event::bumpLTstab (const Event *that) const {
  SumEE *see1, *see2;
  if (a == that->a) {
    see1 = dynamic_cast<SumEE *>(a);
    see2 = dynamic_cast<SumEE *>(b);
  }
  else if (b == that->a) {
    see1 = dynamic_cast<SumEE *>(b);
    see2 = dynamic_cast<SumEE *>(a);
  }
  else
    assert(0);

  bool isOuter = (type2dir == 0 || type2dir == 1);
  bool isBump = (type2dir == 0 || type2dir == 3);

  Point2 f12 = see2->fe->c->getC() - see1->fe->c->getC();
  Point2 fu = f12.unit();
  Point2 m12 = see2->me->c->getC() - see1->me->c->getC();
  Point2 mu = m12.unit();
  Point2 mid = fu.divides(mu);

  if ((mid.crossproductSign(that->theta) < 0) ^ isOuter ^ isBump)
    return this->theta.crossproductSign(that->theta) > 0;

  // cout << "bumpLTstab test " << isOuter << isBump << endl;
  return isBump;
}

#ifdef BLEEN
bool CSpace::Event::compareStab (const Event *that) const {
  SumEE *e2 = dynamic_cast<SumEE *>(this->a);
  SumV *v1 = dynamic_cast<SumV *>(this->b);
  SumEE *e1 = dynamic_cast<SumEE *>(that->a);
  SumV *v2 = dynamic_cast<SumV *>(that->b);

  Point2 theta1 = this->theta;
  Point2 n1 = v1->getN(theta1);
  Point2 n2 = v2->getN(theta1);

  static double maxN1xN2;
  double currN1xN2 = fabs(P2PV(n1) % P2PV(n2));
  if (maxN1xN2 < currN1xN2) {
    cerr << "maxN1xN2 " << currN1xN2 << endl;
    maxN1xN2 = currN1xN2;
  }

  bool outer = (this->type2dir == that->type2dir);

  // (n1.innerproductSign(n2) < 0) ^ e1->isConcave() ^ e2->isConcave();

  /*
  if (outer)
    assert(this->type2dir == that->type2dir);
  else
    assert(this->type2dir != that->type2dir);
  */

  bool swapped = false;
  if (!outer && e1->r->compareTo(e2->r) > 0) {
    e1 = dynamic_cast<SumEE *>(this->a);
    v2 = dynamic_cast<SumV *>(this->b);
    e2 = dynamic_cast<SumEE *>(that->a);
    v1 = dynamic_cast<SumV *>(that->b);
    theta1 = that->theta;
    n1 = v1->getN(theta1);
    n2 = v2->getN(theta1);
    swapped = true;
  }

  bool bumping = outer ^ (swapped ? that->type2dir : this->type2dir);

  assert(v1->tailOf == e1 || v1->headOf == e1);
  bool v1min = (v1->tailOf == e1) ^ e1->isConcave();

  Point2 c1 = e1->getC(theta1);
  Point2 c2 = e2->getC(theta1);

  Point2 u12 = outer ? (c2 - c1).unit() : (c1 - c2).unit();

  Parameter *cross1 =
    e1->isConcave() ? n1.crossproduct(u12) : u12.crossproduct(n1);
  Parameter *cross2 =
    e2->isConcave() ? n2.crossproduct(u12) : u12.crossproduct(n2);

  int sign1 = cross1.sign();
  int sign2 = cross2.sign();

  if (sign1 != sign2)
    return swapped ^ bumping ^ v1min ^ (cross1 > 0);

  bool v1closer =
    sign1 > 0 ? cross1->compareTo(cross2) < 0 : cross2->compareTo(cross1);

  return swapped ^ bumping ^ v1min ^ v1closer;
}
#endif

bool CSpace::Event::compareStab (const Event *that) const {
  try {
  SumEE *e2 = dynamic_cast<SumEE *>(this->a);
  SumV *v1 = dynamic_cast<SumV *>(this->b);
  SumEE *e1 = dynamic_cast<SumEE *>(that->a);
  SumV *v2 = dynamic_cast<SumV *>(that->b);

  Point2 theta1 = this->theta;
  Point2 theta2 = that->theta;

  Point2 f1 = e1->fe->c->getC();
  Point2 m1 = e1->me->c->getC();
  Point2 f2 = e2->fe->c->getC();
  Point2 m2 = e2->me->c->getC();

  Point2 f = f2 - f1;
  Point2 m = m2 - m1;

  Point2 n1;
  Parameter *s1 = 0;
  if (v1->type() == SVE) {
    SumVE *sve = dynamic_cast<SumVE *>(v1);
    n1 = sve->fv->n->getN();
  }
  else {
    SumEV *sev = dynamic_cast<SumEV *>(v1);
    n1 = sev->mv->n->getN();
  }
    
  Point2 n2;
  if (v2->type() == SVE) {
    SumVE *sve = dynamic_cast<SumVE *>(v2);
    n2 = sve->fv->n->getN();
  }
  else {
    SumEV *sev = dynamic_cast<SumEV *>(v2);
    n2 = sev->mv->n->getN();
  }
    
  Point2 p1 = n1 * e1->s;
  Point2 p2 = n2 * e2->s;

  bool plusStab2 = (that->type2dir == 0);

  Point2 mid = v2->type() == SVE ?
    (f + p2).unit().divides(m.unit()) :
    f.unit().divides((m + p2).unit());

  assert((mid.crossproductSign(theta2) > 0) == plusStab2);

  if ((mid.crossproductSign(theta1) < 0) == plusStab2)
    return theta1.crossproductSign(theta2) > 0;

  Point2 c = f + theta1.times(m);
  Parameter cx = c.norm();
  Point2 u = c / cx;

  Point2 a1 = v1->type() == SVE ? p1 : theta1.times(p1);
  Point2 a2 = v2->type() == SVE ? p2 : theta1.times(p2);

  Point2 xy1(u.innerproduct(a1), u.crossproduct(a1));
  Point2 xy2(u.innerproduct(a2), u.crossproduct(a2));

  return (plusStab2 ^
          ((xy1.y - xy2.y).sign() < 0) ^
          ((xy1.y + xy2.y).sign() > 0) ^
          ((xy1.x + xy2.x - cx).sign() > 0));          
  } catch (SignException e) {
    ((Event *)this)->increasePrecision();
    ((Event *)that)->increasePrecision();
    bool ret = compareStab(that);
    ((Event *)this)->decreasePrecision();
    ((Event *)that)->decreasePrecision();
    return ret;
  }
}

bool CSpace::Event::isType1Type3special (const Event *that) const {
  if (this->isType3() && that->isType1())
    return that->isType1Type3special(this);
  if (!(this->isType1() && that->isType3()))
    return false;

  Intn *intn012[] = { dynamic_cast<Intn *>(that->a),
                      dynamic_cast<Intn *>(that->b),
                      dynamic_cast<Intn *>(that->c) };
  for (int i = 0; i < 3; i++) {
    SumEE *seeA = intn012[i]->a;
    SumEE *seeB = intn012[i]->b;
    if (seeA->fe->h == seeB->fe->t && seeA->fe->h == a &&
        seeB->me->h == seeA->me->t && seeB->me->h == b ||
        seeB->fe->h == seeA->fe->t && seeB->fe->h == a &&
        seeA->me->h == seeB->me->t && seeA->me->h == b)
      return true;
  }

  return false;
}

int CSpace::Event::type1type3crossProductSign (const Event *that) const {
  if (this->isType3() && that->isType1())
    return -that->type1type3crossProductSign(this);

  Vert *vf = dynamic_cast<Vert *>(a);
  Vert *vm = dynamic_cast<Vert *>(b);
  SumEE *see14 = 0, *see23 = 0;

  Intn *intn012[] = { dynamic_cast<Intn *>(that->a),
                      dynamic_cast<Intn *>(that->b),
                      dynamic_cast<Intn *>(that->c) };
  for (int i = 0; i < 3; i++) {
    SumEE *seeA = intn012[i]->a;
    SumEE *seeB = intn012[i]->b;
    if (seeB->fe->t == vf && seeA->fe->h == vf &&
        seeA->me->t == vm && seeB->me->h == vm ||
        seeA->fe->t == vf && seeB->fe->h == vf &&
        seeB->me->t == vm && seeA->me->h == vm) {
      if (seeA->r < seeB->r) {
        see14 = seeB;
        see23 = seeA;
      }
      else {
        see14 = seeA;
        see23 = seeB;
      }
    }
  }
  assert(see14 != 0);

  Edge *e1 = see14->fe, *e2 = see23->fe, *e3 = see23->me, *e4 = see14->me;
  Point2 u = this->theta;
  Point2 p = that->p;
  Point2 q = that->theta;
  Point2 c14 = see14->getC(q);
  Point2 c23 = see23->getC(q);
  Point2 cc = c23 - c14;
  Point2 c14p = p - c14;
  Point2 c23p = p - c23;

  int proxy = 0;
  if (see14->s.sign() > 0) {
    assert(see23->s.sign() < 0);
    assert(e3->c->getS() < e4->c->getS());
    assert(e2->c->getS() < e1->c->getS());
    if ((e2->c->getS() + e4->c->getS()).sign() > 0) {
      if (vf->n->getN().innerproduct(cc).sign() > 0)
        proxy = vf->n->getN().crossproduct(c14p).sign();
    }
    else /* s2 + s4 < 0 */ {
      if (u.innerproduct(q).sign() > 0)
        proxy = q.times(vm->n->getN()).crossproduct(c23p).sign();
    }
  }
  else /* see14->s.sign() < 0 */ {
    if (see23->s.sign() > 0) {
      assert(e4->c->getS() < e3->c->getS());
      assert(e1->c->getS() < e2->c->getS());
      if ((e2->c->getS() + e4->c->getS()).sign() < 0) {
        if (vf->n->getN().innerproductSign(cc) < 0)
          proxy = -vf->n->getN().crossproductSign(c14p);
      }
      else /* s2 + s4 > 0 */ {
        if (u.innerproductSign(q) > 0)
          proxy = -q.times(vm->n->getN()).crossproductSign(c23p);
      }
    }
    else /* s2 + s3 < 0 */ {
      if (e4->c->getS() < e3->c->getS()) {
        assert(e2->c->getS() < e1->c->getS());
        if (vf->n->getN().innerproductSign(cc) < 0)
          proxy = -vf->n->getN().crossproductSign(c14p);
      }
      else /* -s3 + s4 > 0 */ {
        if ((e2->c->getS() + e4->c->getS()).sign() > 0) {
          proxy = vf->n->getN().crossproductSign(c14p);
        }
        else /* s2 + s4 < 0 */ {
          if (u.innerproductSign(q) > 0)
            proxy = q.times(vm->n->getN()).crossproductSign(c23p);
        }
      }
    }
  }
  
  if (proxy == 0)
    return u.crossproductSign(q);

  double uxq = P2PV(u) % P2PV(q);

  /*
  if ((see14->s.sign() > 0) !=
      ((e1->s - e2->s) + (e4->s - e3->s) * u.innerproduct(q)).sign() > 0)
    return u.crossproductSign(q);

  double uxq = P2PV(u) % P2PV(q);
  int proxy = vf->n.crossproductSign(p - see14->getC(q));

  if (e4->s < e3->s)
    proxy = -proxy;
  */

  static double maxUXQ;
  if ((uxq > 0) != (proxy > 0) && fabs(uxq) > maxUXQ) {
    maxUXQ = fabs(uxq);
    cout << "maxUXQ " << maxUXQ << endl;
    maxUXQ = maxUXQ;
  }

  return proxy;
}

void CSpace::createType1Events () {
  int fm = fvs.size() * mvs.size();
  type1events.resize(fm);

  for (int jf = 0; jf < fvs.size(); jf++)
    for (int jm = 0; jm < mvs.size(); jm++) {
      Event *event = new Event(fvs[jf], mvs[jm]);
      type1events[sumInd(jf, jm)] = event;
      events.insert(event);
    }
}

void CSpace::createSums () {
  int nf = fvs.size();
  int nm = mvs.size();
  assert(fes.size() == nf);
  assert(mes.size() == nm);
  int fm = nf * nm;
  sees.resize(fm);
  sves.resize(fm);
  sevs.resize(fm);
  for (int i = 0; i < fm; i++) {
    sees[i] = 0;
    sves[i] = 0;
    sevs[i] = 0;
  }

  for (int ife = 0; ife < nf; ife++) {
    Edge *fe = fes[ife];
    int ift = fe->t->index();
    assert(fvs[ift] == fe->t);
    int ifh = fe->h->index();
    
    for (int ime = 0; ime < nm; ime++) {
      Edge *me = mes[ime];
      int imt = me->t->index();
      int imh = me->h->index();

      if (fe->isConcave() && me->isConcave())
        continue;

      if (fe->c == 0 || me->c == 0)
        continue;

      if (fe->isConcave() != me->isConcave()) {
        if ((fe->c->getS() + me->c->getS()).sign() > 0)
          continue;
      }

      SumEE *see = new SumEE(fe, me);
      nSee++;
      sees[sumInd(ife, ime)] = see;

      SumVE *&svet = sves[sumInd(ift, ime)];
      if (svet == 0)
        svet = new SumVE(fvs[ift], me);
      if (!me->isConcave()) {
        assert(svet->tailOf == 0);
        svet->tailOf = see;
        // if (svet->fv->id == 44 && svet->me->id == 3)
        // cout << "this is it sve 1" << endl;
      }
      else {
        assert(svet->headOf == 0);
        svet->headOf = see;
        // if (svet->fv->id == 44 && svet->me->id == 3)
        // cout << "this is it sve 2" << endl;
      }

      SumEV *&sevt = sevs[sumInd(ife, imt)];
      if (sevt == 0)
        sevt = new SumEV(fe, mvs[imt]);
      if (!fe->isConcave()) {
        assert(sevt->tailOf == 0);
        sevt->tailOf = see;
      }
      else {
        assert(sevt->headOf == 0);
        sevt->headOf = see;
      }

      SumVE *&sveh = sves[sumInd(ifh, ime)];
      if (sveh == 0)
        sveh = new SumVE(fvs[ifh], me);
      if (!me->isConcave()) {
        assert(sveh->headOf == 0);
        sveh->headOf = see;
        // if (sveh->fv->id == 44 && sveh->me->id == 3)
        // cout << "this is it sve 3" << endl;
      }
      else {
        assert(sveh->tailOf == 0);
        sveh->tailOf = see;
        // if (sveh->fv->id == 44 && sveh->me->id == 3)
        // cout << "this is it sve 4" << endl;
      }

      SumEV *&sevh = sevs[sumInd(ife, imh)];
      if (sevh == 0)
        sevh = new SumEV(fe, mvs[imh]);
      if (!fe->isConcave()) {
        assert(sevh->headOf == 0);
        sevh->headOf = see;
      }
      else {
        assert(sevh->tailOf == 0);
        sevh->tailOf = see;
      }
    }
  }
}

void CSpace::createType2Events () {
  evdfm[0].init(fvs, fes);
  evdfm[1].init(mvs, mes);

  int fm = fvs.size() * mvs.size();
  for (int i = 0; i < fm; i++)
    for (int j = 0; j < fm; j++) {
      if (sees[i] == 0)
        continue;
      if (i < j && sees[j] != 0) {
        createOuterBumps(sees[i], sees[j]);
        createInnerBumps(sees[i], sees[j]);
      }
      if (sves[j] != 0)
        createStabs(sees[i], sves[j]);
      if (sevs[j] != 0)
        createStabs(sees[i], sevs[j]);
    }
}

bool CSpace::SumVE::isLive () {
  Event *eventA = fv->csp->gettype1event(fv->index(), me->getAB(0)->index());
  Event *eventB = fv->csp->gettype1event(fv->index(), me->getAB(1)->index());
  return between(eventB->theta, eventA->theta);
}

bool CSpace::SumVE::isLive (Point2 theta) {
  Event *eventA = fv->csp->gettype1event(fv->index(), me->getAB(0)->index());
  Event *eventB = fv->csp->gettype1event(fv->index(), me->getAB(1)->index());
  return between(eventB->theta, theta, eventA->theta);
}

bool CSpace::SumEV::isLive () {
  Event *eventA = fe->csp->gettype1event(fe->getAB(0)->index(), mv->index());
  Event *eventB = fe->csp->gettype1event(fe->getAB(1)->index(), mv->index());
  return between(eventA->theta, eventB->theta);
}

bool CSpace::SumEV::isLive (Point2 theta) {
  Event *eventA = fe->csp->gettype1event(fe->getAB(0)->index(), mv->index());
  Event *eventB = fe->csp->gettype1event(fe->getAB(1)->index(), mv->index());
  return between(eventA->theta, theta, eventB->theta);
}

CSpace::Event *CSpace::SumEE::getEvent (int abf, int abm) {
  int jabf = fe->getAB(abf)->index();
  int jabm = me->getAB(abm)->index();
  return fe->csp->gettype1event(jabf, jabm);
}

bool CSpace::SumEE::isLive (Point2 theta) {
  Event *ab = getEvent(0, 1);
  Event *ba = getEvent(1, 0);
  return between(ab->theta, theta, ba->theta);
}

bool CSpace::SumEE::contains (Point2 theta, Point2 n) {
  Event *ab = getEvent(0, 1);
  Event *ba = getEvent(1, 0);
  if (!between(ab->theta, theta, ba->theta))
    return false;

  Event *aa = getEvent(0, 0);
  Point2 na;
  if (between(ab->theta, theta, aa->theta))
    na = fe->getAB(0)->n->getN();
  else
    na = me->getAB(0)->n->getN().times(theta);

  Event *bb = getEvent(1, 1);
  Point2 nb;
  if (between(bb->theta, theta, ba->theta))
    nb = fe->getAB(1)->n->getN();
  else
    nb = me->getAB(1)->n->getN().times(theta);

  return between(na, n, nb);
}

bool CSpace::betweenEvents (Event *a, Event *b, Event *c) {
  static int count;
  if (++count == 90)
    cout << "betweenEvents this is it" << endl;
  Compare compare;
  bool ret;
  if (compare(a, c))
    ret = compare(a, b) && compare(b, c);
  else
    ret = compare(a, b) || compare(b, c);
  Parameter::signFail = false;
  if (ret != between(a->theta, b->theta, c->theta))
    cout << "type1type3 conflict in betweenEvents" << endl;
  Parameter::signFail = true;
  return ret;
}

bool CSpace::SumEE::contains (Event *type3, Point2 n) {
  Event *ab = getEvent(0, 1);
  Event *ba = getEvent(1, 0);
  if (!betweenEvents(ab, type3, ba))
    return false;

  Event *aa = getEvent(0, 0);
  Point2 na;
  if (betweenEvents(ab, type3, aa))
    na = fe->getAB(0)->n->getN();
  else
    na = me->getAB(0)->n->getN().times(type3->theta);

  Event *bb = getEvent(1, 1);
  Point2 nb;
  if (betweenEvents(bb, type3, ba))
    nb = fe->getAB(1)->n->getN();
  else
    nb = me->getAB(1)->n->getN().times(type3->theta);

  return between(na, n, nb);
}

Point2 CSpace::SumEE::type3theta (Event *e, SumEE *see012[3]) {
  for (int i = 0; i < 3; i++) {
    SumEE *that = see012[i];
    if (this == that)
      continue;

    if (!(this->fe->h == that->fe->t && this->fe->h == e->a &&
          that->me->h == this->me->t && that->me->h == e->b ||
          that->fe->h == this->fe->t && that->fe->h == e->a &&
          this->me->h == that->me->t && this->me->h == e->b))
      continue;

    assert(this->isConcave() || that->isConcave());

    Point2 fc = this->fe->c->getC() - that->fe->c->getC();
    Point2 mc = this->isConcave() != that->isConcave() ?
      this->me->c->getC() - that->me->c->getC() :
      that->me->c->getC() - this->me->c->getC();
      
    Point2 fm = fc.divides(mc);
    return fm.unit();
  }

  return e->theta;
}

#ifdef BLEEN
bool CSpace::SumEE::contains (Point2 theta, Point2 n, SumEE *see012[3]) {
  static int count;
  if (count++ == 264)
    cout << "contains this is it" << endl;

  Event *ab = getEvent(0, 1);
  Event *ba = getEvent(1, 0);
  if (!between(ab->theta, theta, ba->theta))
    return false;

  Event *aa = getEvent(0, 0);
  Point2 aatheta = type3theta(aa, see012);
  Point2 na;
  if (between(ab->theta, theta, aatheta))
    na = fe->getAB(0)->n;
  else
    na = me->getAB(0)->n.times(theta);

  Event *bb = getEvent(1, 1);
  Point2 bbtheta = type3theta(bb, see012);
  Point2 nb;
  if (between(bbtheta, theta, ba->theta))
    nb = fe->getAB(1)->n;
  else
    nb = me->getAB(1)->n.times(theta);

  bool result = between(na, n, nb);
  // assert(result == contains(theta, n));
  return result;
}
#endif

bool CSpace::SumEE::isLive () {
  Event *ab = getEvent(0, 1);
  Event *ba = getEvent(1, 0);
  return between(ab->theta, ba->theta);
}

bool CSpace::SumEE::contains (Point2 n) {
  Event *ab = getEvent(0, 1);
  Event *ba = getEvent(1, 0);
  if (!between(ab->theta, ba->theta))
    return false;

  Event *aa = getEvent(0, 0);
  Point2 na;
  if (between(ab->theta, aa->theta))
    na = fe->getAB(0)->n->getN();
  else
    na = me->getAB(0)->n->getN();

  Event *bb = getEvent(1, 1);
  Point2 nb;
  if (between(bb->theta, ba->theta))
    nb = fe->getAB(1)->n->getN();
  else
    nb = me->getAB(1)->n->getN();

  return between(na, n, nb);
}

#define FAST_TYPE2
int nNearOrFar, nTooFar;
double mpTime;

void CSpace::createOuterBumps (SumEE *see1, SumEE *see2, int iStart) {
  int i;
  try {
  if (see1->fe == see2->fe || see1->me == see2->me)
    return;

  /*
  if ((see1->fe->index() == (see2->fe->index()+1)%fes.size() ||
       see2->fe->index() == (see1->fe->index()+1)%fes.size()) &&
      (see1->me->index() == (see2->me->index()+1)%mes.size() ||
       see2->me->index() == (see1->me->index()+1)%mes.size()))
    return;      
  */

  if ((see1->fe->t == see2->fe->h ||
       see2->fe->t == see1->fe->h) &&
      (see1->me->t == see2->me->h ||
       see2->me->t == see1->me->h))
    return;      

  nNearOrFar++;
  bool tooFar = false;

  if (lesserZ(evdfm[0].getEEmax(see1->fe->index(), see2->fe->index()),
              evdfm[1].getEEmin(see1->me->index(), see2->me->index())) ||
      greaterZ(evdfm[0].getEEmin(see1->fe->index(), see2->fe->index()),
               evdfm[1].getEEmax(see1->me->index(), see2->me->index()))) {
    tooFar = true;
    nTooFar++;
#ifdef FAST_TYPE2
    return;
#endif
  }


  // |(fc2 + mc2 t) - (fc1 + mc1 t)| = r1 + r2
  // |(fc2 - fc1) + (mc2 - mc1) t| = r1 + r2
  // |fc + mc t| = r
  Point2 fc = see2->fe->c->getC() - see1->fe->c->getC();
  Point2 mc = see2->me->c->getC() - see1->me->c->getC();
  Parameter mcnorm = mc.norm();
  Parameter r = see1->r + see2->r;
  Point2 q[2], cq[2];
  if (!intersectCircles(r, fc, mcnorm, q, cq))
    return;

  Point2 mcu = mc / mcnorm;
  for (i = iStart; i < 2; i++) {
    Point2 n = q[i].unit();
    Point2 nneg = -n;
    Point2 theta = cq[i].unit().divides(mcu);

    Point2 n1 = see1->isConcave() ? nneg : n;
    Point2 n2 = see2->isConcave() ? n : nneg;

    if (!see1->contains(theta, n1))
      continue;
    if (!see2->contains(theta, n2))
      continue;

    // Point2 p = linearCombination(see1->getC(theta), see1->s, n1);
    Point2 p = see1->getC(theta) + n1 * see1->s;
    Event *event = new Event(theta, p, see1, see2, n1, n2, i);
    events.insert(event);
    assert(!tooFar);
  }
  } catch (SignException se) {
    mpTime -= getTime();
    see1->increasePrecision();
    see2->increasePrecision();
    createOuterBumps(see1, see2, i);
    see1->decreasePrecision();
    see2->decreasePrecision();
    mpTime += getTime();
  }
}

void CSpace::createInnerBumps (SumEE *see1, SumEE *see2, int iStart) {
  int i = 0;
  try {
  // if ((see1->index() == 165 || see1->index() == 188) &&
  // (see2->index() == 165 || see2->index() == 188))
  // cout << "this is it bump" << endl;
  if (see1->fe == see2->fe || see1->me == see2->me)
    return;

  /*
  if ((see1->fe->index() == (see2->fe->index()+1)%fes.size() ||
       see2->fe->index() == (see1->fe->index()+1)%fes.size()) &&
      (see1->me->index() == (see2->me->index()+1)%mes.size() ||
       see2->me->index() == (see1->me->index()+1)%mes.size()))
    return;
  */

  if ((see1->fe->t == see2->fe->h ||
       see2->fe->t == see1->fe->h) &&
      (see1->me->t == see2->me->h ||
       see2->me->t == see1->me->h))
    return;

  nNearOrFar++;
  bool tooFar = false;
#ifdef EVDIST
  if (lesserZ(evdfm[0].getEEmax(see1->fe->index(), see2->fe->index()),
              evdfm[1].getEEmin(see1->me->index(), see2->me->index())) ||
      greaterZ(evdfm[0].getEEmin(see1->fe->index(), see2->fe->index()),
               evdfm[1].getEEmax(see1->me->index(), see2->me->index()))) {
    tooFar = true;
    nTooFar++;
#ifdef FAST_TYPE2
    return;
#endif
  }
#endif

  if (see1->r.value() < see2->r.value()) {
    SumEE *temp = see1;
    see1 = see2;
    see2 = temp;
  }
  // |(fc2 + mc2 t) - (fc1 + mc1 t)| = r1 - r2
  // |(fc2 - fc1) + (mc2 - mc1) t| = r1 - r2
  // |fc + mc t| = r
  Point2 fc = see2->fe->c->getC() - see1->fe->c->getC();
  Point2 mc = see2->me->c->getC() - see1->me->c->getC();
  Parameter mcnorm = mc.norm();
  Parameter r = see1->r - see2->r;
  Point2 q[2], cq[2];
  if (!intersectCircles(r, fc, mcnorm, q, cq))
    return;

  Point2 mcu = mc / mcnorm;
  for (i = iStart; i < 2; i++) {
    Point2 n = q[i].unit();
    Point2 nneg = -n;
    Point2 theta = cq[i].unit().divides(mcu);

    Point2 n1 = see1->isConcave() ? nneg : n;
    Point2 n2 = see2->isConcave() ? nneg : n;

    if (!see1->contains(theta, n1))
      continue;
    if (!see2->contains(theta, n2))
      continue;

    // Point2 p = linearCombination(see1->getC(theta), see1->s, n1);
    Point2 p = see1->getC(theta) + n1 * see1->s;
    Event *event = new Event(theta, p, see1, see2, n1, n2, i+2);
    events.insert(event);
    assert(!tooFar);
  }
  } catch (SignException se) {
    mpTime -= getTime();
    see1->increasePrecision();
    see2->increasePrecision();
    createInnerBumps(see1, see2, i);
    see1->decreasePrecision();
    see2->decreasePrecision();
    mpTime += getTime();
  }
}

void CSpace::createStabs (SumEE *see, SumVE *sve, int iStart) {
  int i;
  try {
  // if (see->index() == 188 && sve->headOf != 0 && sve->headOf->index() == 165)
  // cout << "this is it createStabs1" << endl;

  if (see == sve->tailOf || see == sve->headOf)
    return;

#ifdef BLEEN
  if ((sve->fv->index() == see->fe->index() ||
       sve->fv->index() == (see->fe->index()+1)%fes.size()) &&
      (sve->me->index() == (see->me->index()-1+mes.size())%mes.size() ||
       sve->me->index() == (see->me->index()+1)%mes.size()))
    return;
#endif
  assert(sve->me->t != 0);
  assert(sve->me->h != 0);

  if ((sve->fv == see->fe->t ||
       sve->fv == see->fe->h) &&
      (sve->me->t != 0 && sve->me->t == see->me->h ||
       sve->me->h != 0 && sve->me->h == see->me->t))
    return;

  if (see->me->index() == sve->me->index())
    return;

  nNearOrFar++;
  bool tooFar = false;
#ifdef EVDIST
  if (lesserZ(evdfm[0].getVEmax(sve->fv->index(), see->fe->index()),
              evdfm[1].getEEmin(see->me->index(), sve->me->index())) ||
      greaterZ(evdfm[0].getVEmin(sve->fv->index(), see->fe->index()),
               evdfm[1].getEEmax(see->me->index(), sve->me->index()))) {
    tooFar = true;
    nTooFar++;
#ifdef FAST_TYPE2
    return;
#endif
  }
#endif

  // c1 = fc1 + mc1 t
  // v2 = p2 + c2 t + n2 s2
  // |(p2 + c2 t + n2 s2) - (fc1 + mc1 t)| = r1
  // |(p2 + n2 s2 - fc1) + (c2 - mc1) t| = r1
  // |fc + mc t| = r
  /*
  Point2 p1 = sve->fv->n->getP(), p2 = sve->fv->n->getN(), p3 = see->fe->c->getC();
  Parameter p4 = sve->me->c->getS();
  */
  Point2 fc = sve->fv->n->getP() + sve->fv->n->getN() * sve->me->c->getS() - see->fe->c->getC();
  Point2 mc = sve->me->c->getC() - see->me->c->getC();
  Parameter mcnorm = mc.norm();
  Parameter r = see->r;
  Point2 q[2], cq[2];
  if (!intersectCircles(r, fc, mcnorm, q, cq))
    return;

  Point2 mcu = mc / mcnorm;
  for (i = iStart; i < 2; i++) {
    Point2 theta = cq[i].unit().divides(mcu);

    if (!sve->isLive(theta))
      continue;

    Point2 n = q[i].unit();
    Point2 nneg = -n;

    Point2 n1 = see->isConcave() ? nneg : n;
    Point2 n2 = sve->fv->n->getN();

    if (!see->contains(theta, n1))
      continue;

    // Point2 p = linearCombination(see->getC(theta), see->s, n1);
    Point2 p = see->getC(theta) + n1 * see->s;
    Event *event = new Event(theta, p, see, sve, n1, n2, i);
    events.insert(event);
    assert(!tooFar);
  }
  } catch (SignException se) {
    mpTime -= getTime();
    see->increasePrecision();
    sve->increasePrecision();
    createStabs(see, sve, i);
    see->decreasePrecision();
    sve->decreasePrecision();
    mpTime += getTime();
  }
}

void CSpace::createStabs (SumEE *see, SumEV *sev, int iStart) {
  int i;
  try {
  // if (see->index() == 188 && sev->headOf != 0 && sev->headOf->index() == 165)
  // cout << "this is it createStabs1" << endl;

  if (see == sev->tailOf || see == sev->headOf)
    return;

#ifdef BLEEN
  if ((sev->fe->index() == (see->fe->index()-1+fes.size())%fes.size() ||
       sev->fe->index() == (see->fe->index()+1)%fes.size()) &&
      (sev->mv->index() == see->me->index() ||
       sev->mv->index() == (see->me->index()+1)%mes.size()))
    return;
#endif
  if ((sev->fe->t != 0 && sev->fe->t == see->fe->h ||
       sev->fe->h != 0 && sev->fe->h == see->fe->t) &&
      (sev->mv == see->me->t ||
       sev->mv == see->me->h))
    return;
  
  if (see->fe->index() == sev->fe->index())
    return;

  nNearOrFar++;
  bool tooFar = false;
#ifdef EVDIST
  if (lesserZ(evdfm[0].getEEmax(see->fe->index(), sev->fe->index()),
              evdfm[1].getVEmin(sev->mv->index(), see->me->index())) ||
      greaterZ(evdfm[0].getEEmin(see->fe->index(), sev->fe->index()),
               evdfm[1].getVEmax(sev->mv->index(), see->me->index()))) {
    tooFar = true;
    nTooFar++;
#ifdef FAST_TYPE2
    return;
#endif
  }
#endif

  // c1 = fc1 + mc1 t
  // v2 = c2 + n2 t s2 + p2 t
  // |(c2 + n2 t s2 + p2 t) - (fc1 + mc1 t)| = r1
  // |(c2 - fc1) + (n2 s + p2 - mc1) t| = r1
  // |fc + mc t| = r
  Point2 fc = sev->fe->c->getC() - see->fe->c->getC();
  Point2 mc = sev->mv->n->getN() * sev->fe->c->getS() + sev->mv->n->getP() - see->me->c->getC();
  Parameter mcnorm = mc.norm();
  Parameter r = see->r;
  Point2 q[2], cq[2];
  if (!intersectCircles(r, fc, mcnorm, q, cq))
    return;

  Point2 mcu = mc / mcnorm;
  for (i = iStart; i < 2; i++) {
    Point2 theta = cq[i].unit().divides(mcu);

    if (!sev->isLive(theta))
      continue;

    Point2 n = q[i].unit();
    Point2 nneg = -n;

    Point2 n1 = see->isConcave() ? nneg : n;
    Point2 n2 = sev->mv->n->getN().times(theta);

    if (!see->contains(theta, n1))
      continue;

    // Point2 p = linearCombination(see->getC(theta), see->s, n1);
    Point2 p = see->getC(theta) + n1 * see->s;

    n1.decreasePrecision();
    n2.decreasePrecision();
    theta.decreasePrecision();
    p.decreasePrecision();
    see->decreasePrecision();
    sev->decreasePrecision();

    Event *event = new Event(theta, p, see, sev, n1, n2, i);
    events.insert(event);
    assert(!tooFar);
  }
  } catch (SignException se) {
    mpTime -= getTime();
    see->increasePrecision();
    sev->increasePrecision();
    createStabs(see, sev, i);
    see->decreasePrecision();
    sev->decreasePrecision();
    mpTime += getTime();
  }
}

#ifdef BLEEN
int type3 (Point2 f[3], Point2 m[3], Parameter r[3], PV *q, PV *p, int i) {
#ifdef DUMP_TYPE3
  cout << setprecision(23);
  cout << "type3:";
  for (int i = 0; i < 3; i++)
     cout << " " << m[i].x.value() << " " << m[i].y.value();
  cout << endl;
  cout << "type3:";
  for (int i = 0; i < 3; i++)
    cout << " " << f[i].x.value() << " " << f[i].y.value();
  cout << endl;
  cout << "type3:";
  for (int i = 0; i < 3; i++)
    cout << " " << r[i].value();
  cout << endl;
#endif
  double x[128];
  int nqp = criticality3(m, f, r, x, i);
  for (int i = 0; i < nqp; i++) {
    double *pxyqyx = x + 4*i;
    p[i].x = pxyqyx[0];
    p[i].y = pxyqyx[1];
    q[i].x = pxyqyx[3];
    q[i].y = pxyqyx[2];
  }
  return nqp;
}
#endif

Point2 inverse (const Point2 &p) {
  return Point2(p.x, -p.y) / (p.x * p.x + p.y * p.y);
}

#ifdef BLEEN
int type3dual (Point2 f[3], Point2 m[3], Parameter r[3],
               PV *q, PV *p, int iVal) {
  // return type3(f, m, r, q, p, iVal);
  int i0 = 0;
  for (int i = 1; i < 3; i++)
    if (r[i0] < r[i])
      i0 = i;
  int i1 = (i0+1)%3;
  int i2 = (i1+1)%3;

  Parameter rb[3] = { r[i0], r[i1], r[i2] };
  Point2 fb[3] =
    { f[i0] - f[i0], f[i1] - f[i0], f[i2] - f[i0]};
  Point2 mb[3] =
    { m[i0] - m[i0], m[i1] - m[i0], m[i2] - m[i0]};

  Parameter rc[3] =
    { rb[0] / rb[0], rb[1] / rb[0], rb[2] / rb[0] };
  Point2 fc[3] =
    { fb[0] / rb[0], fb[1] / rb[0], fb[2] / rb[0] };
  Point2 mc[3] =
    { mb[0] / rb[0], mb[1] / rb[0], mb[2] / rb[0] };

  Parameter rd[3] = { rc[0], rc[1] / mc[1].norm(), rc[2] / mc[2].norm() };
  Point2 md[3] = { mc[0], inverse(mc[1]), inverse(mc[2]) };
  Point2 fd[3] = { fc[0], -fc[1].times(md[1]), -fc[2].times(md[2]) };

  double x[128];
  PV pSave[6];
  int nqp = criticality3(md, fd, rd, x, iVal);
  static double minPP = 1e9;
  for (int i = 0; i < nqp; i++) {
    double *pxyqyx = x + 4*i;
    q[i].x = pxyqyx[0];
    q[i].y = pxyqyx[1];
    p[i].x = pxyqyx[3];
    p[i].y = pxyqyx[2];
    pSave[i] = p[i];

    for (int j = 0; j < i; j++) {
      double dist = pSave[i].distance(pSave[j]);
      if (dist < minPP) {
        cout << "minPP2 " << dist << endl;
        minPP = dist;
      }
    }

    p[i] = p[i] * rb[0].value(iVal);
    p[i] = p[i] + P2PV(f[i0], iVal) + P2PV(m[i0], iVal).times(q[i]);

#ifdef BLEEN
    for (int j = 0; j < 3; j++) {
      PV cp = p[i] - (P2PV(f[j]) + q[i].times(P2PV(m[j])));
      // PV cp = q[i] - (P2PV(fd[j], iVal) + p[i].times(P2PV(md[j], iVal)));
      // PV cp = p[i] - (P2PV(fc[j], iVal) + q[i].times(P2PV(mc[j], iVal)));
      // PV cp = p[i] - (P2PV(fb[j], iVal) + q[i].times(P2PV(mb[j], iVal)));
      double cp2r2 = cp * cp - r[j].value(iVal) * r[j].value(iVal);
      // double cp2r2 = cp * cp - rd[j].value(iVal) * rd[j].value(iVal);
      // double cp2r2 = cp * cp - rc[j].value(iVal) * rc[j].value(iVal);
      // double cp2r2 = cp * cp - rb[j].value(iVal) * rb[j].value(iVal);
      
      cout << "cp2r2 " << cp2r2
           << " " << p[i] * p[i]
           << " " << q[i] * q[i] << endl;
    }
#endif
  }
  return nqp;
}
#endif

// Two circles have the same fixed center.
int type3sameFE (Point2 f[3], Point2 m[3], Parameter r[3], int iSameFE,
                 Point2 *q, Point2 *p) {
  int i = iSameFE;
  int j = (i+1)%3;
  int k = (j+1)%3;
  // f[j] == f[k]
  
  // Calculate the intersections of mj,rj and mk,rk
  Point2 mm = m[k] - m[j];
  Point2 op[2], cp[2];
  if (!intersectCircles(r[j], mm, r[k], op, cp))
    return 0;
  Point2 pjk[2] = { m[j] + op[0], m[j] + op[1] };

  int npq = 0;
  for (int h = 0; h < 2; h++) {
    // f[jk] + q pjk[h] intersects f[i] + q m[i], r[i]
    // |(f[jk] - f[i]) + q (pjk[h] - m[i])| = r[i]
    Point2 ff = f[j] - f[i];
    Point2 mp = pjk[h] - m[i];
    if (!intersectCircles(r[i], ff, mp.norm(), op, cp))
      continue;
    for (int g = 0; g < 2; g++) {
      q[npq] = cp[g].unit().divides(mp.unit());
      p[npq] = f[j] + q[npq].times(pjk[h]);
      npq++;
    }
  }
  return npq;
}

// Two circles have the same moving center.
int type3sameME (Point2 f[3], Point2 m[3], Parameter r[3], int iSameME,
                 Point2 *q, Point2 *p) {
  int i2 = iSameME;
  int i0 = (i2+1)%3;
  int i1 = (i0+1)%3;

  // Use substitution p = p - (f0 + q m0)

  // p^2 = r0^2
  // (p - f01)^2 = r1^2
  // (p - (f02 + q m02))^2 = r2^2
  // ((f02 - p) + q m02))^2 = r2^2
  // q^2 = 1
  
  Point2 f01 = f[i1] - f[i0];
  Point2 op[2], cp[2];
  if (!intersectCircles(r[i0], f01, r[i1], op, cp))
    return 0;

  int npq = 0;

  Point2 m02 = m[i2] - m[i0];
  Parameter m02norm = m02.norm();
  Point2 m02inv = inverse(m02);
  for (int i = 0; i < 2; i++) {
    Point2 pf02 = f[i2] - f[i0] - op[i];
    Point2 op2[2], cp2[2];
    if (!intersectCircles(r[i2], pf02, m02norm, op2, cp2))
      continue;
    for (int j = 0; j < 2; j++) {
      q[npq] = cp2[j].times(m02inv);
      p[npq] = op[i] + f[i0] + q[npq].times(m[i0]);
      npq++;
    }
  }

  return npq;
}

double createType3EventsTime;

bool CSpace::testType3 (Type3 *type3, Intn *intns[3]) {
  try {
    Point2 p = type3->p;
    Point2 q = type3->theta;

    Intn iintns[] = { Intn(type3->sees[0], type3->sees[1]),
                      Intn(type3->sees[1], type3->sees[2]),
                      Intn(type3->sees[2], type3->sees[0]) };
    Intn *pintns[] = { &iintns[0], &iintns[1], &iintns[2] };
    Event eventTemp(q, p, pintns, type3);

    Point2 n[3];
    for (int i = 0; i < 3; i++) {
      Point2 c = type3->sees[i]->getC(q);
      n[i] = (p - c) / type3->sees[i]->s;
#ifndef NXNPROXY
      if (!type3->sees[i]->contains(q, n[i]))
        return false;
#else
      if (!type3->sees[i]->contains(&eventTemp, n[i]))
        return false;
#endif
    }

    PV p2(P2PV(p)), q2(P2PV(q));
    double minDist = -1;
    for (int i = 0; i < 3; i++) {
      int j = (i+1)%3;
      int k = (i+2)%3;
#ifndef NXNPROXY
      int cpSign = n[j].crossproductSign(n[k]);
#else
      int cpSign = type3->sees[j]->crossproductSign(type3->sees[k], p, q);
      Parameter::signFail = false;
      if (cpSign != n[j].crossproductSign(n[k]))
        cout << "bad Intn sign" << endl;      
      Parameter::signFail = true;
#endif
      if (cpSign > 0)
        intns[i] = new Intn(type3->sees[j], type3->sees[k]);
      else
        intns[i] = new Intn(type3->sees[k], type3->sees[j]);
      PV intn = intns[i]->getP(q2);
      double dist = p2.distance(intn);
      if (i == 0 || dist < minDist)
        minDist = dist;
    }
     
    static double maxPIdist;
    if (minDist > maxPIdist) {
        cout << "maxPIdist " << minDist << endl;
        maxPIdist = minDist;
    }
    return true;
  } catch (SignException se) {
    type3->increasePrecision();
    bool ret = testType3(type3, intns);
    type3->decreasePrecision();
    return ret;
  }
}

void CSpace::createType3Events (SumEE *see012[3]) {
  for (int i = 0; i < 3; i++)
    for (int j = i+1; j < 3; j++)
      if (see012[j]->lessThan(see012[i])) {
        SumEE *temp = see012[i];
        see012[i] = see012[j];
        see012[j] = temp;
      }
  assert(see012[0]->lessThan(see012[1]));
  assert(see012[1]->lessThan(see012[2]));

  if (see012[0]->index() == 197 &&
      see012[1]->index() == 426 &&
      see012[2]->index() == 793) {
    cout << "this is it createType3Events" << endl;
    // double c = 0.9046797991960086; // 0.90467979919428831;
    // testT = sqrt((1 - c) / (1 + c));
    // testT = -4.3438409515042231;
  }

  if (testedTriples.find(see012) != testedTriples.end())
    return;
  SumEE **triple = new SumEE*[3];
  for (int i = 0; i < 3; i++)
    triple[i] = see012[i];
  testedTriples.insert(triple);

  static int count;
  if (++count == 57)
    cout << "type3s this is it" << endl;
  vector<Type3 *> type3s = Type3::construct(see012);
  // for (int i = 0; i < 10; i++) {
  /*
  vector<Type3 *> type3sB = Type3::construct(see012);
  if (type3s.size() < type3sB.size())
    type3s = type3sB;
  */
  //}}
  // assert(type3s.size() == type3sB.size());

  if (see012[0]->index() == -1 /*17* /*21*/ /*53*/ /*81*/ &&
      see012[1]->index() == 423 /*199*/ /*70*/ /*72*/ /*100*/ &&
      see012[2]->index() == 790 /*794* /*89*/ /*81*/ /*137*/)
    while (type3s.size() == 0) {
      cout << "try again" << endl;
      type3s = Type3::construct(see012);
    }

  if (see012[0]->index() == -1 &&
      see012[1]->index() == 100 &&
      see012[2]->index() == 137) {
    PV q3 = P2PV(type3s[3]->theta);
    PV int81_137 = dynamic_cast<Intn*>(see012[0]->getSubs()[1]->h)->getP(q3);
    PV head = see012[0]->getSubs()[2]->h->getP(q3);
    PV c3 = P2PV(see012[0]->fe->c->getC()) + P2PV(see012[0]->me->c->getC()).times(q3);
    PV intN = (int81_137 - c3).unit();
    PV headN = (head - c3).unit();
    PV mehn = P2PV(see012[0]->me->h->n->getN());
    PV mehnq3 = mehn.times(q3);
    cout << int81_137.x << " " << int81_137.y << endl;
    cout << head.x << " " << head.y << endl;
  }

  for (int h = 0; h < type3s.size(); h++) {
    Intn *intns[3];
    if (!testType3(type3s[h], intns))
      continue;
    Event *event = new Event(type3s[h]->theta, type3s[h]->p, intns, type3s[h]);
    events.insert(event);
  }
}

void CSpace::makeArrangement0 () {
  vector<SumEE *> liveSEEs;
  for (int i = 0; i < sees.size(); i++) {
    if (sees[i] != 0 && sees[i]->isLive())
      liveSEEs.push_back(sees[i]);
  }

  vector< vector<INN> > innss(liveSEEs.size());

  INN inn[2];
  for (int i = 0; i < liveSEEs.size(); i++)
    for (int j = i+1; j < liveSEEs.size(); j++) {
      // if (liveSEEs[i] == sees[24] && liveSEEs[j] == sees[49])
      // cout << "this is it" << endl;
      int n = liveSEEs[i]->intersect(liveSEEs[j], inn);
      int indI = liveSEEs[i]->index();
      int indJ = liveSEEs[j]->index();
      for (int k = 0; k < n; k++) {
        inn[k].insert(liveSEEs[i], innss[i]);
        inn[k].insert(liveSEEs[j], innss[j]);
      }
    }

  // make subedges
  for (int i = 0; i < liveSEEs.size(); i++) {
    SumEE *see = liveSEEs[i];
    vector<SubE *> &subs = see->getSubs();
    vector<INN> &inns = innss[i];

    subs.push_back(new SubE(see));
    subs[0]->setT(see->getTH(0));
    for (int j = 0; j < inns.size(); j++) {
      subs.back()->setH(inns[j].intn);
      subs.push_back(new SubE(see));
      subs.back()->setT(inns[j].intn);
    }
    subs.back()->setH(see->getTH(1));

    see->subs0 = subs;
  }

  vector<SubE *> oldSubs;
  updateSubs(oldSubs, allSubs, 0);

  /*
  for (int i = 0; i < liveSEEs.size(); i++) {
    vector<SubE *> &subs = liveSEEs[i]->subs;
    for (int i = 0; i < subs.size(); i++)
      checkTriangles(subs[i]);
  }
  */
}

CSpace::SumV *CSpace::SumEE::getTH (int th) {
  int ab = isConcave() ? 1 - th : th;
  CSpace *csp = fe->csp;
  SumVE *sve = csp->sves[csp->sumInd(fe->getAB(ab)->index(), me->index())];
  assert(this == (th ? sve->headOf : sve->tailOf));
  if (sve->isLive())
    return sve;
  SumEV *sev = csp->sevs[csp->sumInd(fe->index(), me->getAB(ab)->index())];
  assert(this == (th ? sev->headOf : sev->tailOf));
  return sev;
}

void CSpace::INN::insert(SumEE *see, vector<INN> &inns) {
  Point2 n = getN(see);
  int a = 0;
  int b = inns.size() - 1;
  while (a <= b) {
    int c = (a + b)/2;
    if (((n.crossproductSign(inns[c].getN(see)) > 0) ^ see->isConcave()) == 1)
      b = c - 1;
    else
      a = c + 1;
  }
  inns.insert(inns.begin() + a, *this);
}

int CSpace::SumEE::intersect(SumEE *that, INN inn[2]) {
  // if (this->index() == 8 && that->index() == 211 ||
  // this->index() == 211 && that->index() == 8)
  // cout << "intersect this is it" << endl;

  // if (this->fe == that->fe || this->me == that->me)
  // return 0;
  int fes_size = fe->csp->fes.size();
  int mes_size = me->csp->mes.size();

#ifdef BLEEN
  if (this->fe == that->fe &&
      (this->me->index() == (that->me->index() + 1)%mes_size ||
       that->me->index() == (this->me->index() + 1)%mes_size) ||
      this->me == that->me &&
      (this->fe->index() == (that->fe->index() + 1)%fes_size ||
       that->fe->index() == (this->fe->index() + 1)%fes_size))
    return 0;
#endif
  if (this->fe == that->fe &&
      (this->me->t != 0 && this->me->t == that->me->h ||
       that->me->t != 0 && that->me->t == this->me->h) ||
      this->me == that->me &&
      (this->fe->t != 0 && this->fe->t == that->fe->h ||
       that->fe->t != 0 && that->fe->t == this->fe->h))
    return 0;

  Point2 thisC = this->getC();
  Point2 thatC = that->getC();
  Point2 c = thatC - thisC;
  Point2 q[2], cq[2];
  static int count;
  if (++count == -1)
    cout << "intersectCircles this is it" << endl;
  if (!intersectCircles(this->r, c, that->r, q, cq))
    return 0;

  int n = 0;
  for (int i = 0; i < 2; i++) {
    Point2 n1 = q[i] / this->s;
    if (!this->contains(n1))
      continue;

    Point2 n2 = cq[i] / that->s;
    if (!that->contains(n2))
      continue;

    if ((i ^ this->isConcave() ^ that->isConcave()) == 0)
      inn[n++] = INN(new Intn(this, that), n1, n2);
    else
      inn[n++] = INN(new Intn(that, this), n2, n1);
  }

  return n;
}

void drawCircle (Point2 &c, Parameter &r, int d) {
  int x = int(rint(1200 * c.x.value()));
  int y = int(rint(1200 * -c.y.value()));
  int z = int(rint(1200 * r.value()));
  cout << "1 2 0 1 0 7 " << d << " -1 -1 0.000 1 0.0000 "
       << x << " " << y << " "
       << z << " " << z << " "
       << x << " " << y << " "
       << z+x << " " << y << endl;
}

// Intersect the three chord lines of the three pairs of circles.
Point2 calcP (Point2 c[3], Parameter r[3]) {
  // (p - c0)^2 = r0^2
  // (p - c1)^2 = r1^2
  // (p - c2)^2 = r2^2
  
  // -2 (c1 - c0) * p + c1^2 - c0^2 = r1^2 - r0^2
  // -2 (c2 - c0) * p + c2^2 - c0^2 = r2^2 - r0^2
  
  // (c1 - c0) * p = (-c0^2 + r0^2 + c1^2 - r1^2) / 2
  // (c2 - c0) * p = (-c0^2 + r0^2 + c2^2 - r2^2) / 2
  
  // a1 * p = b1
  // a2 * p = b2
  
  Point2 a1 = c[1] - c[0];
  Parameter b1 = (-c[0].innerproduct(c[0]) + r[0] * r[0] +
                  c[1].innerproduct(c[1]) - r[1] * r[1]) / 2;
  Point2 a2 = c[2] - c[0];
  Parameter b2 = (-c[0].innerproduct(c[0]) + r[0] * r[0] +
                  c[2].innerproduct(c[2]) - r[2] * r[2]) / 2;
  
  Point2 ab1(a1.x, b1), ba1(b1, a1.y);
  Point2 ab2(a2.x, b2), ba2(b2, a2.y);
  Parameter a1Xa2 = a1.crossproduct(a2);
  Parameter ab1Xab2 = ab1.crossproduct(ab2);
  Parameter ba1Xba2 = ba1.crossproduct(ba2);
  
  Point2 p(ba1Xba2 / a1Xa2, ab1Xab2 / a1Xa2);
  return p;
}

vector<CSpace::Type3 *>
CSpace::Type3::constructSameFE (SumEE *sees[3], int iSameFE) {
  Point2 f[3] = { sees[0]->fe->c->getC(),
                  sees[1]->fe->c->getC(),
                  sees[2]->fe->c->getC() };
  Point2 m[3] = { sees[0]->me->c->getC(),
                  sees[1]->me->c->getC(),
                  sees[2]->me->c->getC() };
  Parameter r[3] = { sees[0]->r, sees[1]->r, sees[2]->r };

  int i = iSameFE;
  int j = (i+1)%3;
  int k = (j+1)%3;
  // f[j] == f[k]
  
  // Calculate the intersections of mj,rj and mk,rk
  vector<Type3 *> out;
  Point2 mm = m[k] - m[j];
  Point2 op[2], cp[2];
  if (!intersectCircles(r[j], mm, r[k], op, cp))
    return out;
  Point2 pjk[2] = { m[j] + op[0], m[j] + op[1] };

  for (int h = 0; h < 2; h++) {
    // f[jk] + q pjk[h] intersects f[i] + q m[i], r[i]
    // |(f[jk] - f[i]) + q (pjk[h] - m[i])| = r[i]
    Point2 ff = f[j] - f[i];
    Point2 mp = pjk[h] - m[i];
    if (!intersectCircles(r[i], ff, mp.norm(), op, cp))
      continue;
    for (int g = 0; g < 2; g++) {
      Point2 q = cp[g].unit().divides(mp.unit());
      Point2 p = f[j] + q.times(pjk[h]);
      out.push_back(new Type3(sees, p, q, 2*h+g));
    }
  }
  
  return out;
}

vector<CSpace::Type3 *>
CSpace::Type3::constructSameME (SumEE *sees[3], int iSameME) {
  Point2 f[3] = { sees[0]->fe->c->getC(),
                  sees[1]->fe->c->getC(),
                  sees[2]->fe->c->getC() };
  Point2 m[3] = { sees[0]->me->c->getC(),
                  sees[1]->me->c->getC(),
                  sees[2]->me->c->getC() };
  Parameter r[3] = { sees[0]->r, sees[1]->r, sees[2]->r };

  int i2 = iSameME;
  int i0 = (i2+1)%3;
  int i1 = (i0+1)%3;

  // Use substitution p = p - (f0 + q m0)

  // p^2 = r0^2
  // (p - f01)^2 = r1^2
  // (p - (f02 + q m02))^2 = r2^2
  // ((f02 - p) + q m02))^2 = r2^2
  // q^2 = 1
  
  vector<Type3 *> out;
  Point2 f01 = f[i1] - f[i0];
  Point2 op[2], cp[2];
  if (!intersectCircles(r[i0], f01, r[i1], op, cp))
    return out;

  Point2 m02 = m[i2] - m[i0];
  Parameter m02norm = m02.norm();
  Point2 m02inv = inverse(m02);
  for (int i = 0; i < 2; i++) {
    Point2 pf02 = f[i2] - f[i0] - op[i];
    Point2 op2[2], cp2[2];
    if (!intersectCircles(r[i2], pf02, m02norm, op2, cp2))
      continue;
    for (int j = 0; j < 2; j++) {
      Point2 q = cp2[j].times(m02inv);
      Point2 p = op[i] + f[i0] + q.times(m[i0]);
      out.push_back(new Type3(sees, p, q, 2*i+j));
    }
  }

  return out;
}

vector<CSpace::Type3 *> CSpace::Type3::construct (SumEE *sees[3]) {
  static int count;
  if (++count == 327)
    cout << "this is it Type3::construct" << endl;
  int iSameFE = -1, iSameME = -1;
  for (int i = 0; i < 3; i++) {
    int j = (i+1)%3;
    int k = (j+1)%3;
    if (sees[j]->fe == sees[k]->fe)
      iSameFE = i;
    if (sees[j]->me == sees[k]->me)
      iSameME = i;
  }

  if (iSameFE != -1)
    return constructSameFE(sees, iSameFE);

  if (iSameME != -1)
    return constructSameME(sees, iSameME);

  try {
  Point2 f[3] = { sees[0]->fe->c->getC(),
                  sees[1]->fe->c->getC(),
                  sees[2]->fe->c->getC() };
  Point2 m[3] = { sees[0]->me->c->getC(),
                  sees[1]->me->c->getC(),
                  sees[2]->me->c->getC() };
  Parameter r[3] = { sees[0]->r, sees[1]->r, sees[2]->r };

  if (false) {
    Point2 p[6], q[6];
    int n = criticality3(m, f, r, q);
    vector<Type3 *> out;

    for (int i = 0; i < n; i++) {
      Point2 c[3];
      for (int j = 0; j < 3; j++) {
	c[j] = f[j] + m[j].times(q[i]);
	if (count == 0)
	  drawCircle(c[j], r[j], i);
      }
      p[i] = calcP(c, r);
      out.push_back(new Type3(sees, p[i], q[i], i));
    }
    return out;
  }
  
  int i0 = 0;
  for (int i = 1; i < 3; i++)
    if (r[i0].value() < r[i].value())
      i0 = i;
  int i1 = (i0+1)%3;
  int i2 = (i1+1)%3;

  Parameter rb[3] = { r[i0], r[i1], r[i2] };
  Point2 fb[3] =
    { f[i0] - f[i0], f[i1] - f[i0], f[i2] - f[i0]};
  Point2 mb[3] =
    { m[i0] - m[i0], m[i1] - m[i0], m[i2] - m[i0]};

  Parameter rc[3] =
    { rb[0] / rb[0], rb[1] / rb[0], rb[2] / rb[0] };
  Point2 fc[3] =
    { fb[0] / rb[0], fb[1] / rb[0], fb[2] / rb[0] };
  Point2 mc[3] =
    { mb[0] / rb[0], mb[1] / rb[0], mb[2] / rb[0] };

  Parameter rd[3] = { rc[0], rc[1] / mc[1].norm(), rc[2] / mc[2].norm() };
  Point2 md[3] = { mc[0], inverse(mc[1]), inverse(mc[2]) };
  Point2 fd[3] = { fc[0], -fc[1].times(md[1]), -fc[2].times(md[2]) };

  Point2 pd[6], qd[6];
  Point2 p[6], q[6];
  bool flag[6];

#ifdef BLEEN
  cout << setprecision(23);
  cout << "Point2 a[3] = {"
       << "Point2(Parameter(" << md[0].x.l << "," << md[0].x.u.r << "), "
       << "Parameter(" << md[0].y.l << "," << md[0].y.u.r << ")), "
       << "Point2(Parameter(" << md[1].x.l << "," << md[1].x.u.r << "), "
       << "Parameter(" << md[1].y.l << "," << md[1].y.u.r << ")), "
       << "Point2(Parameter(" << md[2].x.l << "," << md[2].x.u.r << "), "
       << "Parameter(" << md[2].y.l << "," << md[2].y.u.r << ")" << ")};"
       << endl;
  cout << "Point2 b[3] = {"
       << "Point2(Parameter(" << fd[0].x.l << "," << fd[0].x.u.r << "), "
       << "Parameter(" << fd[0].y.l << "," << fd[0].y.u.r << ")), "
       << "Point2(Parameter(" << fd[1].x.l << "," << fd[1].x.u.r << "), "
       << "Parameter(" << fd[1].y.l << "," << fd[1].y.u.r << ")), "
       << "Point2(Parameter(" << fd[2].x.l << "," << fd[2].x.u.r << "), "
       << "Parameter(" << fd[2].y.l << "," << fd[2].y.u.r << ")" << ")};"
       << endl;
  cout << "Parameter r[3] = {"
       << "Parameter(" << rd[0].l << "," << rd[0].u.r << "), "
       << "Parameter(" << rd[1].l << "," << rd[1].u.r << "), "
       << "Parameter(" << rd[2].l << "," << rd[2].u.r << ")};" << endl;    
#endif

  if (count == -1)
    for (int i = 0; i < -1; i++)
      cout << md[0].x.intervalWidth() << " "
	   << md[0].y.intervalWidth() << " "
	   << fd[0].x.intervalWidth() << " "
	   << fd[0].y.intervalWidth() << " "
	   << rd[0].intervalWidth() << endl;
    
  int n = criticality3(md, fd, rd, qd);
  for (int i = 0; i < n; i++)
    if (qd[i].x.intervalWidth() > 1e-3 ||
	qd[i].y.intervalWidth() > 1e-3) {
      assert(rd[2].l != 1e20);
      cout << "qd too wide" << endl;
      throw signException;
    }
  if (count == 640 || count == 1970)
    /* n = criticality3(md, fd, rd, qd) */;
  if (md[0].x.l == 1e20)
    for (int i = 0; i < n; i++)
      qd[i].increasePrecision();
  for (int i = 0; i < n; i++)
    flag[i] = true;

#ifdef BLEEN
  double maxInt = 0;
  for (int i = 0; i < 3; i++) {
    if (maxInt < md[i].x.u.r - md[i].x.l)
      maxInt = md[i].x.u.r - md[i].x.l;
    if (maxInt < md[i].y.u.r - md[i].y.l)
      maxInt = md[i].y.u.r - md[i].y.l;
    if (maxInt < fd[i].x.u.r - fd[i].x.l)
      maxInt = fd[i].x.u.r - fd[i].x.l;
    if (maxInt < fd[i].y.u.r - fd[i].y.l)
      maxInt = fd[i].y.u.r - fd[i].y.l;
    if (maxInt < rd[i].u.r - rd[i].l)
      maxInt = rd[i].u.r - rd[i].l;
  }
#endif

  vector<Type3 *> out;

  for (int i = 0; i < n; i++) {
    if (!flag[i]) {
      cout << "bracketing failure " << i << endl;
      Point2 tinyrot(cos(-0.2), sin(-0.2));
      for (int j = 0; j < 3; j++) {
        Point2 cdj = fd[j] + md[j].times(qd[i].times(tinyrot));
        if (!flag[i])
          drawCircle(cdj, rd[j], i);
      }
    }
      
    Point2 cd[3];
    for (int j = 0; j < 3; j++) {
      cd[j] = fd[j] + md[j].times(qd[i]);
      if (count == 1057)
        drawCircle(cd[j], rd[j], i);
    }

    if (!flag[i])
      continue;

    pd[i] = calcP(cd, rd);
    if (pd[i].x.intervalWidth() > 1e-3 ||
	pd[i].y.intervalWidth() > 1e-3) {
      assert(rd[2].l != 1e20);
      cout << "pd too wide" << endl;
      throw signException;
    }

    q[i] = pd[i];

    Point2 rot(cos(0.01), sin(0.01));
    if (f[0].x.l == 1e20)
      rot.increasePrecision();
    for (int j = 0; j < 3; j++) {
      Point2 c = f[j] + m[j].times(q[i].times(rot));
      if (count == 2669)
	drawCircle(c, r[j], i);
    }

    p[i] = qd[i] * rb[0] + f[i0] + m[i0].times(q[i]);

#ifndef BLEEN
    static double maxcp2r2d, maxcp2r2;
    for (int j = 0; j < 3; j++) {
      PV cpd = P2PV(pd[i]) - (P2PV(fd[j]) + P2PV(qd[i]).times(P2PV(md[j])));
      double cp2r2d = cpd * cpd - rd[j].value() * rd[j].value();
      // cout << "cp2r2d " << cp2r2d << endl;
      if (cp2r2d > maxcp2r2d) {
	cout << "maxcp2r2d " << cp2r2d << endl;
	maxcp2r2d = cp2r2d;
      }
      PV cp = P2PV(p[i]) - (P2PV(f[j]) + P2PV(q[i]).times(P2PV(m[j])));
      double cp2r2 = cp * cp - r[j].value() * r[j].value();
      // cout << "cp2r2 " << cp2r2 << endl;
      if (cp2r2 > maxcp2r2) {
	cout << "maxcp2r2 " << cp2r2 << endl;
	maxcp2r2 = cp2r2;
      }
    }
#endif

    out.push_back(new Type3(sees, p[i], q[i], qd[i], i));
  }
  return out;
  } catch (SignException signException) {
    for (int i = 0; i < 3; i++) {
      /*
	cout << sees[i]->s.intervalWidth() << " "
	   << sees[i]->fe->c->getC().x.intervalWidth() << " "
	   << sees[i]->fe->c->getC().y.intervalWidth() << " "
	   << sees[i]->me->c->getC().x.intervalWidth() << " "
	   << sees[i]->me->c->getC().y.intervalWidth() << endl;
      */
      sees[i]->increasePrecision();
      /*
      cout << sees[i]->s.intervalWidth() << " "
	   << sees[i]->fe->c->getC().x.intervalWidth() << " "
	   << sees[i]->fe->c->getC().y.intervalWidth() << " "
	   << sees[i]->me->c->getC().x.intervalWidth() << " "
	   << sees[i]->me->c->getC().y.intervalWidth() << endl;
      */
    }
    cout << "hi precision construct" << endl;
    vector<Type3 *> out = construct(sees);
    for (int i = 0; i < 3; i++)
      sees[i]->decreasePrecision();
    for (int i = 0; i < out.size(); i++) {
      out[i]->p.decreasePrecision();
      out[i]->theta.decreasePrecision();
    }
    return out;
  }
}    

void CSpace::Type3::increasePrecision () {
  if (theta.x.l == 1e20)
    return;

  for (int i = 0; i < 3; i++)
    sees[i]->increasePrecision();

  Point2 f[3] = { sees[0]->fe->c->getC(),
                  sees[1]->fe->c->getC(),
                  sees[2]->fe->c->getC() };
  Point2 m[3] = { sees[0]->me->c->getC(),
                  sees[1]->me->c->getC(),
                  sees[2]->me->c->getC() };
  Parameter r[3] = { sees[0]->r, sees[1]->r, sees[2]->r };

  int iSameFE = -1, iSameME = -1;
  for (int i = 0; i < 3; i++) {
    int j = (i+1)%3;
    int k = (j+1)%3;
    if (sees[j]->fe == sees[k]->fe)
      iSameFE = i;
    if (sees[j]->me == sees[k]->me)
      iSameME = i;
  }

  if (iSameFE != -1) {
    int i = iSameFE;
    int j = (i+1)%3;
    int k = (j+1)%3;
    // f[j] == f[k]
  
    // Calculate the intersections of mj,rj and mk,rk
    Point2 mm = m[k] - m[j];
    Point2 op[2], cp[2];
    if (!intersectCircles(r[j], mm, r[k], op, cp))
      assert(0);
    Point2 pjk[2] = { m[j] + op[0], m[j] + op[1] };

    int h = index / 2;
    // f[jk] + q pjk[h] intersects f[i] + q m[i], r[i]
    // |(f[jk] - f[i]) + q (pjk[h] - m[i])| = r[i]
    Point2 ff = f[j] - f[i];
    Point2 mp = pjk[h] - m[i];
    if (!intersectCircles(r[i], ff, mp.norm(), op, cp))
      assert(0);
    int g = index % 2;
    theta = cp[g].unit().divides(mp.unit());
    p = f[j] + theta.times(pjk[h]);
  }
  else if (iSameME != -1) {
    int i2 = iSameME;
    int i0 = (i2+1)%3;
    int i1 = (i0+1)%3;

    // Use substitution p = p - (f0 + q m0)

    // p^2 = r0^2
    // (p - f01)^2 = r1^2
    // (p - (f02 + q m02))^2 = r2^2
    // ((f02 - p) + q m02))^2 = r2^2
    // q^2 = 1
  
    Point2 f01 = f[i1] - f[i0];
    Point2 op[2], cp[2];
    if (!intersectCircles(r[i0], f01, r[i1], op, cp))
      assert(0);

    Point2 m02 = m[i2] - m[i0];
    Parameter m02norm = m02.norm();
    Point2 m02inv = inverse(m02);
    int i = index / 2;
    Point2 pf02 = f[i2] - f[i0] - op[i];
    Point2 op2[2], cp2[2];
    if (!intersectCircles(r[i2], pf02, m02norm, op2, cp2))
      assert(0);
    int j = index % 2;
    theta = cp2[j].times(m02inv);
    p = op[i] + f[i0] + theta.times(m[i0]);
  }
  else {
    int i0 = 0;
    for (int i = 1; i < 3; i++)
      if (r[i0].value() < r[i].value())
        i0 = i;
    int i1 = (i0+1)%3;
    int i2 = (i1+1)%3;
    
    Parameter rb[3] = { r[i0], r[i1], r[i2] };
    Point2 fb[3] =
      { f[i0] - f[i0], f[i1] - f[i0], f[i2] - f[i0]};
    Point2 mb[3] =
      { m[i0] - m[i0], m[i1] - m[i0], m[i2] - m[i0]};
    
    Parameter rc[3] =
      { rb[0] / rb[0], rb[1] / rb[0], rb[2] / rb[0] };
    Point2 fc[3] =
      { fb[0] / rb[0], fb[1] / rb[0], fb[2] / rb[0] };
    Point2 mc[3] =
      { mb[0] / rb[0], mb[1] / rb[0], mb[2] / rb[0] };
    
    Parameter rd[3] = { rc[0], rc[1] / mc[1].norm(), rc[2] / mc[2].norm() };
    Point2 md[3] = { mc[0], inverse(mc[1]), inverse(mc[2]) };
    Point2 fd[3] = { fc[0], -fc[1].times(md[1]), -fc[2].times(md[2]) };

    static int count;
    if (++count == 2)
      cout << "this is it" << endl;
    
    p.increasePrecision();
    theta.increasePrecision();
    // Point2 qd = (p - (f[i0] + m[i0].times(theta))) / rb[0];
    
    PPoly poly = criticality3eq(md, fd, rd);
    qd = polishCS(poly, qd);
    // bool flag = circleImplicit(qd.x, qd.y, poly, qd);
    
    Point2 cd[3];
    for (int j = 0; j < 3; j++) {
      cd[j] = fd[j] + md[j].times(qd);
      // drawCircle(cd[j], rd[j], i);
    }
    
    Point2 pd = calcP(cd, rd);
    
    theta = pd;
    p = qd * rb[0] + f[i0] + m[i0].times(theta);
    
    if (false) {
    for (int j = 0; j < 3; j++) {
      Point2 cpd = pd - (fd[j] + qd.times(md[j]));
      Parameter cp2r2d = cpd.innerproduct(cpd) - rd[j] * rd[j];
      cout << "cp2r2d " << cp2r2d.value() << endl;
      Point2 cp = p - (f[j] + theta.times(m[j]));
      Parameter cp2r2 = cp.innerproduct(cp) - r[j] * r[j];
      cout << "cp2r2 " << cp2r2.value() << endl;
    }
    }
  }
}

void CSpace::Type3::decreasePrecision () {
  if (theta.x.l != 1e20)
    return;

  for (int i = 0; i < 3; i++)
    sees[i]->decreasePrecision();

  theta.decreasePrecision();
  p.decreasePrecision();
}

CSpace::DSubE CSpace::DSubE::nextRight () {
  if (e == 0 || e->t->type() != I || e->h->type() != I)
    return DSubE();
  Intn *head = dynamic_cast<Intn *>(getTH(1));
  int dNext = d ^ (e->see == head->a);
  vector<SubE *> &subsNext = e->see == head->a ? head->b->getSubs(false) : head->a->getSubs(false);
  for (int i = 0; i < subsNext.size(); i++)
    if (subsNext[i]->getTH(dNext) == head)
      return DSubE(subsNext[i], dNext);
}

void CSpace::checkTriangle (DSubE dsub0) {
  static int count;
  if (++count == -1)
    cout << "checkTriangle this is it" << endl;
  DSubE dsub1 = dsub0.nextRight();
  DSubE dsub2 = dsub1.nextRight();
  if (dsub2.nextRight() != dsub0)
    return;

  SumEE *see012[] = { dsub0.getSEE(), dsub1.getSEE(), dsub2.getSEE() };

  createType3Events(see012);
}

void CSpace::processType1Event (Event *event) {
  static int count;
  if (++count == -1)
    cout << "processType1Event this is it" << endl;

  assert(event->a->type() == V && event->b->type() == V && event->c == 0);

  Vert *fv = dynamic_cast<Vert *>(event->a);
  Vert *mv = dynamic_cast<Vert *>(event->b);

  /*
  int fi = fv->index();
  int mi = mv->index();
  if (fi == -1 && mi == 2)
    cout << "fi " << fi << " mi " << mi << endl;

  int fh = (fi + fvs.size() - 1) % fvs.size();
  int mh = (mi + mvs.size() - 1) % mvs.size();

  Edge *feHT[2] = { fes[fh], fes[fi] };
  Edge *meHT[2] = { mes[mh], mes[mi] };
  */

  int fi = fv->index();
  int mi = mv->index();
  int ft = fv->t->index();
  int mt = mv->t->index();
  int fh = fv->h->index();
  int mh = mv->h->index();

  Edge *feHT[2] = { fv->h, fv->t };
  Edge *meHT[2] = { mv->h, mv->t };

  SumV *svB[4], *svA[4];
  int nSvB = 0, nSvA = 0;

  for (int ht = 0; ht < 2; ht++)
    if ((feHT[ht]->isConcave() ^ ht) == 0) {
      svB[nSvB] = sevs[sumInd(feHT[ht]->index(), mi)];
      nSvB += (svB[nSvB] != 0);
    }
    else {
      svA[nSvA] = sevs[sumInd(feHT[ht]->index(), mi)];
      nSvA += (svA[nSvA] != 0);
    }

  for (int ht = 0; ht < 2; ht++)
    if ((meHT[ht]->isConcave() ^ ht) == 1) {
      svB[nSvB] = sves[sumInd(fi, meHT[ht]->index())];
      nSvB += (svB[nSvB] != 0);
    }
    else {
      svA[nSvA] = sves[sumInd(fi, meHT[ht]->index())];
      nSvA += (svA[nSvA] != 0);
    }

  SumEE *see03 = sees[sumInd(fh, mt)];
  SumEE *see12 = sees[sumInd(ft, mh)];

  if (count == -1) {
    cout << "see03tp this is it" << endl;
  }

  SumEE *intSEE0B = 0, *intSEE1B = 0;
  {
    Edge *ffmmB[] = { fes[fh], fes[ft], mes[mh], mes[mt] };
    int stabB = type1stab(ffmmB, 0);
    if (stabB > 0) {
      intSEE0B = see03 /* sees[sumInd(ffmmB[0], ffmmB[3])] */;
      intSEE1B = see12 /* sees[sumInd(ffmmB[1], ffmmB[2])] */;
    }
    else if (stabB < 0) {
      intSEE0B = see12 /* sees[sumInd(ffmmB[1], ffmmB[2])] */;
      intSEE1B = see03 /* sees[sumInd(ffmmB[0], ffmmB[3])] */;
    }
  }

  SumEE *intSEE0A = 0, *intSEE1A = 0;
  Intn *intnA = 0;
  {
    Edge *ffmmA[] = { fes[fh], fes[ft], mes[mh], mes[mt] };
    int stabA = type1stab(ffmmA, 1);
    if (stabA > 0) {
      intSEE0A = see03 /* sees[sumInd(ffmmA[0], ffmmA[3])] */;
      intSEE1A = see12 /* sees[sumInd(ffmmA[1], ffmmA[2])] */;
    }
    else if (stabA < 0) {
      intSEE0A = see12 /* sees[sumInd(ffmmA[1], ffmmA[2])] */;
      intSEE1A = see03 /* sees[sumInd(ffmmA[0], ffmmA[3])] */;
    }

    if (intSEE0A != 0) {
      intnA = new Intn(intSEE0A, intSEE1A);
      //      if (intSEE0A->index() == 316 && intSEE1A->index() == 344 ||
      // intSEE0A->index() == 344 && intSEE1A->index() == 316)
      // cout << "type1 this is it" << endl;
    }
  }

  vector<SubE *> oldSubs, newSubs;

  for (int fht = 0; fht < 2; fht++)
    for (int mht = 0; mht < 2; mht++) {
      SumEE *see = sees[sumInd(feHT[fht]->index(), meHT[mht]->index())];
      if (see == 0)
        continue;
      vector<SubE *> &subs = see->getSubs();
      SumV *tailB = 0, *headB = 0, *tailA = 0, *headA = 0;
      for (int i = 0; i < nSvB; i++)
        if (svB[i]->tailOf == see)
          tailB = svB[i];
        else if (svB[i]->headOf == see)
          headB = svB[i];
      for (int i = 0; i < nSvA; i++)
        if (svA[i]->tailOf == see)
          tailA = svA[i];
        else if (svA[i]->headOf == see)
          headA = svA[i];

      if (tailB != 0 && tailA != 0) {
        assert(headB == 0 && headA == 0);
        if (see == intSEE0B)
          assert(dynamic_cast<Intn *>(subs[0]->h)->b == intSEE1B);
        else if (see == intSEE1B)
          assert(dynamic_cast<Intn *>(subs[0]->h)->a == intSEE0B);

        if (see == intSEE0B || see == intSEE1B) {
          oldSubs.push_back(subs[0]);
          subs.erase(subs.begin());
        }
        SubE *save = subs[0];
        oldSubs.push_back(subs[0]);
        subs.erase(subs.begin());

        subs.insert(subs.begin(), new SubE(see));
        newSubs.push_back(subs[0]);
        subs[0]->setT(tailA);
        subs[0]->setH(save->h);

        if (see == intSEE0A || see == intSEE1A) {
          subs[0]->setT(intnA);
          subs.insert(subs.begin(), new SubE(see));
          newSubs.push_back(subs[0]);
          subs[0]->setT(tailA);
          subs[0]->setH(intnA);
        }
      }
      else if (headB != 0 && headA != 0) {
        assert(tailB == 0 && tailA == 0);
        if (see == intSEE0B)
          assert(dynamic_cast<Intn *>(subs.back()->t)->b == intSEE1B);
        else if (see == intSEE1B)
          assert(dynamic_cast<Intn *>(subs.back()->t)->a == intSEE0B);

        if (see == intSEE0B || see == intSEE1B) {
          oldSubs.push_back(subs.back());
          subs.pop_back();
        }
        SubE *save = subs.back();
        oldSubs.push_back(subs.back());
        subs.pop_back();

        subs.push_back(new SubE(see));
        newSubs.push_back(subs.back());
        subs.back()->setH(headA);
        subs.back()->setT(save->t);

        if (see == intSEE0A || see == intSEE1A) {
          subs.back()->setH(intnA);
          subs.push_back(new SubE(see));
          newSubs.push_back(subs.back());
          subs.back()->setH(headA);
          subs.back()->setT(intnA);
        }
      }
      else if (tailB != 0 && headB != 0) {
        assert(tailA == 0 && headA == 0);
        if (see == intSEE0B)
          assert(dynamic_cast<Intn *>(subs.back()->t)->a == see);
        else if (see == intSEE1B)
          assert(dynamic_cast<Intn *>(subs.back()->t)->b == see);
        else
          assert(intSEE0B == 0 && intSEE1B == 0);
        if (see == intSEE0B || see == intSEE1B)
          assert(subs.size() == 2);
        else
          assert(subs.size() == 1);

        while (subs.size() > 0) {
          oldSubs.push_back(subs.back());
          oldSubs.back()->zerolength = true;
          subs.pop_back();
        }
      }
      else if (tailA != 0 && headA != 0) {
        assert(tailB == 0 && headB == 0);
        assert(see != intSEE0B && see != intSEE1B);
        assert(subs.size() == 0);

        subs.push_back(new SubE(see));
        newSubs.push_back(subs.back());
        newSubs.back()->zerolength = true;
        subs.back()->setT(tailA);
        subs.back()->setH(headA);

        if (see == intSEE0A || see == intSEE1A) {
          subs.back()->setH(intnA);
          subs.push_back(new SubE(see));
          newSubs.push_back(subs.back());
          newSubs.back()->zerolength = true;
          subs.back()->setT(intnA);
          subs.back()->setH(headA);
        }
      }
      else
        assert(0);
    }

  updateSubs(oldSubs, newSubs, event);
}

int CSpace::type1stab (Edge *ffmm[4], int ba) {
  if (ffmm[0]->c == 0 || ffmm[1]->c == 0 ||
      ffmm[2]->c == 0 || ffmm[3]->c == 0)
    return 0;

  int rSign = 1;
  int crossSign = 1;

  if (ba == 0) {
    swap(ffmm[0], ffmm[2]);
    swap(ffmm[1], ffmm[3]);
    crossSign = -crossSign;
  }
    
  if (ffmm[2]->isConcave() != ffmm[3]->isConcave() &&
      ffmm[0]->isConcave() == ffmm[1]->isConcave()) {
    swap(ffmm[0], ffmm[3]);
    swap(ffmm[1], ffmm[2]);
    crossSign = -crossSign;
  }

  if (ffmm[2]->isConcave() && ffmm[3]->isConcave()) {
    swap(ffmm[0], ffmm[1]);
    swap(ffmm[2], ffmm[3]);
    crossSign = -crossSign;
    rSign = -1;
  }

  bool concave = (rSign == 1);

  if (ffmm[0]->isConcave() == !concave && ffmm[3]->isConcave() == concave) {
    swap(ffmm[0], ffmm[3]);
    swap(ffmm[1], ffmm[2]);
    crossSign = -crossSign;
    assert(rSign == 1);
  }

  if (ffmm[0]->isConcave() == concave &&
      ffmm[1]->isConcave() == concave &&
      ffmm[2]->isConcave() == !concave &&
      ffmm[3]->isConcave() == !concave) {
    int sign03 = rSign * (ffmm[0]->c->getS() + ffmm[3]->c->getS()).sign();
    int sign02 = rSign * (ffmm[0]->c->getS() + ffmm[2]->c->getS()).sign();
    int sign12 = rSign * (ffmm[1]->c->getS() + ffmm[2]->c->getS()).sign();
    if (sign03 == sign12 && sign03 != sign02)
      return -crossSign;
    return 0;
  }

  if (ffmm[0]->isConcave() == concave &&
      ffmm[1]->isConcave() == !concave &&
      ffmm[2]->isConcave() == !concave &&
      ffmm[3]->isConcave() == !concave) {
    if (rSign * (ffmm[0]->c->getS() + ffmm[2]->c->getS()).sign() > 0 &&
        rSign * (ffmm[0]->c->getS() + ffmm[3]->c->getS()).sign() < 0)
      return -crossSign;
    return 0;
  }

  if (ffmm[0]->isConcave() == concave &&
      ffmm[1]->isConcave() == !concave &&
      ffmm[2]->isConcave() == concave &&
      ffmm[3]->isConcave() == !concave) {
    int sign03 = rSign * (ffmm[0]->c->getS() + ffmm[3]->c->getS()).sign();
    int sign12 = rSign * (ffmm[1]->c->getS() + ffmm[2]->c->getS()).sign();
    int sign0312 =
      rSign * (ffmm[0]->c->getS() + ffmm[3]->c->getS() -
               ffmm[1]->c->getS() - ffmm[2]->c->getS()).sign();
    if (sign03 > 0 && sign12 > 0 && sign0312 > 0)
      return crossSign;
    if (sign03 < 0 && sign12 < 0 && sign0312 > 0)
      return -crossSign;
    return 0;
  }

  return 0;
}

void CSpace::processBump (Event *event) {
  static int count;
  if (++count == -1)
    cout << "processBump this is it" << endl;

  SumEE *see1 = dynamic_cast<SumEE *>(event->a);
  SumEE *see2 = dynamic_cast<SumEE *>(event->b);
  vector<SubE *> &subs1 = see1->getSubs();
  vector<SubE *> &subs2 = see2->getSubs();
  vector<SubE *> oldSubs, newSubs;
 
  if (event->type2dir == 0 || event->type2dir == 3) {
    Intn *intn12 = new Intn(see1, see2);
    Intn *intn21 = new Intn(see2, see1);

    Point2 n1 = event->n1;
    Point2 n2 = event->n2;
    int i1 = see1->findSub(event->theta, n1);
    int i2 = see2->findSub(event->theta, n2);

    oldSubs.push_back(subs1[i1]);
    SubE *save1 = subs1[i1];
    subs1[i1] = new SubE(see1);
    newSubs.push_back(subs1[i1]);
    subs1[i1]->setH(save1->h);
    subs1.insert(subs1.begin() + i1, new SubE(see1));
    newSubs.push_back(subs1[i1]);
    subs1.insert(subs1.begin() + i1, new SubE(see1));
    newSubs.push_back(subs1[i1]);
    assert(subs1[i1+0] == newSubs[newSubs.size()-1] &&
           subs1[i1+1] == newSubs[newSubs.size()-2] &&
           subs1[i1+2] == newSubs[newSubs.size()-3]);
    subs1[i1+1]->zerolength = true;
    subs1[i1]->setT(save1->t);
    if (see2->isConcave()) {
      subs1[i1+1]->setT(intn12);
      subs1[i1]->setH(intn12);
      subs1[i1+2]->setT(intn21);
      subs1[i1+1]->setH(intn21);
    }
    else {
      subs1[i1+1]->setT(intn21);
      subs1[i1]->setH(intn21);
      subs1[i1+2]->setT(intn12);
      subs1[i1+1]->setH(intn12);
    }
    
    oldSubs.push_back(subs2[i2]);
    SubE *save2 = subs2[i2];
    subs2[i2] = new SubE(see2);
    newSubs.push_back(subs2[i2]);
    subs2[i2]->setH(save2->h);
    subs2.insert(subs2.begin() + i2, new SubE(see2));
    newSubs.push_back(subs2[i2]);
    subs2.insert(subs2.begin() + i2, new SubE(see2));
    newSubs.push_back(subs2[i2]);
    subs2[i2]->setT(save2->t);
    assert(subs2[i2+0] == newSubs[newSubs.size()-1] &&
           subs2[i2+1] == newSubs[newSubs.size()-2] &&
           subs2[i2+2] == newSubs[newSubs.size()-3]);
    subs2[i2+1]->zerolength = true;
    if (event->type2dir == 0 && see1->isConcave() ||
        event->type2dir == 3 && !see1->isConcave()) {
      subs2[i2+1]->setT(intn21);
      subs2[i2]->setH(intn21);
      subs2[i2+2]->setT(intn12);
      subs2[i2+1]->setH(intn12);
    }
    else {
      subs2[i2+1]->setT(intn12);
      subs2[i2]->setH(intn12);
      subs2[i2+2]->setT(intn21);
      subs2[i2+1]->setH(intn21);
    }
  }
  else {
    int i1;
    for (i1 = 1; i1+1 < subs1.size(); i1++) {
      SubE *sub1 = subs1[i1];
      Intn *t = sub1->getITH(0);
      Intn *h = sub1->getITH(1);
      if (t->a == see2 && h->b == see2 ||
          t->b == see2 && h->a == see2)
        break;
    }
    assert(i1+1 < subs1.size());

    int i2;
    for (i2 = 1; i2+1 < subs2.size(); i2++) {
      SubE *sub2 = subs2[i2];
      Intn *t = sub2->getITH(0);
      Intn *h = sub2->getITH(1);
      if (t->a == see1 && h->b == see1 ||
          t->b == see1 && h->a == see1)
        break;
    }
    assert(i2+1 < subs2.size());

    if (see2->isConcave())
      assert(subs1[i1]->getITH(0)->a == see1 &&
             subs1[i1]->getITH(1)->b == see1);
    else
      assert(subs1[i1]->getITH(0)->b == see1 &&
             subs1[i1]->getITH(1)->a == see1);
    if (event->type2dir == 1 && see1->isConcave() ||
        event->type2dir == 2 && !see1->isConcave())
      assert(subs2[i2]->getITH(0)->a == see2 &&
             subs2[i2]->getITH(1)->b == see2);
    else
      assert(subs2[i2]->getITH(0)->b == see2 &&
             subs2[i2]->getITH(1)->a == see2);
    
    oldSubs.push_back(subs1[i1-1]);
    oldSubs.push_back(subs1[i1]);
    oldSubs.push_back(subs1[i1+1]);
    oldSubs.push_back(subs2[i2-1]);
    oldSubs.push_back(subs2[i2]);
    oldSubs.push_back(subs2[i2+1]);

    subs1[i1]->zerolength = true;
    subs2[i2]->zerolength = true;

    SubE *sub1 = new SubE(see1);
    newSubs.push_back(sub1);
    sub1->setT(subs1[i1-1]->t);
    sub1->setH(subs1[i1+1]->h);

    SubE *sub2 = new SubE(see2);
    newSubs.push_back(sub2);
    sub2->setT(subs2[i2-1]->t);
    sub2->setH(subs2[i2+1]->h);

    subs1.erase(subs1.begin() + i1);
    subs1.erase(subs1.begin() + i1);
    subs1[i1-1] = sub1;

    subs2.erase(subs2.begin() + i2);
    subs2.erase(subs2.begin() + i2);
    subs2[i2-1] = sub2;
  }

  updateSubs(oldSubs, newSubs, event);
}

int CSpace::SumEE::findSub (Point2 theta, Point2 n) {
  int a = 0, b = subs.size()-1;
  while (a < b) {
    int c = (a+b)/2;
    Point2 nc = dynamic_cast<Intn *>(subs[c]->h)->getN(theta, this);
    if (((n.crossproductSign(nc) > 0) ^ isConcave()) == 1)
      b = c;
    else
      a = c + 1;
  }
  return a;
}

bool cpsRecursing = false;

int CSpace::SumEE::crossproductSign (SumEE *that, Point2 &p, Point2 &q) {
  Point2 thisC = this->getC(q);
  Point2 cc = that->getC(q) - this->getC(q);
  Point2 ccu = cc.unit();
  Point2 cp = p - thisC;
  int thisSsign = this->s.sign();
  int thatSsign = that->s.sign();

  if (thisSsign != thatSsign)
    return thisSsign * thatSsign * ccu.crossproductSign(cp);

  Parameter thisfes = this->fe->c->getS();
  Parameter thismes = this->me->c->getS();
  Parameter thatfes = that->fe->c->getS();
  Parameter thatmes = that->me->c->getS();

  Vert *vf = 0, *vm = 0;
  if (this->fe->h == that->fe->t && this->me->t == that->me->h) {
    vf = this->fe->h;
    vm = this->me->t;
  }
  else if (this->fe->t == that->fe->h && this->me->h == that->me->t) {
    vf = this->fe->t;
    vm = this->me->h;
  }
  else
    return thisSsign * thatSsign * ccu.crossproductSign(cp);

  Point2 u = vf->n->getN().divides(vm->n->getN());

  assert(thisSsign < 0);

  SumEE *see14 = 0, *see23 = 0;
  if (this->r < that->r) {
    see14 = that;
    see23 = this;
  }
  else {
    see14 = this;
    see23 = that;
  }

  Edge *e1 = see14->fe, *e2 = see23->fe, *e3 = see23->me, *e4 = see14->me;
  Point2 qvm = q.times(vm->n->getN());
  Point2 c23p = p - see23->getC(q);
  Point2 c14p = p - see14->getC(q);

  int proxy = 0;

  if (e4->c->getS() < e3->c->getS()) {
    assert(e2->c->getS() < e1->c->getS());

    Point2 qnm = q.times(vm->n->getN());
    if (qnm.innerproductSign(c23p) < 0)
      proxy = -qnm.crossproductSign(c23p);
  }
  else /* -s3 + s4 > 0 */ {
    if ((e2->c->getS() + e4->c->getS()).sign() > 0) {
      Parameter A = e4->c->getS() - e3->c->getS();
      Parameter B = see14->r + see23->r;
      Parameter C = e2->c->getS() - e1->c->getS();
      Parameter COS = (A * A + B * B - C * C) / (A * B);
      Point2 qnm = q.times(vm->n->getN());
      if (qnm.innerproduct(c23p) < see23->r * COS)
        proxy = -qnm.crossproductSign(c23p);
    }
    else /* s2 + s4 < 0 */ {
      proxy = -vf->n->getN().crossproductSign(c14p);
    }
  }

  if (proxy == 0)
    return ccu.crossproductSign(cp);
  /*
  if (qvm.innerproductSign(c23p) > 0)
    return ccu.crossproductSign(cp);
  int proxy = -qvm.crossproductSign(c23p);
  */

  int swapSign = see14 == this ? 1 : -1;
  int ret = swapSign * proxy;
  Parameter::signFail = false;
  if (!(ccu.crossproductSign(cp) == ret))
    cout << "bad singular case" << endl;
  Parameter::signFail = true;

  // return thisSsign * thatSsign * ccu.crossproductSign(cp);
  return ret;
}

Point2 CSpace::Intn::getN (Point2 theta, SumEE *see) {
  SumEE *see01[] = { see, 0 };
  if (see == a)
    see01[1] = b;
  else if (see == b)
    see01[1] = a;
  else
    assert(0);
  
  Point2 c = see01[1]->getC(theta) - see01[0]->getC(theta);
  Point2 q[2], cq[2];
  bool check = intersectCircles(see01[0]->r, c, see01[1]->r, q, cq);
  assert(check);
  int i = (see == b) ^ a->isConcave() ^ b->isConcave();
  Point2 qi = q[i];
  return qi / see->s;
}

void CSpace::processStab (Event *event) {
  static int count;
  if (++count == 0)
    cout << "processStab this is it" << endl;

  SumEE *see = dynamic_cast<SumEE *>(event->a);
  vector<SubE *> &subs = see->getSubs();
  int n1xn2 = -1;
  try {
    n1xn2 = (event->n1.crossproductSign(event->n2) < 0);
  } catch (SignException signException) {
    event->increasePrecision();
    n1xn2 = (event->n1.crossproductSign(event->n2) < 0);
    event->decreasePrecision();
  }
  assert(n1xn2 != -1);
  SumV *sv = dynamic_cast<SumV *>(event->b);
  SumEE *seeT = sv->tailOf;
  SumEE *seeH = sv->headOf;
  vector<SubE *> oldSubs, newSubs;

  if (count == -1 /* 118 */) {
    vector<SubE *> &subsH = seeH->getSubs();
    Intn *intn0 = (Intn*)subsH[subsH.size()-2]->t;
    Intn *intn1 = (Intn*)subsH[subsH.size()-1]->t;
    PV p0 = intn0->getP(P2PV(event->theta));
    PV p1 = intn1->getP(P2PV(event->theta));
    cout << "this is it" << endl;
  }

  Point2 seeC = see->getC(event->theta);

  if ((event->type2dir ^ n1xn2 ^ see->isConcave()) == 0) {
    // intersection moves from tail of tailOf to head of headOf
    int isub = -1;
    Obj *saveT = 0, *saveH = 0;
    if (seeT != 0) {
      vector<SubE *> &subsT = seeT->getSubs();

#ifndef BLEEN      
      if (++count == 949)
	for (int i = 1; i < subsT.size()-1; i++)
	  subsT[i]->check(event);
#endif

      assert(see == (n1xn2 ? subsT[0]->getITH(1)->b : subsT[0]->getITH(1)->a));
      oldSubs.push_back(subsT[0]);
      oldSubs.back()->zerolength = true;
      oldSubs.push_back(subsT[1]);
      SubE *subT = new SubE(seeT);
      newSubs.push_back(subT);
      subT->setT(subsT[0]->t);
      subT->setH(subsT[1]->h);
      subsT.erase(subsT.begin());
      subsT[0] = subT;

      for (isub = 0; isub < subs.size(); isub++)
        if (seeT ==
            (n1xn2 ? subs[isub]->getITH(1)->a : subs[isub]->getITH(1)->b))
          break;
      assert(isub != subs.size());
      saveT = subs[isub]->t;
      saveH = subs[isub+1]->h;
      oldSubs.push_back(subs[isub]);
      oldSubs.push_back(subs[isub+1]);
      subs.erase(subs.begin() + isub);
    }
    else {
      isub = see->findSub(event->theta, event->n1);
      oldSubs.push_back(subs[isub]);
      saveT = subs[isub]->t;
      saveH = subs[isub]->h;
    }

    if (seeH != 0) {
      Intn *intnH = n1xn2 ? new Intn(seeH, see) : new Intn(see, seeH);
      vector<SubE *> &subsH = seeH->getSubs();
      SubE *subH = subsH.back();
      oldSubs.push_back(subH);
      subsH.back() = new SubE(seeH);
      newSubs.push_back(subsH.back());
      subsH.back()->setT(subH->t);
      subsH.back()->setH(intnH);
      subsH.push_back(new SubE(seeH));
      newSubs.push_back(subsH.back());
      subsH.back()->setT(intnH);
      subsH.back()->setH(subH->h);
      newSubs.back()->zerolength = true;

      subs[isub] = new SubE(see);
      newSubs.push_back(subs[isub]);
      subs[isub]->setT(intnH);
      subs[isub]->setH(saveH);

      subs.insert(subs.begin() + isub, new SubE(see));
      newSubs.push_back(subs[isub]);
      subs[isub]->setT(saveT);
      subs[isub]->setH(intnH);
    }
    else {
      subs[isub] = new SubE(see);
      newSubs.push_back(subs[isub]);
      subs[isub]->setT(saveT);
      subs[isub]->setH(saveH);
    }
  }
  else {
    // intersection moves from head of headOf to tail of tailOf
    int isub = -1;
    Obj *saveT = 0, *saveH = 0;
    if (seeH != 0) {
      vector<SubE *> &subsH = seeH->getSubs();
      if (!(see == (n1xn2 ? subsH.back()->getITH(0)->b :
                    subsH.back()->getITH(0)->a))) {
        Intn *intn13_42 = (Intn *) seeH->getSubs()[1]->h;
        Point2  n13_42 = intn13_42->getN(event->theta, intn13_42->b);
        Intn *intn154_42 = (Intn *) see->getSubs()[1]->h;
        Point2  n154_42 = intn154_42->getN(event->theta, intn154_42->b);
        cout << "this is it 13 33 154" << endl;

        Intn *intn13_154 = (Intn *) see->getSubs()[0]->h;
        Point2 n13_154 = intn13_154->getN(event->theta, intn13_154->b);
        Intn *intn42_154 = (Intn *) see->getSubs()[1]->h;
        Point2 n42_154 = intn42_154->getN(event->theta, intn42_154->a);
        cout << "this is it 13 33 154" << endl;
      }
      assert(see == (n1xn2 ? subsH.back()->getITH(0)->b :
                     subsH.back()->getITH(0)->a));
      oldSubs.push_back(subsH.back());
      oldSubs.back()->zerolength = true;
      oldSubs.push_back(subsH[subsH.size()-2]);
      SubE *subH = new SubE(seeH);
      newSubs.push_back(subH);
      subH->setT(subsH[subsH.size()-2]->t);
      subH->setH(subsH.back()->h);
      subsH.pop_back();
      subsH.back() = subH;

      for (isub = 0; isub < subs.size(); isub++)
        if (seeH ==
            (n1xn2 ? subs[isub]->getITH(1)->a : subs[isub]->getITH(1)->b))
          break;
      assert(isub != subs.size());
      saveT = subs[isub]->t;
      saveH = subs[isub+1]->h;
      oldSubs.push_back(subs[isub]);
      oldSubs.push_back(subs[isub+1]);
      subs.erase(subs.begin() + isub);
    }
    else {
      isub = see->findSub(event->theta, event->n1);
      oldSubs.push_back(subs[isub]);
      saveT = subs[isub]->t;
      saveH = subs[isub]->h;
    }

    if (seeT != 0) {
      Intn *intnT = n1xn2 ? new Intn(seeT, see) : new Intn(see, seeT);
      vector<SubE *> &subsT = seeT->getSubs();
      SubE *subT = subsT[0];
      oldSubs.push_back(subT);
      subsT[0] = new SubE(seeT);
      newSubs.push_back(subsT[0]);
      subsT[0]->setT(intnT);
      subsT[0]->setH(subT->h);
      subsT.insert(subsT.begin(), new SubE(seeT));
      newSubs.push_back(subsT[0]);
      subsT[0]->setT(subT->t);
      subsT[0]->setH(intnT);
      newSubs.back()->zerolength = true;

      subs[isub] = new SubE(see);
      newSubs.push_back(subs[isub]);
      subs[isub]->setT(intnT);
      subs[isub]->setH(saveH);

      subs.insert(subs.begin() + isub, new SubE(see));
      newSubs.push_back(subs[isub]);
      subs[isub]->setT(saveT);
      subs[isub]->setH(intnT);
    }
    else {
      subs[isub] = new SubE(see);
      newSubs.push_back(subs[isub]);
      subs[isub]->setT(saveT);
      subs[isub]->setH(saveH);
    }
  }

  updateSubs(oldSubs, newSubs, event);
}

void CSpace::processFlip (Event *event) {
  static int count;
  if (++count == -1)
    cout << "processFlip this is it" << endl;

  Intn *intns[3];
  intns[0] = dynamic_cast<Intn *>(event->a);
  intns[1] = dynamic_cast<Intn *>(event->b);
  intns[2] = dynamic_cast<Intn *>(event->c);
  bool reversed[3];
  vector<SubE *> oldSubs, newSubs;

  if (count == -1) {
    PV tiny(cos(0.001), sin(0.001));
    PV before[3], middle[3], after[3];
    for (int i = 0; i < 3; i++) {
      before[i] = intns[i]->getP(P2PV(event->theta).times(tiny.conjugate()));
      middle[i] = intns[i]->getP(P2PV(event->theta));
      after[i] = intns[i]->getP(P2PV(event->theta).times(tiny));
    }
    cout << "processFlip this is it" << endl;
  }

  for (int i = 0; i < 3; i++) {
    int j = (i+1)%3;
    int k = (i+2)%3;

    SumEE *see = intns[j]->a;
    if (see != intns[k]->a && see != intns[k]->b)
      see = intns[j]->b;
    assert(see == intns[k]->a || see == intns[k]->b);
    vector<SubE *> &subs = see->getSubs();

    int isub;
    for (isub = 0; isub < subs.size(); isub++)
      if (*intns[j] == *subs[isub]->t && *intns[k] == *subs[isub]->h) {
        reversed[i] = false;
        break;
      }
      else if (*intns[j] == *subs[isub]->h && *intns[k] == *subs[isub]->t) {
        reversed[i] = true;
        break;
      }
    assert(isub != subs.size());

    Obj *saveT = subs[isub-1]->t;
    Obj *saveH = subs[isub+1]->h;

    subs[isub]->zerolength = true;

    oldSubs.push_back(subs[isub-1]);
    subs[isub-1] = new SubE(see);
    newSubs.push_back(subs[isub-1]);
    subs[isub-1]->setT(saveT);
    subs[isub-1]->setH(subs[isub]->h);

    oldSubs.push_back(subs[isub+1]);
    subs[isub+1] = new SubE(see);
    newSubs.push_back(subs[isub+1]);
    subs[isub+1]->setH(saveH);
    subs[isub+1]->setT(subs[isub]->t);

    oldSubs.push_back(subs[isub]);
    subs[isub] = new SubE(see);
    newSubs.push_back(subs[isub]);
    subs[isub]->setT(subs[isub-1]->h);
    subs[isub]->setH(subs[isub+1]->t);

    subs[isub]->zerolength = true;
  }

  // assert(reversed[0] == reversed[1] && reversed[1] == reversed[2]);

  updateSubs(oldSubs, newSubs, event);
}

void CSpace::processEvents () {
  int count = 0;
  for (set<Event *, Compare>::iterator it = events.begin();
       it != events.end(); ++it) {
    count++;
    Event *event = *it;
    assert(event->theta.x.signFail);
    if (event->id == -1 || event->id == -1) {
      cout << "processEvents this is it" << endl;
      for (set<Event *, Compare>::iterator it2 = events.begin();
           it2 != events.end(); ++it2)
        cout << (*it2)->id << endl;
    }
    if (event->a->type() == V)
      processType1Event(event);
    else if (event->b->type() == SEE)
      processBump(event);
    else if (event->c == 0)
      processStab(event);
    else
      processFlip(event);
    drawArcs(P2PV(event->theta));
  }

  for (int i = 0; i < sees.size(); i++) {
    SumEE *see = sees[i];
    if (see == 0)
      continue;
    assert(see->getSubs().size() == see->subs0.size());
    for (int j = 0; j < see->getSubs().size(); j++) {
      see->getSubs()[j]->addNeighbor(see->subs0[j]);
      see->subs0[j]->addNeighbor(see->getSubs()[j]);
    }
  }

  int n = 0;
  for (set<Event *, Compare>::iterator it = events.begin();
       it != events.end(); it++)
    (*it)->index = n++;
}

void CSpace::drawArcs (PV theta) {
  // fast!
  return;
  if (nFigs < -1)
    cout << "drawArcs this is it" << endl;
  char buf[16];
  sprintf(buf, "arcs%04d.fig", nFigs++);
  ofstream out;
  out.open(buf);
  if (!out)
    cout << "Could not open " << buf << endl;

  out << "#FIG 3.2  Produced by xfig version 3.2.5" << endl
      << "Landscape" << endl
      << "Center" << endl
      << "Inches" << endl
      << "Letter  " << endl
      << "100.00" << endl
      << "Single" << endl
      << "-2" << endl
      << "1200 2" << endl;
  for (int i = 0; i < sees.size(); i++) {
    if (i == -1)
      cout << "see " << i << endl;

    SumEE *see = sees[i];
    if (see == 0)
      continue;

    PV fc(P2PV(see->fe->c->getC()));
    PV mc(P2PV(see->me->c->getC()));
    PV c = fc + mc.times(theta);
    double r = see->r.value();
    for (int j = 0; j < see->getSubs(false).size(); j++) {
      SubE *sub = see->getSubs(false)[j];

      PV t = sub->t->getP(theta);
      static double maxCTR;
      double ctr = fabs(c.distance(t) - r);
      if (ctr > maxCTR) {
        cout << "maxCTR " << ctr << endl;
        maxCTR = ctr;
      }

      PV h = sub->h->getP(theta);
      static double maxCHR;
      double chr = fabs(c.distance(h) - r);
      if (chr > maxCHR) {
        cout << "maxCHR " << chr << endl;
        maxCHR = chr;
        h = sub->h->getP(theta);
      }

      PV m = c + ((t - c) + (h - c)).unit() * r;
      static double maxCMR;
      double cmr = fabs(c.distance(m) - r);
      if (cmr > maxCMR) {
        cout << "maxCMR " << cmr << endl;
        maxCMR = cmr;
      }

      if (see->isConcave()) {
        PV temp = t;
        t = h;
        h = temp;
      }

      PV ct = t - c, cm = m - c, ch = h - c;

      if (t.distance(h) > 1.0)
        out << "5 1 0 1 0 7 " << i << " -1 -1 0.000 0 1 0 0"
            << " " << (1200*c.x) << " " << (-1200*c.y)
            << " " << (int) rint(1200*t.x) << " " << (int) rint(-1200*t.y)
            << " " << (int) rint(1200*m.x) << " " << (int) rint(-1200*m.y)
            << " " << (int) rint(1200*h.x) << " " << (int) rint(-1200*h.y)
            << endl;
      else
        out << "2 1 0 1 0 7 " << i << " -1 -1 0.000 0 0 -1 0 0 3" << endl
        << "\t"
        << " " << (int) rint(1200*t.x) << " " << (int) rint(-1200*t.y)
        << " " << (int) rint(1200*m.x) << " " << (int) rint(-1200*m.y)
        << " " << (int) rint(1200*h.x) << " " << (int) rint(-1200*h.y)
        << endl;
      out.flush();
      out.flush();
    }
  }

  out.close();
}

void CSpace::drawBody (int fm) {
  const char *name = fm ? "moving.fig" : "fixed.fig";
  vector<Vert *> &verts = fm ? mvs : fvs;
  vector<Edge *> &edges = fm ? mes : fes;
  
  ofstream out;
  out.open(name);
  if (!out)
    cout << "Could not open " << name << endl;

  const char *inname = !fm ? "path0.txt" : "path1.txt";
  ifstream in;
  in.open(inname);
  if (!in)
    cout << "Could not open " << inname << endl;

  cout << name << endl;
  out << "#FIG 3.2  Produced by xfig version 3.2.5" << endl
      << "Landscape" << endl
      << "Center" << endl
      << "Inches" << endl
      << "Letter  " << endl
      << "100.00" << endl
      << "Single" << endl
      << "-2" << endl
      << "1200 2" << endl;
  double x, y, theta;
  int n = 99;
  while (in >> x >> y >> theta) {
    n++;
    PV tran(x, y);
    double theta0123 = fm ? theta-randomAngle : theta;
    PV rot(cos(theta0123), sin(theta0123));
    if (fm)
      rot = -rot;

  for (int i = 0; i < edges.size(); i++) {
    Edge *e = edges[i];
    Vert *vt = e->t;
    Vert *vh = e->h;
    PV t(P2PV(vt->n->getP()));
    PV nt(P2PV(vt->n->getN()));
    PV h(P2PV(vh->n->getP()));
    PV nh(P2PV(vh->n->getN()));
    if (e->c == 0)
      continue;
    PV c(P2PV(e->c->getC()));
    double s = e->c->getS().value();
    double r = fabs(s);
    PV m = c + ((t - c) + (h - c)).unit() * r;

    if (e->isConcave()) {
      PV temp = t;
      t = h;
      h = temp;
    }

    c = c.times(rot) + tran;
    t = t.times(rot) + tran;
    h = h.times(rot) + tran;
    m = m.times(rot) + tran;

    int depth = fm == 0 ? i : n;
    if (t.distance(h) > 0.001)
      out << "5 1 0 1 0 7 " << depth << " -1 -1 0.000 0 1 0 0"
          << " " << (1200*c.x) << " " << (-1200*c.y)
          << " " << (int) rint(1200*t.x) << " " << (int) rint(-1200*t.y)
          << " " << (int) rint(1200*m.x) << " " << (int) rint(-1200*m.y)
          << " " << (int) rint(1200*h.x) << " " << (int) rint(-1200*h.y)
          << endl;
    else
      out << "2 1 0 1 0 7 " << depth << " -1 -1 0.000 0 0 -1 0 0 3" << endl
          << "\t"
          << " " << (int) rint(1200*t.x) << " " << (int) rint(-1200*t.y)
          << " " << (int) rint(1200*m.x) << " " << (int) rint(-1200*m.y)
          << " " << (int) rint(1200*h.x) << " " << (int) rint(-1200*h.y)
          << endl;

    /*
    out << "2 1 0 1 0 7 " << depth << " -1 -1 0.000 0 0 -1 0 0 3" << endl
        << "\t"
        << " " << (int) rint(1200*t.x) << " " << (int) rint(-1200*t.y)
        << " " << (int) rint(1200*m.x) << " " << (int) rint(-1200*m.y)
        << " " << (int) rint(1200*h.x) << " " << (int) rint(-1200*h.y)
        << endl;
    cout << setprecision(23)
         << "i " << depth << endl
         << "t " << t.x << " " << t.y << endl
         << "c " << c.x << " " << c.y << endl
         << "s " << s << endl;
    */
  }
  }
  in.close();
  out.close();
}

PV CSpace::SumVE::getP (PV theta) {
  PV v(P2PV(fv->n->getP()));
  PV n(P2PV(fv->n->getN()));
  PV c(P2PV(me->c->getC()));
  double s = me->c->getS().value();
  return v + c.times(theta) + n * s;
}

PV CSpace::SumEV::getP (PV theta) {
  PV c(P2PV(fe->c->getC()));
  double s = fe->c->getS().value();
  PV v(P2PV(mv->n->getP()));
  PV n(P2PV(mv->n->getN()));
  return c + n.times(theta) * s + v.times(theta);
}

PV CSpace::Intn::getP (PV theta) {
  PV fc1(P2PV(a->fe->c->getC()));
  PV mc1(P2PV(a->me->c->getC()));
  double r1 = a->r.value();
  PV fc2(P2PV(b->fe->c->getC()));
  PV mc2(P2PV(b->me->c->getC()));
  double r2 = b->r.value();
  
  PV c1 = fc1 + mc1.times(theta);
  PV c2 = fc2 + mc2.times(theta);

  bool swapped = false;
  if (r1 > r2) {
    double rTemp = r1;
    r1 = r2;
    r2 = rTemp;
    PV cTemp = c1;
    c1 = c2;
    c2 = cTemp;
    swapped = true;
  }

  PV c = c2 - c1;
  double d = c.length();
  double x = 0;

  if (r1 < d) {
    double e = d - r2;
    static double maxRE;
    if (r1 < -e) {
      double dist = -e - r1;
      if (dist > maxRE) {
        cout << "maxRE " << dist << endl;
        maxRE = dist;
      }
      e = -r1;
    }
    else if (r1 < e) {
      double dist = e - r1;
      if (dist > maxRE) {
        cout << "maxRE " << dist << endl;
        maxRE = dist;
      }
      e = r1;
    }

    x = e + (r1*r1 - e*e)/(2*d);
  }
  else {
    double f = r2 - r1;
    if (d < f) {
      static double maxFD;
      double fd = f - d;
      if (maxFD < fd) {
        cout << "maxFD " << fd << endl;
        maxFD = fd;
      }
      f = d;
    }
    double d2Mf2D2d = (d*d - f*f)/(2*d);
    double fDd = f/d;
    x = (d*d - f*f)/(2*d) - fDd * r1;
  }

  double y = sqrt(r1*r1 - x*x);
  assert(!isnan(y));

  PV u = c / d;
  PV v = u.rot90();

  PV q = swapped != (a->isConcave() == b->isConcave()) ?
    c1 + u * x + v * y : c1 + u * x - v * y;

#ifdef BLEEN  
  double dpp = d - r1 - r2;
  double dpm = d - r1 + r2;
  double dmp = d + r1 - r2;
  double dmm = d + r1 + r2;

  double dddd = dpp * dpm * dmp * dmm;
  static double maxDDDD;
  if (dddd > maxDDDD) {
    cout << "maxDDDD " << dddd << endl;
    maxDDDD = dddd;
  }
  if (dddd > 0)
    dddd = 0;

  PV u = c / d;
  PV v = u.rot90();
  double f = (d*d + r1*r1 - r2*r2) / d / 2;
  double h = sqrt(-dddd) / d / 2;

  PV q = a->isConcave() == b->isConcave() ?
    c1 + u * f + v * h : c1 + u * f - v * h;
#endif

  return q;
}

CSpace::SubE *CSpace::SubE::getNext () {
  if (index == -1)
    cout << "getNext this is it" << endl;
  if (h->type() == SVE || h->type() == SEV) {
    SumV *head = dynamic_cast<SumV *>(h);
    assert(head->headOf == see);
    if (head->tailOf == 0) {
      dangling = true;
      return 0;
    }
    return head->tailOf->getSubs(false)[0];
  }
  else {
    Intn *head = dynamic_cast<Intn *>(h);
    if (head->a == see) {
      dangling = true;
      return 0;
    }
    vector<SubE *> &subs = head->a->getSubs(false);
    for (int i = 0; i < subs.size(); i++)
      if (subs[i]->t == head)
        return subs[i];
    assert(0);
    return 0;
  }
}

CSpace::SubE *CSpace::SubE::getPrev () {
  if (index == -1)
    cout << "getNext this is it" << endl;
  if (t->type() == SVE || t->type() == SEV) {
    SumV *tail = dynamic_cast<SumV *>(t);
    assert(tail->tailOf == see);
    if (tail->headOf == 0) {
      dangling = true;
      return 0;
    }
    return tail->headOf->getSubs(false).back();
  }
  else {
    Intn *tail = dynamic_cast<Intn *>(t);
    if (tail->b == see) {
      dangling = true;
      return 0;
    }
    vector<SubE *> &subs = tail->b->getSubs(false);
    for (int i = 0; i < subs.size(); i++)
      if (subs[i]->h == tail)
        return subs[i];
    assert(0);
    return 0;
  }
}

void CSpace::SubE::addNeighbor (SubE *that) {
  // if (this->index == 76 && that->index == 2231)
  // cout << "this is it" << endl;
  for (int i = 0; i < neighbors.size(); i++)
    if (that == neighbors[i])
      return;
  neighbors.push_back(that);
}

bool CSpace::SubE::isNeighbor (SubE *that) {
  if (this->see != that->see)
    return false;
  if (!(this->a == that->b || this->b == that->a))
    return false;
  if (!(this->t == that->t || this->h == that->h))
    return false;
  if (this->zerolength || that->zerolength)
    return false;

  return true;
}

void CSpace::updateSubs (vector<SubE *> &oldSubs, vector<SubE *> &newSubs,
                         Event *event) {
  for (int i = 0; i < oldSubs.size(); i++)
    oldSubs[i]->b = event;

  for (int i = 0; i < newSubs.size(); i++) {
    SubE *newSub = newSubs[i];
    newSub->a = event;
    checkTriangles(newSub);
    SubE *prev = newSub->getPrev();
    if (prev != 0) {
      prev->addNeighbor(newSub);
      newSub->addNeighbor(prev);
    }
    SubE *next = newSub->getNext();
    if (next != 0) {
      next->addNeighbor(newSub);
      newSub->addNeighbor(next);
    }

    for (int j = 0; j < oldSubs.size(); j++) {
      SubE *oldSub = oldSubs[j];
      if (oldSub->isNeighbor(newSub)) {
        oldSub->addNeighbor(newSub);
        newSub->addNeighbor(oldSub);
      }
    }

    newSub->zerolength = false;
  }

  for (int i = 0; i < oldSubs.size(); i++)
    oldSubs[i]->zerolength = false;
}

void CSpace::findDangling () {
  for (int i = 0; i < allSubs.size(); i++) {
    allSubs[i]->next = 0;
    allSubs[i]->qnext = 0;
    allSubs[i]->visited = false;
  }

  SubE *front = 0, *back = 0;
  for (int i = 0; i < allSubs.size(); i++) {
    SubE *sub = allSubs[i];
    if (sub->visited || !sub->dangling)
      continue;

    sub->visited = true;
    front = back = sub;
    while (front != 0) {
      sub = front;
      front = front->qnext;
      if (front == 0)
        back = 0;
      
      for (int i = 0; i < sub->neighbors.size(); i++) {
        SubE *nei = sub->neighbors[i];
        if (nei->visited)
          continue;
        nei->visited = true;
        nei->dangling = true;
        if (front == 0)
          front = back = nei;
        else {
          back->qnext = nei;
          back = nei;
        }
      }
    }
  }
}

int CSpace::findComponents () {
  for (int i = 0; i < allSubs.size(); i++) {
    allSubs[i]->next = 0;
    allSubs[i]->qnext = 0;
    allSubs[i]->visited = false;
  }

  int nComponents = 0;

  SubE *front = 0, *back = 0;
  for (int i = 0; i < allSubs.size(); i++) {
    SubE *sub = allSubs[i];
    if (sub->component != -1 || sub->dangling)
      continue;

    sub->component = nComponents++;
    front = back = sub;
    while (front != 0) {
      sub = front;
      front = front->qnext;
      if (front == 0)
        back = 0;
      
      for (int i = 0; i < sub->neighbors.size(); i++) {
        SubE *nei = sub->neighbors[i];
        if (nei->component != -1) {
          assert(nei->component == sub->component);
          continue;
        }
        assert(!nei->dangling);
        nei->component = sub->component;
        if (front == 0)
          front = back = nei;
        else {
          back->qnext = nei;
          back = nei;
        }
      }
    }
  }

  return nComponents;
}

bool CSpace::findPath (SubE *start, SubE *finish) {
  if (start == finish)
    return true;
  notDangling = 0;
  for (int i = 0; i < allSubs.size(); i++) {
    allSubs[i]->next = 0;
    allSubs[i]->qnext = 0;
    allSubs[i]->visited = false;
    if (!allSubs[i]->dangling)
      notDangling++;
  }

  cout << "sum edges " << nSee << endl;
  cout << "subedges " << allSubs.size() << endl;
  cout << "not dangling " << notDangling << endl;

  int nVisited = 0;
  SubE *front = finish, *back = finish;
  finish->visited = true;
  nVisited++;

  while (front != 0) {
    SubE *sub = front;
    assert(!sub->dangling);
    front = front->qnext;
    if (front == 0)
      back = 0;

    for (int i = 0; i < sub->neighbors.size(); i++) {
      SubE *nei = sub->neighbors[i];
      if (nei->visited)
        continue;
      if (nei->index == -1)
        cout << "this is it" << endl;
      nei->visited = true;
      nVisited++;
      int aindex = nei->a == 0 ? 0 : nei->a->index+1;
      int bindex = nei->b == 0 ? events.size() : nei->b->index;
#ifdef PRINT_VISITED
      cout << "visited "
           << aindex << " "
           << bindex << " "
           << nei->see->index() << endl;
#endif
      nei->next = sub;
      if (nei == start) {
        cout << "nvisited " << nVisited << endl;
        return true;
      }
      if (front == 0)
        front = back = nei;
      else {
        back->qnext = nei;
        back = nei;
      }
    }
  }

  cout << "nvisited " << nVisited << endl;

  return false;
}

void checkC (PV p1, PV q1, PV p2, PV q2, PV c) {
  PV cp1 = p1 - c;
  PV cp2 = p2 - c;
  double cp12 = fabs(cp1.length() - cp2.length());
  static double maxcp12;
  if (maxcp12 < cp12) {
    cout << "maxcp12 " << cp12 << endl;
    maxcp12 = cp12;
  }

  PV cp1Dq1 = cp1.times(q1.conjugate());
  PV cp2Dq2 = cp2.times(q2.conjugate());
  double cpq12 = cp1Dq1.distance(cp2Dq2);
  static double maxcpq12;
  if (maxcpq12 < cpq12) {
    cout << "maxcpq12 " << cpq12 << endl;
    maxcpq12 = cpq12;
  }
}

Point2 CSpace::SubE::getN (Point2 theta, int th) {
  Obj *o = th == 0 ? t : h;
  if (o->type() == SVE)
    return dynamic_cast<SumVE *>(o)->fv->n->getN();
  else if (o->type() == SEV)
    return dynamic_cast<SumEV *>(o)->mv->n->getN().times(theta);
  else
    return dynamic_cast<Intn *>(o)->getN(theta, see);
}

bool CSpace::SubE::check (Event *event) {
  try {
    Point2 nt = getN(event->theta, 0);
    Point2 nh = getN(event->theta, 1);
    Parameter ntxnh = nt.crossproduct(nh);
    bool ret = ntxnh.sign() == see->s.sign();
    if (!ret)
      cout << "check failed" << endl;
    return ret;
  } catch (SignException e) {
    increasePrecision();
    event->increasePrecision();
    bool ret = check(event);
    decreasePrecision();
    event->decreasePrecision();
    return ret;
  }
}

bool CSpace::SubE::contains(Point2 theta, Point2 n) {
  if (!between(a->theta, theta, b->theta))
    return false;
  Point2 nt = getN(theta, 0);
  Point2 nh = getN(theta, 1);
  if (see->isConcave())
    return between(nh, n, nt);
  else
    return between(nt, n, nh);
}

bool CSpace::SubE::contains(Crossing *cr) {
  try {
    return contains(cr->getTheta(), cr->getN());
  } catch (SignException e) {
    increasePrecision();
    cr->increasePrecision();
    bool ret = contains(cr);
    decreasePrecision();
    cr->decreasePrecision();
    return ret;
  }
}

bool CSpace::findPath (double x1, double y1, double t1,
                       double x2, double y2, double t2,
                       vector<PV> &ps, vector<PV> &qs) {
  PV theta0PV(1, 0);
  Point2 theta0(theta0PV.x, theta0PV.y);
  Event *event000 = new Event(theta0);
  event000->index = -1;
  Event *event360 = new Event(theta0);
  event360->index = events.size();
  for (int i = 0; i < allSubs.size(); i++) {
    SubE *sub = allSubs[i];
    if (sub->a == 0)
      sub->a = event000;
    if (sub->b == 0)
      sub->b = event360;
  }

  for (int i = 0; i < sees.size(); i++) {
    SumEE *see = sees[i];
    if (see == 0)
      continue;
    if (see->subs0.size() == 0)
      continue;
    vector<SubE *> &subs0 = see->subs0;
    vector<SubE *> &subs = see->getSubs();
    vector<SubE *> &allSubs = see->allSubs;
    for (int j = 1; j < subs0.size(); j++) {
      assert(subs0[j]->t->type() == I);
      Intn *intn0 = dynamic_cast<Intn *>(subs0[j]->t);
      Intn *intn = dynamic_cast<Intn *>(subs[j]->t);
      for (int k = 0; k < allSubs.size(); k++) {
        if (allSubs[k]->t == intn)
          allSubs[k]->setT(intn0);
        if (allSubs[k]->h == intn)
          allSubs[k]->setH(intn0);
      }
    }
  }

  while (fabs(t2 + 2*M_PI - t1) < fabs(t2 - t1))
    t2 += 2*M_PI;
  while (fabs(t2 - 2*M_PI - t1) < fabs(t2 - t1))
    t2 -= 2*M_PI;
  
  double oldT1 = t1;
  double oldT2 = t2;
  if (fabs(t2 - t1) < 0.1) {
    Point2 p1(x1, y1);
    Point2 q1(cos(t1), sin(t1));
    Point2 q1s[] = { Point2(cos(t1-0.2), sin(t1-0.2)),
                     Point2(cos(t1+0.2), sin(t1+0.2)) };
    angleRange(p1, q1, q1s);

    PV qq1 = P2PV(q1).times(P2PV(q1s[0]).conjugate());
    t1 -= atan2(qq1.y, qq1.x);

    Point2 p2(x2, y2);
    Point2 q2(cos(t2), sin(t2));
    Point2 q2s[] = { Point2(cos(t2-0.2), sin(t2-0.2)),
                     Point2(cos(t2+0.2), sin(t2+0.2)) };
    angleRange(p2, q2, q2s);
    PV qq2 = P2PV(q2s[1]).times(P2PV(q2).conjugate());
    t2 += atan2(qq2.y, qq2.x);
  }

#ifdef BLEEN
  Point2 p1(x1, y1);
  Point2 q1(cos(t1), sin(t1));
  Point2 p2(x2, y2);
  Point2 q2(cos(t2), sin(t2));
#endif

  NormalXYQ n1(Parameter::input(x1), Parameter::input(y1),
               Parameter::input(t1));
  NormalXYQ n2(Parameter::input(x2), Parameter::input(y2),
               Parameter::input(t2));

  bool reversed = false;
  if (n1.getN().crossproductSign(n2.getN()) < 0) {
    n1 = NormalXYQ(Parameter::input(x2), Parameter::input(y2),
                   Parameter::input(t2));
    n2 = NormalXYQ(Parameter::input(x1), Parameter::input(y1),
                   Parameter::input(t1));
  }

#ifdef BLEEN
  Point2 p1 = n1.getP();
  Point2 q1 = n1.getN();
  Point2 p2 = n2.getP();
  Point2 q2 = n2.getN();

  Point2 q12 = q2.divides(q1);
  Point2 qMid = (q1.innerproductSign(q2) > 0 ?
                Point2(q1.x + q2.x, q1.y + q2.y) :
                Point2(q2.y - q1.y, q1.x - q2.x)).unit();
  Point2 q12o2 = qMid.divides(q1);

  Point2 pMid = (p1 + p2) / 2;
  Point2 pNorm = (p2 - p1).rot90() / 2;
  Point2 c = pMid + pNorm * (q12o2.x / q12o2.y);
  checkC(P2PV(p1), P2PV(q1), P2PV(p2), P2PV(q2), P2PV(c));
  Point2 m = (p1 - c).divides(q1);
#endif

  NormalNN nNN(&n1, &n2);
  checkC(P2PV(n1.getP()), P2PV(n1.getN()), P2PV(n2.getP()), P2PV(n2.getN()), P2PV(nNN.getP()));  

  map<int, Crossings> seeCrossings;
  int fm = fvs.size() * mvs.size();
  for (int i = 0; i < fm; i++) {
    if (sees[i] == 0)
      continue;
    
    SumEE *see = sees[i];
#ifdef BLEEN
    Point2 fc = c - see->fe->c->getC();
    Point2 mc = m - see->me->c->getC();
    Parameter mcnorm = mc.norm();
    Crossings cr;
    Point2 q[2], cq[2];
    if (!intersectCircles(see->r, fc, mcnorm, q, cq))
      continue;

    Point2 mcu = mc / mcnorm;
    for (int j = 0; j < 2; j++) {
      cr.theta[j] = cq[j].unit().divides(mcu);
      if (!between(q1, cr.theta[j], q2))
        continue;
      cr.n[j] = q[j] / see->s;
      cr.onSee[j] = see->contains(cr.theta[j], cr.n[j]);
    }
#endif

    Crossings cr(see, &nNN);
    if (cr.onSee[0] || cr.onSee[1])
      seeCrossings[i] = cr;
  }

  vector<Crossing> subCrossings;
  for (int i = 0; i < allSubs.size(); i++) {
    if (i == -1)
      cout << "this is it" << endl;

    SubE *sub = allSubs[i];
    map<int, Crossings>::iterator it = seeCrossings.find(sub->see->index());
    if (it == seeCrossings.end())
      continue;

    Crossings &crs = (*it).second;
    for (int j = 0; j < 2; j++) {
      if (!crs.onSee[j])
        continue;
      Crossing cr(sub, &crs, j);
      if (!sub->contains(&cr))
        continue;
      
      int index = subCrossings.size();
      subCrossings.push_back(cr);
      while (index > 0 && cr.between(&n1, &subCrossings[index-1])) {
        subCrossings[index] = subCrossings[index-1];
        index--;
      }
      subCrossings[index] = cr;
    }
  }

  findComponents();

  vector<int> starts, finishes;
  for (int i = 0; i < subCrossings.size(); i++) {
    Crossing &cr = subCrossings[i];
    if (cr.isOutward()) {
      cout << "outward crossing:  no path!" << endl;
      return false;
    }

    starts.push_back(i);

    while (++i < subCrossings.size() &&
           subCrossings[i].sub->component != cr.sub->component);

    if (i == subCrossings.size()) {
      cout << "unmatched crossing:  no path!" << endl;
      return false;
    }
  
    assert(subCrossings[i].isOutward());

    finishes.push_back(i);
  }

  Point2 p1 = n1.getP();
  Point2 q1 = n1.getN();
  Point2 p2 = n2.getP();
  Point2 q2 = n2.getN();
  
  if (t1 != oldT1) {
    ps.push_back(P2PV(p1));
    qs.push_back(PV(cos(oldT1), sin(oldT1)));
  }
  ps.push_back(P2PV(p1));
  qs.push_back(P2PV(q1));

  for (int i = 0; i < starts.size(); i++) {
    Crossing &startCR = subCrossings[starts[i]];
    Crossing &finishCR = subCrossings[finishes[i]];
    SubE *startSub = startCR.sub;
    SubE *finishSub = finishCR.sub;

    bool pathFound = findPath(startSub, finishSub);
    assert(pathFound);

    PV startP = startCR.getP();
    PV startQ = startCR.getQ();
    PV finishP = finishCR.getP();
    PV finishQ = finishCR.getQ();

    interpolate(ps, qs, startP, startQ);

    if (startSub != finishSub) {
      ObjAB objAB = startSub->shared(startSub->next);
      objAB = interpolate(startP, startQ, objAB, startSub->see, maxStep,
                          ps, qs);
      SubE *sub = startSub->next;
      while (sub != finishSub) {
        ObjAB nextAB = sub->shared(sub->next);
        objAB = objAB.interpolate(nextAB, sub->see, maxStep, ps, qs);
        sub = sub->next;
      }

      vector<PV> ps2, qs2;
      interpolate(finishP, finishQ, objAB, finishSub->see, maxStep, ps2, qs2);
      for (int j = ps2.size(); --j >= 0;) {
        ps.push_back(ps2[j]);
        qs.push_back(qs2[j]);
      }
    }

    ps.push_back(finishP);
    qs.push_back(finishQ);
  }

  interpolate(ps, qs, P2PV(p2), P2PV(q2));

  if (oldT2 != t2) {
    ps.push_back(P2PV(p2));
    qs.push_back(PV(cos(oldT2), sin(oldT2)));
  }

  if (reversed) {
    int n = ps.size() - 1;
    int m = ps.size() / 2;

    for (int i = 0; i < m; i++) {
      PV temp = ps[i];
      ps[i] = ps[n - i];
      ps[n - i] = temp;
      
      temp = qs[i];
      qs[i] = qs[n - i];
      qs[n - i] = temp;
    }
  }

  for (int i = 0; i < allSubs.size(); i++) {
    SubE *sub = allSubs[i];
    if (sub->a == event000)
      sub->a = 0;
    if (sub->b == event360)
      sub->b = 0;
  }
  return true;
}

void CSpace::angleRange (Point2 p, Point2 q0, Point2 qs[2]) {
  int fm = fvs.size() * mvs.size();
  for (int i = 0; i < fm; i++) {
    if (sees[i] == 0)
      continue;
    
    SumEE *see = sees[i];
    Point2 fc = see->fe->c->getC() - p;
    Point2 mc = see->me->c->getC();
    Parameter mcnorm = mc.norm();
    Crossings cr;
    Point2 q[2], cq[2];
    if (!intersectCircles(see->r, fc, mcnorm, q, cq))
      continue;
    
    Point2 mcu = mc / mcnorm;
    for (int j = 0; j < 2; j++) {
      cr.theta[j] = cq[j].unit().divides(mcu);
      cr.n[j] = q[j] / see->s;
      cr.onSee[j] = see->contains(cr.theta[j], cr.n[j]);
      if (!cr.onSee[j])
        continue;
      if (between(qs[0], cr.theta[j], q0))
          qs[0] = cr.theta[j];
      if (between(q0, cr.theta[j], qs[1]))
        qs[1] = cr.theta[j];
    }
  }
}  

void CSpace::interpolate (vector<PV> &ps, vector<PV> &qs, PV p2, PV q2) {
  PV p1 = ps.back();
  PV q1 = qs.back();
  /*
  cout << setprecision(6);
  cout << "interpolate "
       << p1.x << " " << p2.x << " "
       << p1.y << " " << p2.y << " "
       << q1.x << " " << q2.x << " "
       << q1.y << " " << q2.y << endl;
  */
  if (p1.distance(p2) + q1.distance(q2) < maxStep) {
    ps.push_back(p2);
    qs.push_back(q2);
    return;
  }

  PV q12 = q2.times(q1.conjugate());
  PV qHalf = (q1 * q2 > 0 ? q1 + q2 : (q1 - q2).rot90()).unit();
  PV q12Half = qHalf.times(q1.conjugate());

  PV pMid = (p1 + p2) / 2;
  PV pNorm = (p2 - p1).rot90();
  PV c = pMid + pNorm * (q12Half.x / q12Half.y / 2);
  checkC(p1, q1, p2, q2, c);

  PV pHalf = c + (p1 - c).times(q12Half);
  interpolate(ps, qs, pHalf, qHalf);
  interpolate(ps, qs, p2, q2);
}

void CSpace::writePath (const char *name, vector<PV> &ps, vector<PV> &qs) {
  assert(ps.size() == qs.size());

  ofstream out;
  out.open(name);
  if (!out) {
    cerr << "writePath could not open " << name << endl;
    return;
  }

  if (ps.size() == 0) {
    out << "no" << endl;
    out.close();
    return;
  }

  out << "yes" << endl;
  double prevQ;
  for (int i = 0; i < ps.size(); i++) {
    double q = atan2(qs[i].y, qs[i].x)+randomAngle;
    if (i > 0) {
      if (fabs(q - 2*M_PI - prevQ) < fabs(q - prevQ))
        q = q - 2*M_PI;
      else if (fabs(q + 2*M_PI - prevQ) < fabs(q - prevQ))
        q = q + 2*M_PI;
    }
    out << ps[i].x << " " << ps[i].y << " " << q << endl;
    prevQ = q;
  }
  out.close();
}

int quadrant[][2] = {{1, 1}, {-1, 1}, {-1, -1}, {1, -1}};
int getQuadrant (int x, int y) {
  for (int i = 0; i < 4; i++)
    if (quadrant[i][0] == x && quadrant[i][1] == y)
      return i;
  assert(0);
  return -1;
}

bool CSpace::SubE::isLive (Point2 theta) {
  if (a == 0 && b == 0)
    return true;
  if (a == 0)
    return between(b->theta, theta);
  if (b == 0)
    return between(theta, a->theta);
  int aQ = getQuadrant(a->theta.x.sign(),  a->theta.x.sign());
  int bQ = getQuadrant(b->theta.x.sign(),  b->theta.x.sign());
  int diffQ = (bQ - aQ + 4) % 4;
  int axb = 0;
  if (diffQ == 0 || diffQ == 2)
    axb = Compare()(a, b) ? 1 : -1;
  else if (diffQ == 1)
    axb = 1;
  else
    axb = -1;

  bool ret;
  if (axb > 0)
    ret = a->theta.crossproductSign(theta) > 0 && theta.crossproductSign(b->theta) > 0;
  else
    ret = a->theta.crossproductSign(theta) > 0 || theta.crossproductSign(b->theta) > 0;
  // assert(ret == between(a->theta, theta, b->theta));
  return ret;
}

CSpace::SubE *CSpace::findNearest (PV pos, double ang, PV &nearP) {
  Point2 p(pos.x, pos.y);
  Point2 theta(cos(ang), sin(ang));

  Parameter minDist;
  SubE *nearest = 0;
  for (int i = 0; i < allSubs.size(); i++) {
    if (i == -1)
      cout << "findNearest this is it" << endl;

    SubE *sub = allSubs[i];
    if (nearest != 0 && sub->see == nearest->see)
      continue;
    if (!sub->isLive(theta))
      continue;
    SumEE *see = sub->see;
    Point2 c = see->getC(theta);
    Parameter r = see->r;
    Point2 cp = see->isConcave() ? c - p : p - c;
    Parameter d = cp.norm();
    Parameter dist = r < d ? d - r : r - d;
    assert(dist.sign() > 0);

    if (nearest != 0 && minDist < dist)
      continue;

    Point2 n = cp / d;

    Point2 nt;
    Obj *t = sub->t;
    if (t->type() == SVE)
      nt = dynamic_cast<SumVE *>(t)->fv->n->getN();
    else if (t->type() == SEV)
      nt = dynamic_cast<SumEV *>(t)->mv->n->getN().times(theta);
    else
      nt = dynamic_cast<Intn *>(t)->getN(theta, see);
    
    Point2 nh;
    Obj *h = sub->h;
    if (h->type() == SVE)
      nh = dynamic_cast<SumVE *>(h)->fv->n->getN();
    else if (h->type() == SEV)
      nh = dynamic_cast<SumEV *>(h)->mv->n->getN().times(theta);
    else
      nh = dynamic_cast<Intn *>(h)->getN(theta, see);
    
    if (!see->isConcave() && !between(nt, n, nh))
      continue;

    if (see->isConcave() && !between(nh, n, nt))
      continue;

    PV cPV(P2PV(c)), nPV(P2PV(n));
    double s = see->s.value();
    nearP = cPV + nPV * s;
    minDist = dist;
    nearest = sub;
  }

  return nearest;
}  

// TEMPORARY!!!
bool operator== (const Point2 &a, const Point2 &b) {
  return (a.x.l == b.x.l && a.x.u.r == b.x.u.r &&
          a.y.l == b.y.l && a.y.u.r == b.y.u.r);
}

CSpace::ObjAB CSpace::SubE::shared (SubE *that) {
  if (this->b->theta == that->a->theta || this->a->theta == that->b->theta) {
    Event *event = this->b->theta == that->a->theta ? this->b : this->a;
    if (this->t == that->t)
      return ObjAB(this->t, event, event);
    else if (this->h == that->h)
      return ObjAB(this->h, event, event);
    assert(0);
    return ObjAB(0, 0, 0);
  }

  assert(this->t == that->h || this->h == that->t);
  Obj *th = this->t == that->h ? this->t : this->h;
  Event *maxA = this->a->index < that->a->index ? that->a : this->a;
  Event *minB = this->b->index < that->b->index ? this->b : that->b;
  assert(maxA->index < minB->index);
  return ObjAB(th, maxA, minB);
}

CSpace::ObjAB
CSpace::ObjAB::interpolate (ObjAB &next, SumEE *see, double maxStep,
                            vector<PV> &ps, vector<PV> &qs) {
  assert(a == b);
  if (obj != next.obj && !(next.a == next.b && a == next.a)) {
    ObjAB mid(next.obj, a, b);
    interpolate(mid, see, maxStep, ps, qs);
    return mid.interpolate(next, see, maxStep, ps, qs);
  }

  if (obj == next.obj && a->index >= next.a->index && b->index <= next.b->index)
    return *this;

  if (obj != next.obj) {
    assert(next.a == next.b && a == next.a);
    PV theta(P2PV(a->theta));
    PV nextP = next.obj->getP(theta);
    double dist = ps.back().distance(nextP) + qs.back().distance(theta);
    if (dist <= maxStep) {
      // if (ps.size() == 125 || ps.size() == 126)
      // cout << "1 this is it" << endl;
      ps.push_back(nextP);
      qs.push_back(theta);
      return next;
    }
    int nsteps = 1 + (int) (dist/maxStep);
    PV fc(P2PV(see->fe->c->getC()));
    PV mc(P2PV(see->me->c->getC()));
    PV c = fc + mc.times(theta);
    double r = see->r.value();
    PV u0 = (ps.back() - c).unit();
    PV un = (nextP - c).unit();
    PV ustep = un.times(u0.conjugate()).nthRoot(nsteps);
    PV u = u0;
    for (int i = 0; i < nsteps; i++) {
      u = u.times(ustep);
      PV p = c + u * r;
      // if (ps.size() == 125 || ps.size() == 126)
      // cout << "2 this is it" << endl;
      ps.push_back(p);
      qs.push_back(theta);
    }
    return next;
  }

  if (obj == next.obj) {
    assert(a->index < next.a->index || a->index > next.b->index);
    Event *event = a->index < next.a->index ? next.a : next.b;
    PV thetaN(P2PV(event->theta));
    PV nextP = obj->getP(thetaN);
    double dist = ps.back().distance(nextP) + qs.back().distance(thetaN);
    if (dist <= maxStep) {
      // if (ps.size() == 125 || ps.size() == 126)
      // cout << "3 this is it" << endl;
      ps.push_back(nextP);
      qs.push_back(thetaN);
      return ObjAB(obj, event, event);
    }
    int nsteps = 1 + (int) (dist/maxStep);
    PV theta0(P2PV(a->theta));
    PV thetaStep = thetaN.times(theta0.conjugate()).nthRoot(nsteps);
    PV theta = theta0;
    for (int i = 0; i < nsteps; i++) {
      theta = theta.times(thetaStep);
      PV p = obj->getP(theta);
      // if (ps.size() == 125 || ps.size() == 126)
      // cout << "4 this is it" << endl;
      ps.push_back(p);
      qs.push_back(theta);
    }
    return ObjAB(obj, event, event);
  }

  assert(0);
  return ObjAB(0, 0, 0);
}

void CSpace::writePath(const char *name,
                       PV startP, PV startQ, PV nearP1, SubE *start,
                       SubE *finish, PV nearP2, PV finishP, PV finishQ,
                       double maxStep) {
  ofstream out;
  out.open(name);
  if (!out) {
    cerr << "writePath could not open " << name << endl;
    return;
  }
  if (start == finish) {
    out << "yes" << endl;
    out << startP.x << " " << startP.y << " "
        << (atan2(startQ.y, startQ.x)+randomAngle) << endl;
    out << finishP.x << " " << finishP.y << " "
        << (atan2(finishQ.y, finishQ.x)+randomAngle) << endl;
    out.close();
    return;
  }

  PV theta0PV(1, 0);
  Point2 theta0(theta0PV.x, theta0PV.y);
  Event *event000 = new Event(theta0);
  event000->index = -1;
  Event *event360 = new Event(theta0);
  event360->index = events.size();
  for (int i = 0; i < allSubs.size(); i++) {
    SubE *sub = allSubs[i];
    if (sub->a == 0)
      sub->a = event000;
    if (sub->b == 0)
      sub->b = event360;
  }

  for (int i = 0; i < sees.size(); i++) {
    SumEE *see = sees[i];
    if (see == 0)
      continue;
    if (see->subs0.size() == 0)
      continue;
    vector<SubE *> &subs0 = see->subs0;
    vector<SubE *> &subs = see->getSubs();
    vector<SubE *> &allSubs = see->allSubs;
    for (int j = 1; j < subs0.size(); j++) {
      assert(subs0[j]->t->type() == I);
      Intn *intn0 = dynamic_cast<Intn *>(subs0[j]->t);
      Intn *intn = dynamic_cast<Intn *>(subs[j]->t);
      for (int k = 0; k < allSubs.size(); k++) {
        if (allSubs[k]->t == intn)
          allSubs[k]->setT(intn0);
        if (allSubs[k]->h == intn)
          allSubs[k]->setH(intn0);
      }
    }
  }

  vector<PV> ps, qs;
  // if (ps.size() == 125 || ps.size() == 126)
  // cout << "5 this is it" << endl;
  // ps.push_back(startP);
  // qs.push_back(startQ);

  interpolate(startP, startQ, nearP1, maxStep, ps, qs);

  ObjAB objAB = start->shared(start->next);
  // objAB.b = objAB.a;
  // PV theta(objAB.a->theta);
  // PV p = objAB.obj->getP(theta);
  // if (ps.size() == 125 || ps.size() == 126)
  // cout << "6 this is it" << endl;
  // ps.push_back(p);
  // qs.push_back(theta);
  objAB = interpolate(nearP1, startQ, objAB, start->see, maxStep, ps, qs);

  SubE *sub = start->next;
  while (sub != finish) {
    ObjAB nextAB = sub->shared(sub->next);
    objAB = objAB.interpolate(nextAB, sub->see, maxStep, ps, qs);
    sub = sub->next;
  }
  // ps.push_back(finishP);
  // qs.push_back(finishQ);
  vector<PV> ps2, qs2;
  interpolate(nearP2, finishQ, objAB, finish->see, maxStep, ps2, qs2);
  for (int i = ps2.size(); --i >= 0;) {
    // if (ps.size() == 125 || ps.size() == 126)
    // cout << "7 this is it" << endl;
    ps.push_back(ps2[i]);
    qs.push_back(qs2[i]);
  }
  interpolate(nearP2, finishQ, finishP, maxStep, ps, qs);

  double maxPQdist = 0;
  for (int i = 1; i < ps.size(); i++) {
    if (ps[i-1].distance(ps[i]) > maxPQdist)
      maxPQdist = ps[i-1].distance(ps[i]);
    if (qs[i-1].distance(qs[i]) > maxPQdist)
      maxPQdist = qs[i-1].distance(qs[i]);
  }
  cout << "maxPQdist " << maxPQdist << endl;  

  out << "yes" << endl;
  double prevQ;
  for (int i = 0; i < ps.size(); i++) {
    double q = atan2(qs[i].y, qs[i].x)+randomAngle;
    if (i > 0) {
      if (fabs(q - 2*M_PI - prevQ) < fabs(q - prevQ))
        q = q - 2*M_PI;
      else if (fabs(q + 2*M_PI - prevQ) < fabs(q - prevQ))
        q = q + 2*M_PI;
    }
    out << ps[i].x << " " << ps[i].y << " " << q << endl;
    prevQ = q;
  }
  out.close();

  for (int i = 0; i < allSubs.size(); i++) {
    SubE *sub = allSubs[i];
    if (sub->a == event000)
      sub->a = 0;
    if (sub->b == event360)
      sub->b = 0;
  }
}

CSpace::ObjAB CSpace::interpolate
(PV nearP, PV nearQ, ObjAB next, SumEE *see, double maxStep,
 vector<PV> &ps, vector<PV> &qs) {
  {
    PV nextP = next.obj->getP(nearQ);
    double dist = nearP.distance(nextP);
    if (dist <= maxStep) {
      // if (ps.size() == 125 || ps.size() == 126)
      // cout << "8 this is it" << endl;
      ps.push_back(nextP);
      qs.push_back(nearQ);
    }
    else {
      int nsteps = 1 + (int) (dist/maxStep);
      PV fc(P2PV(see->fe->c->getC()));
      PV mc(P2PV(see->me->c->getC()));
      PV c = fc + mc.times(nearQ);
      double r = see->r.value();
      PV u0 = (nearP - c).unit();
      PV un = (nextP - c).unit();
      PV ustep = un.times(u0.conjugate()).nthRoot(nsteps);
      PV u = u0;
      for (int i = 0; i < nsteps; i++) {
        u = u.times(ustep);
        PV p = c + u * r;
        // if (ps.size() == 125 || ps.size() == 126)
        // cout << "9 this is it" << endl;
        ps.push_back(p);
        qs.push_back(nearQ);
      }
    }
  }

  {
    Event *event = next.a;
    PV thetaN(P2PV(event->theta));
    PV nextP = next.obj->getP(thetaN);
    double dist = ps.back().distance(nextP) + qs.back().distance(thetaN);
    if (dist <= maxStep) {
      // if (ps.size() == 125 || ps.size() == 126)
      // cout << "10 this is it" << endl;
      ps.push_back(nextP);
      qs.push_back(thetaN);
    }
    else {
      int nsteps = 1 + (int) (dist/maxStep);
      PV theta0(nearQ);
      PV thetaStep = thetaN.times(theta0.conjugate()).nthRoot(nsteps);
      PV theta = theta0;
      for (int i = 0; i < nsteps; i++) {
        theta = theta.times(thetaStep);
        PV p = next.obj->getP(theta);
        // if (ps.size() == 125 || ps.size() == 126)
        // cout << "11 this is it" << endl;
        ps.push_back(p);
        qs.push_back(theta);
      }
    }
  }
  
  return ObjAB(next.obj, next.a, next.a);
}

void CSpace::interpolate (PV startP, PV startQ, PV nearP, double maxStep,
                          vector<PV> &ps, vector<PV> &qs) {
  ps.push_back(startP);
  qs.push_back(startQ);
  double dist = startP.distance(nearP);
  int nsteps = 1 + (int) (dist/maxStep);
  PV step = (nearP - startP) / nsteps;
  PV p = startP;
  for (int i = 0; i < nsteps; i++) {
    p = p + step;
    // if (ps.size() == 125 || ps.size() == 126)
    // cout << "12 this is it" << endl;
    ps.push_back(p);
    qs.push_back(startQ);
  }
}

void CSpace::writeGrids (const char *name, int n) {
  ofstream out;
  out.open(name);
  if (!out) {
    cerr << "writeGrids:  Could not open " << name << endl;
    return;
  }

  out << notDangling << endl << endl;

  out << setprecision(6);
  for (int i = 0; i < allSubs.size(); i++)
    if (!allSubs[i]->dangling)
      allSubs[i]->writeGrid(out, n);

  out.close();
}

void CSpace::SubE::writeGrid (ostream &out, int n) {
  double r = see->r.value();
  PV thetaA = a == 0 ? PV(1, 0) : P2PV(a->theta);
  PV thetaB = b == 0 ? PV(1, 0) : P2PV(b->theta);
  PV thetaS = thetaB.times(thetaA.conjugate()).nthRoot(n-1);
  PV theta = thetaA;
  for (int i = 0; i < n; i++) {
    PV fc(P2PV(see->fe->c->getC()));
    PV mc(P2PV(see->me->c->getC()));
    PV c = fc + mc.times(theta);
    double q = atan2(theta.y, theta.x)+randomAngle;
    if (q < 0)
      q += 2 * M_PI;
    if (i == n-1 && b == 0)
      q = 2 * M_PI;
    
    static double maxQ0;
    if (b == 0 && i == n-1 && fabs(q) > maxQ0) {
      cout << "maxQ0 " << fabs(q) << endl;
      maxQ0 = fabs(q);
    }

    PV nt = (t->getP(theta) - c).unit();
    PV nh = (h->getP(theta) - c).unit();
    PV ns = nh.times(nt.conjugate()).nthRoot(n-1);
    PV nj = nt;
    for (int j = 0; j < n; j++) {
      PV p = c + nj * r;
      out << p.x << " " << p.y << " " << q << " ";
      nj = nj.times(ns);
    }
    out << endl;

    theta = theta.times(thetaS);
  }

  out << endl;
}

void CSpace::SubE::drawTight () {
  // fast!
  return;
  CSpace *csp = see->fe->csp;
  vector<SubE *> tNeis, hNeis, aNeis, bNeis;

  for (int i = 0; i < neighbors.size(); i++) {
    SubE *nei = neighbors[i];
    if (nei->h == t)
      tNeis.push_back(nei);
    if (nei->t == h)
      hNeis.push_back(nei);
    if (nei->b->theta == a->theta)
      aNeis.push_back(nei);
    if (nei->a->theta == b->theta)
      bNeis.push_back(nei);
  }

  for (int i = 0; i < tNeis.size(); i++)
    for (int j = i+1; j < tNeis.size(); j++)
      if (tNeis[j]->a->index < tNeis[i]->a->index) {
        SubE *temp = tNeis[j];
        tNeis[j] = tNeis[i];
        tNeis[i] = temp;
      }
  for (int i = 1; i < tNeis.size(); i++)
    assert(tNeis[i-1]->b == tNeis[i]->a);

  for (int i = 0; i < hNeis.size(); i++)
    for (int j = i+1; j < hNeis.size(); j++)
      if (hNeis[j]->a->index < hNeis[i]->a->index) {
        SubE *temp = hNeis[j];
        hNeis[j] = hNeis[i];
        hNeis[i] = temp;
      }
  for (int i = 1; i < hNeis.size(); i++)
    assert(hNeis[i-1]->b == hNeis[i]->a);

  assert(aNeis.size() <= 2);
  if (aNeis.size() == 2 && aNeis[0]->h == h) {
        SubE *temp = aNeis[0];
        aNeis[0] = aNeis[1];
        aNeis[1] = temp;
        assert(aNeis[0]->t == t && aNeis[1]->h == h);
  }

  assert(bNeis.size() <= 2);
  if (bNeis.size() == 2 && bNeis[0]->h == h) {
        SubE *temp = bNeis[0];
        bNeis[0] = bNeis[1];
        bNeis[1] = temp;
        assert(bNeis[0]->t == t && bNeis[1]->h == h);
  }

  vector<int> tCons, hCons, aCons, bCons;

  for (int i = 0; i < aNeis.size(); i++) {
    SubE *nei = aNeis[i];
    PV theta(P2PV(a->theta));
    PV pt = nei->t == t ? t->getP(theta) : P2PV(a->p);
    PV ht = nei->h == h ? h->getP(theta) : P2PV(a->p);
    csp->horizInterp(see, a->theta, pt, ht, false, aCons);
  }

  if (aCons.size() == 0) {
    PV p(P2PV(a->p));
    PV q(P2PV(a->theta));
    int con = csp->addConfig(p, q);
    aCons.push_back(con);
  }

  for (int i = 0; i < bNeis.size(); i++) {
    SubE *nei = bNeis[i];
    PV theta(P2PV(b->theta));
    PV pt = nei->t == t ? t->getP(theta) : P2PV(b->p);
    PV ht = nei->h == h ? h->getP(theta) : P2PV(b->p);
    csp->horizInterp(see, b->theta, pt, ht, true, bCons);
  }

  if (bCons.size() == 0) {
    PV p(P2PV(b->p));
    PV q(P2PV(b->theta));
    int con = csp->addConfig(p, q, true);
    bCons.push_back(con);
  }

  for (int i = 0; i < tNeis.size(); i++) {
    SubE *nei = tNeis[i];
    Event *ai = i == 0 ? a : nei->a;
    Event *bi = i == tNeis.size()-1 ? b : nei->b;
    csp->vertInterp(t, ai->theta, bi->theta, bi==b, tCons);
  }

  tCons.front() = aCons.front();
  tCons.back() = bCons.front();

  for (int i = 0; i < hNeis.size(); i++) {
    SubE *nei = hNeis[i];
    Event *ai = i == 0 ? a : nei->a;
    Event *bi = i == hNeis.size()-1 ? b : nei->b;
    csp->vertInterp(h, ai->theta, bi->theta, bi==b, hCons);
  }

  hCons.front() = aCons.back();
  hCons.back() = bCons.back();

  double maxStep = csp->maxStep;
  vector<Config> &configs = csp->configs;
  double tDist = configs[tCons.front()].distance(configs[tCons.back()]);
  double hDist = configs[hCons.front()].distance(configs[hCons.back()]);
  double thDist = tDist > hDist ? tDist : hDist;
  int thSteps = 1 + (int) (thDist / maxStep);

  double aDist = configs[tCons.front()].distance(configs[hCons.front()]);
  double bDist = configs[tCons.back()].distance(configs[hCons.back()]);
  double abDist = aDist > bDist ? aDist : bDist;
  int abSteps = 1 + (int) (abDist / maxStep);

  int grid[abSteps * thSteps];
  PV theta0(P2PV(a->theta));
  PV thetaN(P2PV(b->theta));
  PV thetaS = thetaN.times(theta0.conjugate()).nthRoot(thSteps+1);
  PV theta = theta0;
  double r = see->r.value();
  for (int i = 0; i < thSteps; i++) {
    theta = theta.times(thetaS);

    PV fc(P2PV(see->fe->c->getC()));
    PV mc(P2PV(see->me->c->getC()));
    PV c = fc + mc.times(theta);
    PV n0 = (t->getP(theta) - c).unit();
    PV nN = (h->getP(theta) - c).unit();
    PV nS = nN.times(n0.conjugate()).nthRoot(abSteps+1);
    PV n = n0;
    for (int j = 0; j < abSteps; j++) {
      n = n.times(nS);
      PV p = c + n * r;
      grid[i * abSteps + j] = csp->addConfig(p, theta);
    }
  }

  for (int i = 1; i < thSteps; i++)
    csp->addTriangles(&grid[(i-1)*abSteps], abSteps,
                      &grid[i*abSteps], abSteps);

  csp->addTriangles(&aCons[0], aCons.size(), &grid[0], abSteps);

  csp->addTriangles(&grid[(thSteps-1)*abSteps], abSteps, &bCons[0], bCons.size());

  int side[thSteps];
  for (int i = 0; i < thSteps; i++)
    side[i] = grid[i * abSteps];
  csp->addTriangles(side, thSteps, &tCons[0], tCons.size());
    
  for (int i = 0; i < thSteps; i++)
    side[i] = grid[i * abSteps + abSteps - 1];
  csp->addTriangles(&hCons[0], hCons.size(), side, thSteps);
}

void CSpace::drawTight (double maxStep) {
  this->maxStep = maxStep;

  Point2 p(0, 0);
  PV theta0PV(1, 0);
  Point2 theta0(theta0PV.x, theta0PV.y);
  q360 = P2PV(theta0);
  Event *event000 = new Event(theta0);
  event000->p = p;
  event000->index = -1;
  Event *event360 = new Event(theta0);
  event360->index = events.size();
  event360->p = p;
  for (int i = 0; i < allSubs.size(); i++) {
    SubE *sub = allSubs[i];
    if (sub->a == 0)
      sub->a = event000;
    if (sub->b == 0)
      sub->b = event360;
  }

  for (int i = 0; i < allSubs.size(); i++)
    if (!allSubs[i]->dangling) {
      // if (i == 994)
      // cout << "drawTight this is it" << endl;
      allSubs[i]->drawTight();
    }

  ofstream out;
  out.open("tight.obj");
  if (!out) {
    cerr << "drawTight could not open tight.obj" << endl;
    return;
  }

  for (int i = 0; i < configs.size(); i++) {
    Config &config = configs[i];
    out << "v " << config.x << " " << config.y << " " << (config.z+randomAngle) << endl;
  }

  out << "s 1" << endl;
  for (int i = 0; i < triangles.size(); i++) {
    Triangle &tri = triangles[i];
    out << "f " << (tri.a+1) << " " << (tri.b+1) << " " << (tri.c+1) << endl;

    double az = configs[tri.a].z;
    double bz = configs[tri.b].z;
    double cz = configs[tri.c].z;

    double zz = fabs(az - bz);
    if (fabs(bz - cz) > zz)
      zz = fabs(bz - cz);
    if (fabs(cz - az) > zz)
      zz = fabs(cz - az);
    static double maxZZ;
    if (maxZZ < zz) {
      /* cout << "maxZZ " << zz << endl*/;
      maxZZ = zz;
    }    
  }

  out.close();

#ifdef TEST_TIGHT
  PV q000(theta0);
  double angle000 = atan2(q000.y, q000.x);
  double angle360 = angle000 + 2 * M_PI;
  cout << "size of long int " << sizeof(long int) << endl;
  set<long int> edges;
  for (int i = 0; i < triangles.size(); i++) {
    Triangle &tri = triangles[i];
    int verts[] = { tri.a, tri.b, tri.c, tri.a };
    for (int j = 0; j < 3; j++) {
      long int a = verts[j];
      long int b = verts[j+1];
      Config &ca = configs[a];
      Config &cb = configs[b];
      if (ca.z == angle000 && cb.z == angle000 ||
          ca.z == angle360 && cb.z == angle360)
        continue;
      // if (a == 6354 && b == 7040)
      // cout << "edges this is it" << endl;
      edges.insert(configs.size() * a + b);
    }
  }

  for (set<long int>::iterator it = edges.begin(); it != edges.end(); ++it) {
    long int a = *it / configs.size();
    long int b = *it % configs.size();
    long int ba = configs.size() * b + a;
    assert(edges.find(ba) != edges.end());
  }
#endif

  for (int i = 0; i < allSubs.size(); i++) {
    SubE *sub = allSubs[i];
    if (sub->a == event000)
      sub->a = 0;
    if (sub->b == event360)
      sub->b = 0;
  }
}      

void CSpace::vertInterp (Obj *obj, Point2 theta0, Point2 thetaN, bool top,
                         vector<int> &cons) {
  PV q0(P2PV(theta0));
  PV qN(P2PV(thetaN));
  PV p0 = obj->getP(q0);
  PV pN = obj->getP(qN);
  double dist = p0.distance(pN) + q0.distance(qN);
  int steps = 1 + (int)(dist/maxStep);
  PV qS = qN.times(q0.conjugate()).nthRoot(steps);
  PV q = q0;

  for (int i = 0; i < steps; i++) {
    PV p = obj->getP(q);
    int con = addConfig(p, q);
    if (!(cons.size()>0 && cons.back()==con))
      cons.push_back(con);
    q = q.times(qS);
  }

  int con = addConfig(pN, qN, top);
  cons.push_back(con);
}

void CSpace::horizInterp (SumEE *see, Point2 theta, PV p0, PV pN, bool top,
                          vector<int> &cons) {
  PV q(P2PV(theta));
  double r = see->r.value();
  PV fc(P2PV(see->fe->c->getC()));
  PV mc(P2PV(see->me->c->getC()));
  PV c = fc + mc.times(q);

  PV n0 = (p0 - c).unit();
  PV nN = (pN - c).unit();
  double dist = p0.distance(pN);
  int steps = 1 + (int)(dist/maxStep);
  PV nS = nN.times(n0.conjugate()).nthRoot(steps);
  PV n = n0;

  for (int i = 0; i < steps; i++) {
    PV p = i == 0 ? p0 : c + n * r;
    int con = addConfig(p, q, top);
    if (!(cons.size()>0 && cons.back()==con))
      cons.push_back(con);
    n = n.times(nS);
  }

  int con = addConfig(pN, q, top);
  cons.push_back(con);
}

void CSpace::addTriangles (int *a, int m, int *b, int n) {
  if (m == 0 || n == 0)
    return;
  int i = 0, j = 0;
  while (!(i == m-1 && j == n-1))
    if (i == m-1 ||
        (j != n-1 && (i+1.0)/(m-1.0) > (j+1.0)/(n-1.0))) {
      addTriangle(a[i], b[j+1], b[j]);
      j++;
    }
    else {
      addTriangle(b[j], a[i], a[i+1]);
      i++;
    }
}

void CSpace::addTriangle (int a, int b, int c) {
  int abc[] = { a, b, c };
  static double maxTriEdge;
  for (int i = 0; i < 3; i++) {
    // if (abc[i] == 1800)
    // cout << "addTriangle this is it" << endl;
    for (int j = i+1; j < 3; j++) {
      double dist = configs[abc[i]].distance(configs[abc[j]]);
      if (maxTriEdge < dist && dist > 3) {
        /* cout << "maxTriEdge " << dist << endl*/;
        maxTriEdge = dist;
      }
    }
  }
      
  if (triangles.size() == -1)
    cout << "addTriangle this is it 2" << endl;
  triangles.push_back(Triangle(a, b, c));
}

double CSpace::SubE::dZo (const Corner &c1,  const Corner &c2) {
  assert(c1.o == c2.o);
  double zo1 = c1.zo;
  if (zo1 == DBL_INF)
    zo1 = see->getN(c1.o, c1.ze).angle();
  double zo2 = c2.zo;
  if (zo2 == DBL_INF)
    zo2 = see->getN(c2.o, c2.ze).angle();
  double dzo = zo2 - zo1;
  while (dzo > M_PI)
    dzo -= 2 * M_PI;
  while (dzo < -M_PI)
    dzo += 2 * M_PI;
  return dzo;
}

double CSpace::SubE::dZoLR (const Corner &cLeft,  const Corner &cRight) {
  assert(cLeft.ze == cRight.ze);
  double zoLeft = cLeft.zo;
  if (zoLeft == DBL_INF)
    zoLeft = see->getN(cLeft.o, cLeft.ze).angle();
  double zoRight = cRight.zo;
  if (zoRight == DBL_INF)
    zoRight = see->getN(cRight.o, cRight.ze).angle();
  double dzo = zoRight - zoLeft;
  if (see->isConcave())
    while (dzo > 0) dzo -= 2 * M_PI;
  else
    while (dzo < 0) dzo += 2 * M_PI;
  return dzo;
}

double CSpace::SubE::dZe (const Corner &cLo,  const Corner &cHi) {
  double zeLo = cLo.ze;
  if (zeLo == DBL_INF)
    zeLo = P2PV(cLo.e->theta).angle();
  double zeHi = cHi.ze;
  if (zeHi == DBL_INF)
    zeHi = P2PV(cHi.e->theta).angle();
  double dze = zeHi - zeLo;
  while (dze < 0)
    dze += 2 * M_PI;
  return dze;
}

void CSpace::SubE::setZs (Corner corner) {
  if (corner.o != t) {
    Corner cornerT(corner.e, t);
    cornerT.ze = corner.ze;
    cornerT.zo = corner.zo - dZoLR(cornerT, corner);
    corner = cornerT;
  }

  if (corner.e != a) {
    Corner cornerAT(a, t);
    cornerAT.ze = corner.ze - dZe(cornerAT, corner);
    cornerAT.zo = corner.zo - dZo(cornerAT, corner);
    corner = cornerAT;
  }

  zab[0] = corner.ze;
  zabth[0][0] = corner.zo;
  Corner cornerBT(b, t);
  cornerBT.ze = zab[1] = zab[0] + dZe(corner, cornerBT);
  cornerBT.zo = zabth[1][0] = zabth[0][0] + dZo(corner, cornerBT);

  zabth[0][1] = zabth[0][0] + dZoLR(corner, Corner(a, h));
  zabth[1][1] = zabth[1][0] + dZoLR(cornerBT, Corner(b, h));
}  

void CSpace::SubE::shared (SubE *that, Corner corners[2]) {
  if (b == that->a) {
    if (t == that->t) {
      corners[1] = Corner(b, t, zab[1], zabth[1][0]);
      that->setZs(corners[1]);
      if (zabth[1][1] < that->zabth[0][1])
        corners[0] = Corner(b, h, zab[1], zabth[1][1]);
      else
        corners[0] = Corner(b, that->h, that->zab[0], that->zabth[0][1]);
    }
    else if (h == that->h) {
      corners[0] = Corner(b, h, zab[1], zabth[1][1]);
      that->setZs(corners[0]);
      if (zabth[1][0] > that->zabth[0][0])
        corners[1] = Corner(b, t, zab[1], zabth[1][0]);
      else
        corners[1] = Corner(b, that->t, that->zab[0], that->zabth[0][0]);
    }
    else
      assert(0);
    return;
  }

  if (a == that->b) {
    if (t == that->t) {
      corners[0] = Corner(a, t, zab[0], zabth[0][0]);
      that->setZs(corners[0]);
      if (zabth[0][1] < that->zabth[1][1])
        corners[1] = Corner(a, h, zab[0], zabth[0][1]);
      else
        corners[1] = Corner(a, that->h, that->zab[1], that->zabth[1][1]);
    }
    else if (h == that->h) {
      corners[1] = Corner(a, h, zab[0], zabth[0][1]);
      that->setZs(corners[1]);
      if (zabth[0][0] > that->zabth[1][0])
        corners[0] = Corner(a, t, zab[0], zabth[0][0]);
      else
        corners[0] = Corner(a, that->t, that->zab[1], that->zabth[1][0]);
    }
    else
      assert(0);
    return;
  }

  Event *aMax = a->index > that->a->index ? a : that->a;
  Event *bMin = b->index < that->b->index ? b : that->b;
  Obj *o = t == that->h ? t : h;
  int thisI = t == that->h ? 0 : 1;
  int thatI = 1 - thisI;

  if (aMax == a)
    that->setZs(Corner(aMax, o, zab[0], zabth[0][thisI]));
  else if (bMin == b)
    that->setZs(Corner(bMin, o, zab[1], zabth[1][thisI]));
  else {
    Corner thisCorner(a, o, zab[0], zabth[0][thisI]);
    Corner thatCorner(that->a, o);
    thatCorner.ze = zab[0] + dZe(thisCorner, thatCorner);
    thatCorner.zo = zabth[0][thisI] + dZo(thisCorner, thatCorner);
    that->setZs(thatCorner);
  }

  if (aMax == a)
    corners[thatI] = Corner(aMax, o, zab[0], zabth[0][thisI]);
  else
    corners[thatI] = Corner(aMax, o, that->zab[0], that->zabth[0][thatI]);

  if (bMin == b)
    corners[thisI] = Corner(bMin, o, zab[1], zabth[1][thisI]);
  else
    corners[thisI] = Corner(bMin, o, that->zab[1], that->zabth[1][thatI]);
}

void CSpace::createArcs (PolyArc *poly,
                         vector<Vert *> &verts, vector<Edge *> &edges) {
  int moving = (&verts == &mvs);
  verts.resize(poly->verts.size());
  for (int i = 0; i < verts.size(); i++)
    verts[i] = new Vert(this, poly->verts[i], moving);
  edges.resize(poly->edges.size());
  for (int i = 0; i < edges.size(); i++)
    edges[i] = new Edge(this, poly->edges[i], moving);
  for (int i = 0; i < verts.size(); i++) {
    assert(poly->verts[i]->id == i);
    PolyArc::Vert *pv = poly->verts[i];
    if (pv->t != 0)
      verts[i]->t = edges[pv->t->id];
    if (pv->h != 0)
      verts[i]->h = edges[pv->h->id];
  }
  for (int i = 0; i < edges.size(); i++) {
    assert(poly->edges[i]->id == i);
    PolyArc::Edge *pv = poly->edges[i];
    if (pv->t != 0)
      edges[i]->t = verts[pv->t->id];
    if (pv->h != 0)
      edges[i]->h = verts[pv->h->id];
  }
}

CSpace::CSpace (PolyArc *levelFixed, PolyArc *levelMoving) : nFigs(0) {
  createArcs(levelFixed, fvs, fes);
  createArcs(levelMoving, mvs, mes);
}

double btime = 0.0;
int nn1 = 0, nn2 = 0;

int main (int argc, char *argv[]) {
  double otherTime = -getTime();

  // test_type3tanh();

  if (argc < 10) {
    cout << argv[0] << " fixed.dat moving.dat x y q x y q ho hi" << endl;
    return -1;
  }

#ifdef BLEEN
  vector<PCR> pFixed;
  if (!readDat(argv[1], pFixed)) {
    cout << "Could not read fixed.dat:  " << argv[1] << endl;
    return -2;
  }
      
  vector<PCR> pMoving;
  if (!readDat(argv[2], pMoving)) {
    cout << "Could not read moving.dat:  " << argv[2] << endl;
    return -2;
  }
#endif

  double lo = atof(argv[9]);
  cout << "lo " << lo << endl;
  double hi = argc >= 11 ? atof(argv[10]) : lo;
  cout << "hi " << hi << endl;
  double scale = (lo + hi)/2;
  if (scale == lo)
    cout << "DONE!!!!" << endl;

#ifdef BLEEN
  for (int i = 0; i < pMoving.size(); i++)
    pMoving[i] = pMoving[i] * scale;
#endif

  double dmin = 1e-8;
  if (argc >= 12)
    dmin = atof(argv[11]);
  cout << "dmin " << dmin << endl;

  int seed = 1;
  if (argc >= 13)
    seed = atoi(argv[12]);
  cout << "seed " << seed << endl;
  srandom(seed);

  // CLP *clp = new CLP(dmin);

  // CLP *clp = new CLP(2.000000120e-07); // fa = 6.6908700802059684e-12
  // CLP *clp = new CLP(2.000000130e-07); // fa = -1.1842749003676545e-12

  // CLP *clp = new CLP(2.000000100e-07); // sign -1 1682
  // CLP *clp = new CLP(2.000000300e-07); // sign -1 1682

  // CLP *clp = new CLP(2.000000000e-07); // sign -1 1682
  // CLP *clp = new CLP(2.000000200e-07); // sign -1 1682

  // CLP *clp = new CLP(2.000200000e-07); // sign -1 1682
  // CLP *clp = new CLP(2.000300000e-07); // sign -1 1682
  // CLP *clp = new CLP(2.000400000e-07); // sign -1 1682
  // CLP *clp = new CLP(2.000500000e-07); // sign -1 1682
  // CLP *clp = new CLP(2.000600000e-07); // sign -1 1682
  // CLP *clp = new CLP(2.000700000e-07); // sign -1 1682
  // CLP *clp = new CLP(2.000800000e-07); // sign -1 1682
  // CLP *clp = new CLP(2.000900000e-07); // sign -1 1682

  // CLP *clp = new CLP(1.000000000e-09); // -2.6032301314593553e-13
  // CLP *clp = new CLP(1.000100000e-09); // 2.3817232908118768e-13
  // CLP *clp = new CLP(1.000200000e-09); // sign -1 1682
  // CLP *clp = new CLP(1.000300000e-09); // sign -1 1682
  // CLP *clp = new CLP(1.000400000e-09); // sign -1 1682
  // CLP *clp = new CLP(1.000500000e-09); // sign -1 1682
  // CLP *clp = new CLP(1.000600000e-09); // sign -1 1682
  // CLP *clp = new CLP(1.000700000e-09); // sign -1 1682
  // CLP *clp = new CLP(1.000800000e-09); // sign -1 1682
  // CLP *clp = new CLP(1.000900000e-09); // sign -1 1682

  // CLP *clp = new CLP(1.000000000e-08); // sign -1 1682

  // CLP *clp = new CLP(1.000000000e-10); // sign 1 1682
  // CLP *clp = new CLP(1.000100000e-10); // sign 1 1682
  // CLP *clp = new CLP(1.000200000e-10); // sign -1 1682
  // CLP *clp = new CLP(1.000300000e-10); // sign -1 1682
  // CLP *clp = new CLP(1.000400000e-10); // sign -1 1682
  // CLP *clp = new CLP(1.000500000e-10); // sign 1 1682
  // CLP *clp = new CLP(1.000600000e-10); // sign -1 1682
  // CLP *clp = new CLP(1.000700000e-10); // sign -1 1682
  // CLP *clp = new CLP(1.000800000e-10); // sign -1 1682
  // CLP *clp = new CLP(1.000900000e-10); // sign 1 1682

  // CLP *clp = new CLP(4e-10);
  // CLP *clp = new CLP();
  // CLP *clp = new CLP(1.000000000e-11);
  // CLP *clp = new CLP(1.010000000e-11); // ok
  // CLP *clp = new CLP(1.020000000e-11); // ok
  // CLP *clp = new CLP(1.030000000e-11); // ok
  // CLP *clp = new CLP(1.040000000e-11); // ok


  // CLP *clp = new CLP(1.010000000e-10); // sign -1 1682
  // CLP *clp = new CLP(0.990000000e-10); // sign 1 1682
  // CLP *clp = new CLP(1.020000000e-10); // sign -1 1682
  // CLP *clp = new CLP(0.980000000e-10); // sign -1 1682
  
  // CLP *clp = new CLP(1.000000000e-9);
  // CLP *clp = new CLP(1.000000000e-8);
  // CLP *clp = new CLP(1.000000000e-7); // ok
  // CLP *clp = new CLP(1.001000000e-10); // ok
  // CLP *clp = new CLP(1.002000000e-10); // ok
  // CLP *clp = new CLP(1.003000000e-10);


  // CLP *clp = new CLP(1.4989529149963277e-11);
  // CLP *clp = new CLP(1.4988529149963277e-11);
  // CLP *clp = new CLP(1.4987529149963277e-11);
  // CLP *clp = new CLP(1.4986529149963277e-11);
  // CLP *clp = new CLP(1.4985529149963277e-11);
  // CLP *clp = new CLP(1.4982529149963277e-11);
  // CLP *clp = new CLP(1.4981529149963277e-11);

  // CLP *clp = new CLP(1.0002000000000000e-10);
  // CLP *clp = new CLP(1.0003000000000000e-10);
  // CLP *clp = new CLP(1.0004000000000000e-10);
  // CLP *clp = new CLP(1.0005000000000000e-10);
  // CLP *clp = new CLP(1.0007000000000000e-10);

  // CLP *clp = new CLP(1.0000000000000000e-9);

  Parameter::enable();
  Parameter::signFail = true;
  Parameter::delta = dmin;

  PolyArc polyFixed(argv[1]);
  PolyArc polyMoving(argv[2], CSpace::randomAngle);
  // PolyArc polyMoving(argv[2]);
  polyFixed.drawFig("polyFixed.fig");
  polyMoving.drawFig("polyMoving.fig");
  double rFixed = scale * 2 / M_PI;
  double rMoving = scale * (1 - 2 / M_PI);
  PolyArc convFixed(&polyFixed, rFixed);
  PolyArc convMoving(&polyMoving, rMoving);
  convFixed.drawFig("convF.fig");
  convMoving.drawFig("convM.fig");
  PolyArc *levelFixed = convFixed.levelSet();
  PolyArc *levelMoving = convMoving.levelSet();
  levelFixed->drawFig("levelF.fig");
  levelMoving->drawFig("levelM.fig");

  CSpace cspace(levelFixed, levelMoving);

#ifdef BLEEN
  cspace.createArcs(pFixed, pMoving);
  cspace.checkArcs();
#endif

  cspace.drawBody(0);
  cspace.drawBody(1);

  cspace.createType1Events();
  cspace.createSums();
  cspace.createType2Events();
  cout << "nNearOrFar " << nNearOrFar << " "
       << "nTooFar " << nTooFar << endl;

  cspace.makeArrangement0();
  cspace.drawArcs(PV(1, 0));
  cspace.processEvents();
  cspace.findDangling();

  if (false) {
    PV pos1, pos2;
    double ang1, ang2;
    pos1.x = atof(argv[3]);
    pos1.y = atof(argv[4]);
    ang1 = atof(argv[5])-CSpace::randomAngle;
    pos2.x = atof(argv[6]);
    pos2.y = atof(argv[7]);
    ang2 = atof(argv[8])-CSpace::randomAngle;
    cout << "pos1 " << pos1.x << " " << pos1.y << " "
         << "ang1 " << ang1 << endl;
    cout << "pos2 " << pos2.x << " " << pos2.y << " "
         << "ang2 " << ang2 << endl;

    PV nearP1, nearP2;
    PV theta1(cos(ang1), sin(ang1));
    PV theta2(cos(ang2), sin(ang2));
    CSpace::SubE *start = cspace.findNearest(pos1, ang1, nearP1);
    CSpace::SubE *finish = cspace.findNearest(pos2, ang2, nearP2);

    double findTime = -getTime();
    bool pathFound = cspace.findPath(start, finish);
    findTime += getTime();
    cout << "findTime " << findTime << endl;

    if (pathFound) {
      cout << "Found path:" << endl;
      cout << pos1.x << " " << pos1.y << " " << ang1 << endl;
      int nPath = 0;
      for (CSpace::SubE *sub = start; sub != 0; sub = sub->next) {
        PV theta = sub->a == 0 ? PV(1, 0) : P2PV(sub->a->theta);
        PV p = sub->t->getP(theta);
        cout << p.x << " " << p.y << " " << atan2(theta.y, theta.x) << endl;
        nPath++;
      }
      cout << "path length " << nPath << endl;
      cout << pos2.x << " " << pos2.y << " " << ang2 << endl;
      cout << "Found path" << endl;
      cout << setprecision(23)
           << "r ";
      for (int i = 1; i < 9; i++)
        cout << argv[i] << " ";
      cout << scale << " " << hi << endl;
      cspace.writePath("path.txt",
                       pos1, theta1, nearP1, start,
                       finish, nearP2, pos2, theta2,
                       10 /* 0.1 */);       
    }
    else {
      cout << "Path not found." << endl;
      cout << setprecision(23)
           << "r ";
      for (int i = 1; i < 9; i++)
        cout << argv[i] << " ";
      cout << lo << " " << scale << endl;
      ofstream out;
      out.open("path.txt");
      if (!out)
        cerr << "Could not open path.txt" << endl;
      out << "no";
      out.close();
    }
  }
  else {
    PV pos1, pos2;
    double ang1, ang2;
    pos1.x = atof(argv[3]);
    pos1.y = atof(argv[4]);
    ang1 = atof(argv[5])-CSpace::randomAngle;
    pos2.x = atof(argv[6]);
    pos2.y = atof(argv[7]);
    ang2 = atof(argv[8])-CSpace::randomAngle;
    cout << "pos1 " << pos1.x << " " << pos1.y << " "
         << "ang1 " << ang1 << endl;
    cout << "pos2 " << pos2.x << " " << pos2.y << " "
         << "ang2 " << ang2 << endl;

    cspace.maxStep = 0.1;
    vector<PV> ps, qs;
    bool foundPath = cspace.findPath(pos1.x, pos1.y, ang1,
                                     pos2.x, pos2.y, ang2,
                                     ps, qs);
    cspace.writePath("path.txt", ps, qs);

    if (foundPath) {
      cout << "Found path:" << endl;
      cout << setprecision(23)
           << "r ";
      for (int i = 1; i < 9; i++)
        cout << argv[i] << " ";
      cout << scale << " " << hi << " " << dmin << " " << seed+1 << endl;
    }
    else {
      cout << "Path not found." << endl;
      cout << setprecision(23)
           << "r ";
      for (int i = 1; i < 9; i++)
        cout << argv[i] << " ";
      cout << lo << " " << scale << " " << dmin << " " << seed+1 << endl;
    }
  }

  cout << "scale " << scale << endl;
  cout << "num patches " << cspace.allSubs.size() << endl;

  // cspace.writeGrids("grids.txt", 10);
  // cspace.drawTight(1.0);
  cspace.drawTight(0.5);

  otherTime += getTime();

  cout << "otherTime " << otherTime << " "
       << "createType3EventsTime " << createType3EventsTime << endl;

  cerr << "my time " << btime << endl;
  cerr << "double calls " << nn1 << "; mprf calls " << nn2 << endl;

  return 0;
}

